/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SCINE_DFTD3PARAMETERS_H
#define SCINE_DFTD3PARAMETERS_H

#include <array>
#include <vector>

namespace Scine {
namespace Utils {

enum class ElementType : unsigned;

namespace Dftd3 {

/**
 * @class Dftd3Parameters Dftd3Parameters.h
 * @brief A class handling the parameters involved in the calculation of D3 energies and gradients.
 *
 *      Most parameters in this method are static constexpr because they stay the same for all D3 methods.
 *      The parameters a1, s8, and a2 are specified by the user.
 *      One set of parameters, the reference pairs used to calculate the C6 coefficients, are outsourced to the
 *      Dftd3ReferencePairs class, because it involves a very large number of values.
 */
class Dftd3Parameters {
 public:
  /**
   * @brief Default constructor.
   */
  Dftd3Parameters() = default;
  /**
   * @brief Default destructor.
   */
  ~Dftd3Parameters() = default;

  /**
   * @brief Getter for a covalent radius for a given element type.
   * @return double
   */
  double getCovalentRadius(ElementType elementType);
  /**
   * @brief Getter for the k1 parameter.
   * @return double
   */
  double getK1();
  /**
   * @brief Getter for the k3 parameter.
   * @return double
   */
  double getK3();
  /**
   * @brief Getter for reference pairs for one atom pair.
   * @return const std::array<std::array<double, 3>, 75>& A 75 x 3 array which contains all the reference pairs for the
   * atom pair consisting of element type 1 and element type 2. Some of the reference pairs are just placeholders.
   */
  const std::array<std::array<double, 3>, 25>& getReferencePairs(ElementType elementType1, ElementType elementType2);

  /**
   * @brief Getter for the r2r4 parameter for a given element type.
   *
   *     The name r2r4 is copied from Grimme's implementation of D3.
   *     It refers to the square root of the value Q, with Q being defined in
   *     Eq. (9) of J. Chem. Phys. 132, 154104, (2010).
   *
   * @return double
   */
  double getR2r4(ElementType elementType);
  /**
   * @brief Getter for the cut-off radius r0 in case zero-damping is used.
   * @param elementType1 The element type of the first atom.
   * @param elementType2 The element type of the second atom.
   * @return The cut-off radius r0.
   */
  double getR0Zero(ElementType elementType1, ElementType elementType2);
  /**
   * @brief Getter for the s6 parameter.
   * @return double
   */
  double getS6();
  /**
   * @brief Setter for the s6 parameter.
   */
  void setS6(double s6);
  /**
   * @brief Getter for the s8 parameter.
   * @return double
   */
  double getS8();
  /**
   * @brief Setter for the s8 parameter.
   */
  void setS8(double s8);
  /**
   * @brief Getter for the a1 parameter (BJ damping).
   * @return double
   */
  double getA1();
  /**
   * @brief Setter for the a1 parameter.
   */
  void setA1(double a1);
  /**
   * @brief Getter for the a2 parameter (BJ damping).
   * @return double
   */
  double getA2();
  /**
   * @brief Setter for the a2 parameter.
   */
  void setA2(double a2);
  /**
   * @brief Getter for the a parameter (zero damping).
   * @return double
   */
  double getA();
  /**
   * @brief Setter for the a parameter.
   */
  void setA(double a);
  /**
   * @brief Getter for the sr parameter (zero damping).
   * @return double
   */
  double getSr();
  /**
   * @brief Setter for the sr parameter.
   */
  void setSr(double sr);

 private:
  // The following parameters have to be specified by the user as they depend on the method chosen.
  // General scaling parameters
  double s6_;
  double s8_;
  // Parameters for BJ damping
  double a1_;
  double a2_;
  // Parameters for zero damping
  double sr_;
  double a_;

  // All the following parameters for have been taken from Grimme's D3 implementation.
  static constexpr double covalentRadii_[94] = {
      0.80628308, 1.15903197, 3.02356173, 2.36845659, 1.94011865, 1.88972601, 1.78894056, 1.58736983, 1.61256616,
      1.68815527, 3.52748848, 3.14954334, 2.84718717, 2.62041997, 2.77159820, 2.57002732, 2.49443835, 2.41884923,
      4.43455700, 3.88023730, 3.35111422, 3.07395437, 3.04875805, 2.77159820, 2.69600923, 2.62041997, 2.51963467,
      2.49443835, 2.54483100, 2.74640188, 2.82199085, 2.74640188, 2.89757982, 2.77159820, 2.87238349, 2.94797246,
      4.76210950, 4.20778980, 3.70386304, 3.50229216, 3.32591790, 3.12434702, 2.89757982, 2.84718717, 2.84718717,
      2.72120556, 2.89757982, 3.09915070, 3.22513231, 3.17473967, 3.17473967, 3.09915070, 3.32591790, 3.30072128,
      5.26603625, 4.43455700, 4.08180818, 3.70386304, 3.98102289, 3.95582657, 3.93062995, 3.90543362, 3.80464833,
      3.82984466, 3.80464833, 3.77945201, 3.75425569, 3.75425569, 3.72905937, 3.85504098, 3.67866672, 3.45189952,
      3.30072128, 3.09915070, 2.97316878, 2.92277614, 2.79679452, 2.82199085, 2.84718717, 3.32591790, 3.27552496,
      3.27552496, 3.42670319, 3.30072128, 3.47709584, 3.57788113, 5.06446567, 4.56053862, 4.20778980, 3.98102289,
      3.82984466, 3.85504098, 3.88023730, 3.90543362};

  static constexpr double r2r4_[94] = {
      2.00734898,  1.56637132,  5.01986934, 3.85379032, 3.64446594, 3.10492822,  2.71175247,  2.59361680, 2.38825250,
      2.21522516,  6.58585536,  5.46295967, 5.65216669, 4.88284902, 4.29727576,  4.04108902,  3.72932356, 3.44677275,
      7.97762753,  7.07623947,  6.60844053, 6.28791364, 6.07728703, 5.54643096,  5.80491167,  5.58415602, 5.41374528,
      5.28497229,  5.22592821,  5.09817141, 6.12149689, 5.54083734, 5.06696878,  4.87005108,  4.59089647, 4.31176304,
      9.55461698,  8.67396077,  7.97210197, 7.43439917, 6.58711862, 6.19536215,  6.01517290,  5.81623410, 5.65710424,
      5.52640661,  5.44263305,  5.58285373, 7.02081898, 6.46815523, 5.98089120,  5.81686657,  5.53321815, 5.25477007,
      11.02204549, 10.15679528, 9.35167836, 9.06926079, 8.97241155, 8.90092807,  8.85984840,  8.81736827, 8.79317710,
      7.89969626,  8.80588454,  8.42439218, 8.54289262, 8.47583370, 8.45090888,  8.47339339,  7.83525634, 8.20702843,
      7.70559063,  7.32755997,  7.03887381, 6.68978720, 6.05450052, 5.88752022,  5.70661499,  5.78450695, 7.79780729,
      7.26443867,  6.78151984,  6.67883169, 6.39024318, 6.09527958, 11.79156076, 11.10997644, 9.51377795, 8.67197068,
      8.77140725,  8.65402716,  8.53923501, 8.85024712};

  static constexpr double r0Zero_[94][94] = {
      {4.1239491107191473, 3.5048746601312519, 5.4964570698563353, 4.4503046466386076, 4.7509601504192878,
       4.6283166342640101, 4.4724142794906516, 4.1135554702189978, 3.9015281776180757, 3.7590428507728544,
       5.6499025985146778, 4.7486925535300752, 5.4359857169303734, 5.4582845790082217, 5.3486803081481380,
       4.9491921819133537, 4.6783943466898235, 4.4833747065766607, 6.0749019782082128, 5.2150768154526279,
       5.1704790912969312, 5.0705124451253338, 4.9849082971922272, 5.5236691336184407, 4.8169114916153664,
       4.7103308334561573, 4.6613871447478390, 4.6326630493624084, 4.6133877999854356, 4.5954353530059846,
       5.3860967831845885, 5.5229131178067776, 5.5066614811306813, 5.1589518776250634, 4.9182009401844855,
       4.7407552885991660, 6.2158754987029967, 5.4397648948971362, 5.4796381960486036, 5.3923327872663673,
       5.8401982661693870, 5.7260588015308773, 5.5159210979133357, 5.4004588308773043, 5.3365858573446348,
       5.2919885837347147, 5.2621309151775648, 4.8044390329060320, 5.6206117165433875, 5.8156318072023252,
       5.8379302187343978, 5.4974018643480260, 5.2545721086982633, 5.0735360578262094, 6.4190211824270955,
       5.7882305142144164, 5.7886085221202483, 5.5661875890191768, 5.5263147384134852, 5.4888978128312580,
       5.4571505552907276, 5.4252140685243928, 5.3985692418939228, 5.0640876623635265, 4.7600305379761387,
       5.7600736559606203, 5.3169330506076067, 5.2972797933248028, 5.2849965638412719, 5.2717685398660512,
       5.3658767393159250, 5.4815282355777608, 5.3968684315905691, 5.3120198489233497, 5.2300056497312024,
       5.0856305086297118, 5.2713905319602192, 5.2009035464399398, 5.1404326440597536, 4.7838414316773132,
       5.6412097683178821, 5.8883859390660422, 5.9367628411885018, 5.6056828822648122, 5.3645539369184032,
       5.1801167159854176, 6.4543588392543612, 5.7727348933499822, 6.3967223203492471, 5.8471898355170531,
       5.7045155047189162, 5.6455561834162795, 5.6047380877731214, 5.4354189303445146},
      {3.5048746601312519, 3.2781075372046704, 4.8626427889427806, 4.7422673202224921, 4.5196576084413929,
       4.2569858168287977, 4.0304076978551322, 3.8975597704255094, 3.7588538468199388, 3.6379113662409028,
       4.9883094672531580, 5.1629202848176297, 5.1829515500062655, 5.0276159818187667, 4.8259823297179931,
       4.6859536037148999, 4.5389328578183648, 4.4009829500244573, 5.3507591263573229, 5.7802936998292838,
       5.6247693529617573, 5.4915434176263025, 5.3845852021070391, 4.9554281859951335, 5.2199900171368849,
       5.1279601853504193, 5.0472692391015705, 4.9771602614871222, 4.9144213116719468, 4.8620760023569218,
       5.0897876945023057, 5.0716464688428280, 4.9644990240977602, 4.8887208289873918, 4.7914002381566139,
       4.6891659950958040, 5.4785046228768852, 5.9824945890616918, 5.8492686537262379, 5.7007363267521534,
       5.2908545600172197, 5.1612194745142768, 5.0601192551709602, 4.9764042456754591, 4.9057284814751521,
       4.8416667292624549, 4.7855417914348903, 5.0268599660071036, 5.3021929950090598, 5.3432003198780214,
       5.2755477178328141, 5.2228244006119571, 5.1417554464572763, 5.0531272352775183, 5.6922327257811611,
       6.2873072787149669, 6.1722230195847665, 6.1359401177200343, 6.0926656465076352, 6.0554379501512123,
       6.0263358468599506, 5.9974234233402672, 5.9730457430532340, 5.6977127140512778, 5.6130529100640860,
       5.5905652693062100, 5.8488906458204069, 5.8213005741524695, 5.7993801705262289, 5.7893645379319105,
       5.6844846900760562, 5.6075729217939694, 5.4896538286429211, 5.3894984037912961, 5.3059726235215985,
       5.2163996178501497, 4.8847524217948255, 4.8233367249229486, 4.7588969648044204, 4.9329405452372566,
       5.2687449271651756, 5.3823176052178257, 5.3526487153407043, 5.3307278611686870, 5.2704457374685285,
       5.1946675423581610, 5.7678216916657252, 6.2750240492314360, 6.4643744718486795, 6.0612959463271601,
       6.0316275069958145, 5.9855177512167925, 5.9416764934185347, 5.9137084138447662},
      {5.4964570698563353, 4.8626427889427806, 6.6057260002492937, 5.6317613728551992, 5.6648314327932523,
       6.0344618904708867, 5.4393873375370809, 4.9879314593473270, 4.7405665099191383, 4.5712469019447557,
       6.8088716839733934, 5.8679771165173511, 6.1937658658509518, 7.0889291359795790, 6.4122179412136813,
       5.9207004337382081, 5.6293045467401823, 5.4348516932128792, 7.3177748518424570, 6.3256685482431081,
       6.1368849122117251, 6.1629634028021121, 6.0168874513972677, 6.4554928629718553, 5.8564499028454833,
       5.8001357357921162, 5.7778368737142678, 5.7657428734565404, 5.7712228617266570, 5.7477904264770903,
       6.1353733311341747, 7.1558252716673474, 6.5707659007820842, 6.1361293469458378, 5.8825279428900945,
       5.7143423586332061, 7.6322247156384426, 6.7695651694077847, 6.5240898089629784, 6.4209107714104769,
       6.8876730412388616, 6.7691871615019537, 6.5998680040733460, 6.5522466672167745, 6.5450658686433041,
       6.5416642480365974, 6.5482782600242082, 6.1520029757161030, 6.5819155570938959, 7.5931074302986383,
       7.0414970283488101, 6.6265128307038168, 6.3895415217757776, 6.2219227241047479, 7.9823915958048524,
       7.6405399884751821, 6.8846494285379851, 6.8234225103461368, 6.7742895924120141, 6.7313926785596712,
       6.6918973853140358, 6.6516465268025131, 6.6234896685487170, 6.1578609718920507, 6.2480007641493591,
       6.9766792603244507, 6.5450658686433041, 6.5231450144712877, 6.5091609746844030, 6.4900749545332346,
       6.2684100372438261, 7.2138397984782934, 7.2423746646379206, 7.0826926813520243, 6.9400183505538866,
       6.8100052571451117, 6.8944762824522750, 6.7542583272233792, 6.6040251899459408, 6.1858290514658192,
       6.6677089342528060, 7.6985549658319039, 7.1308808047944554, 6.7174086387727874, 6.4872401205123866,
       6.3067713067719673, 8.0390828696724999, 7.2059029840931608, 6.9927421183205194, 6.6427644673799131,
       6.6380404949214604, 6.6085608342701416, 6.5753015451062851, 6.7270462634612738},
      {4.4503046466386076, 4.7422673202224921, 5.6317613728551992, 5.8547490925421304, 5.5716680278350683,
       5.7096179356289767, 5.2269820370303268, 5.0357415748839269, 5.0848744928180487, 4.8136991002344622,
       5.8995355953778539, 6.2289147439981898, 6.1157200738513708, 5.9739905375449247, 5.8946219431478211,
       5.7712228617266570, 5.9465896951027917, 5.6321393807610303, 6.5424202638482605, 6.8434537755347717,
       6.6147968383519213, 6.4929093380083067, 6.4675868632295828, 5.4800162039544347, 6.2837168794282316,
       6.1792150394782084, 6.1145860501338767, 6.0903975990726469, 6.0446663017452318, 5.9789037392291817,
       6.0321942935816733, 5.9964781783028007, 6.0044154432337091, 5.9501800943895260, 6.2102062812070766,
       5.9171100344514729, 6.8219104787228106, 7.1221879745976597, 6.8232332811203333, 6.6826377685313805,
       5.8088285659889110, 5.7585620748830708, 5.6026592695639366, 5.5265035170935128, 5.4862522080362144,
       5.5093070859257249, 5.4346629145328516, 6.1996238620268995, 6.3256685482431081, 6.3088501249811522,
       6.3339833705340727, 6.3073380933578269, 6.6061040081551257, 6.3090389036611798, 7.1596049001798860,
       7.5675966273856625, 7.1975881618021962, 7.2796023609943434, 7.2363274392361694, 7.1956985728188148,
       7.1635733073724523, 7.1306920261144278, 7.1049915434298718, 6.6692209658761321, 6.8474221827273380,
       6.8011236482682884, 6.9683644380334862, 6.9352943780954330, 6.9128062867917812, 6.8937202666406128,
       6.5909859446507468, 6.7758011734895645, 6.6975666028099559, 6.6338828585030898, 6.5770023554096388,
       6.4617288670536350, 5.6997915322604626, 5.5762032216134942, 5.4293717049427626, 6.1036256230478685,
       6.3292589475298433, 6.3492902127184792, 6.3848170987715482, 6.3865179090749020, 6.7261014689695831,
       6.4156195618203888, 7.2221546207692571, 7.5059912507422064, 7.3321368995351728, 6.9275463423903281,
       6.9683644380334862, 6.9381287615705052, 6.8980666817390102, 7.0798582978769522},
      {4.7509601504192878, 4.5196576084413929, 5.6648314327932523, 5.5716680278350683, 6.0773588043232296,
       5.5805496367118916, 5.2489028912023432, 5.0043723252492276, 4.9573182255242898, 4.7231808495255470,
       7.1091491798482425, 5.9207004337382081, 6.7895964345964206, 6.3504242364359733, 6.0218006530815247,
       5.7921989214069827, 5.6253361395476160, 5.4977798722538580, 7.3364828640877953, 6.4917757648365884,
       6.4600280567502812, 6.3377629990466113, 6.1928210713592611, 5.7744357036533360, 6.0161314355856046,
       5.9095507774263965, 5.8407650527552457, 5.7882305142144164, 5.7617744662639740, 5.7062163150222691,
       6.7393294929448038, 6.4008799567676169, 6.1529477702077937, 5.9841949488192698, 5.8636307014189537,
       5.7702780672349663, 7.4432527514728068, 6.7580379557359178, 6.8147296801493411, 6.6883069860273006,
       6.2353395267599971, 6.0730123892248313, 5.8989683582462193, 5.8075057635913891, 5.7506252604979382,
       5.7075391174197918, 5.6788154725801361, 5.9307160663325256, 6.9660968411442736, 6.6856613812322561,
       6.4706104759304583, 6.3269913506406308, 6.2196551272155354, 6.1365069043058931, 7.5681634139715213,
       7.8527560597561328, 7.1639508647325085, 6.9211211090827458, 6.8776574086445432, 6.8379733367188802,
       6.7981000355674128, 6.7655967622152184, 6.7397075008506357, 6.6879289781214695, 6.5227670065654557,
       7.3474432911738035, 6.6115839964252423, 6.5813487705080362, 6.5605614895077382, 6.5433650583399512,
       6.6820709819455217, 6.8948542903581069, 6.6285916489130017, 6.4904529624390657, 6.3825495018823357,
       6.2538592108710827, 5.9086059829347057, 5.8110961628781244, 5.7065943229281011, 5.8751579150908206,
       6.9840492881237237, 6.7459435049324146, 6.5444990820574453, 6.4222335738079988, 6.3283141530381526,
       6.2504575902643751, 7.6416735616469014, 7.1195423698026152, 7.5373609509226815, 7.2355718739702821,
       7.0683310842050844, 7.0135289487750425, 6.9568372243616183, 6.7385739276789174},
      {4.6283166342640101, 4.2569858168287977, 6.0344618904708867, 5.7096179356289767, 5.5805496367118916,
       5.4996694612372394, 5.1141653747893390, 4.8560287769551707, 4.6808511728048394, 4.5525388896994174,
       6.1387749517408823, 6.0471231278602486, 6.2032142613136347, 6.2975116899893120, 5.9044487970621109,
       5.6463121992279426, 5.4514813377948075, 5.2987913744023531, 6.3797146678614878, 6.5628290863969516,
       6.4964997372950419, 6.3986119093326286, 6.3118737376820278, 6.2942992986084088, 6.1280033033349017,
       6.0269030839915851, 5.9596289403979847, 5.9082279750288738, 5.8691106896890695, 5.8284818232717157,
       6.1493573709210594, 6.3576050350094429, 6.0414539103643286, 5.8418986259269641, 5.6895866704403408,
       5.5665655969250079, 6.5148301921803231, 6.7631403866459783, 6.7620063629284841, 6.6709217761794850,
       6.6635517483802120, 6.5533806909342687, 6.2247571075798200, 6.1293261057324235, 6.0688547528064616,
       6.0235014633848785, 5.9889193718234992, 5.9974234233402672, 6.3526918333251858, 6.6208436132078967,
       6.6958657925066021, 6.1667425807688740, 6.0259582894998944, 5.9118188248613848, 6.6866061757239468,
       7.0724882700776783, 7.0639846691066861, 6.9160191287184611, 6.8786026536820097, 6.8445873487064901,
       6.8177532928502167, 6.7905412290881113, 6.7657859914410228, 6.5955215889749486, 6.4229895896196618,
       6.4645637010744830, 6.6486229141016366, 6.6234896685487170, 6.6047812057576039, 6.5936319999915671,
       6.6414416649823913, 6.6781025747529554, 6.5960883755608082, 6.3232121726738679, 6.1247909119539976,
       5.9923209924302068, 6.0707447923356188, 5.6990355164487996, 5.9445108768936068, 5.6582174208056415,
       6.3574162563294152, 6.6832050056630159, 6.7884624108789264, 6.2670872348463034, 6.1389637304209099,
       6.0301154753724884, 6.7793920233220755, 7.0583154516107660, 7.3774901889567568, 7.0787242741594572,
       6.9812144541028758, 6.9337823464721078, 6.8929642508289497, 6.7682423670102629},
      {4.4724142794906516, 4.0304076978551322, 5.4393873375370809, 5.2269820370303268, 5.2489028912023432,
       5.1141653747893390, 4.9558061939009646, 4.6952132204975552, 4.5136103830396417, 4.3796288824382996,
       5.5476683554538671, 5.5909432772120411, 5.8413318393411053, 5.8564499028454833, 5.7570500432597447,
       5.4904098444545841, 5.2821621803661998, 5.1156769558668893, 5.8101513683864336, 6.1391525091009376,
       6.0629967566305139, 5.9620853159672258, 5.8717562944841140, 5.7740576957475049, 5.7181215365999689,
       5.6045493090930938, 5.5246139281101314, 5.4630085514666753, 5.4127420603608343, 5.3692783599226326,
       5.7770813084483805, 5.9086059829347057, 5.8972675479428656, 5.6893978917603132, 5.5231018964868062,
       5.3849632100128701, 5.9696441224465264, 6.3404086038416549, 6.3235897300339232, 6.2285367360923587,
       6.1527585409819903, 6.0423987048560193, 5.8534262901446077, 5.7330503708785434, 5.6487690253429585,
       5.5837620280927958, 5.5317947266836009, 5.5552271619331686, 5.9730457430532340, 6.1584277584779104,
       6.1931986287193173, 6.0112182339013476, 5.8562606736196798, 5.7260588015308773, 6.1790258102524049,
       6.6386072815073192, 6.6316152616138773, 6.5027357413768208, 6.4636184560370165, 6.4286588071155819,
       6.4006907275418135, 6.3729118771938493, 6.3496682206243102, 6.2147419255312784, 6.0673431717289112,
       5.9726677351474020, 6.2241903209939604, 6.1981122809493501, 6.1782697944407419, 6.1654197783713522,
       6.2000018699327315, 6.2043482850311289, 6.1219560779331506, 6.0416426890443562, 5.9679433121431735,
       5.8320722225584500, 5.6773038915025866, 5.6019037042980493, 5.5323615132694606, 5.5102518804174156,
       5.9654869365739325, 6.2085054709037228, 6.2765356303089863, 6.1159088525313985, 5.9743685454507558,
       5.8507806853495632, 6.2901416621900390, 6.6455993014007610, 6.9388847773821682, 6.6435204831915762,
       6.5577271060326661, 6.5093502039102074, 6.4670200766437231, 6.3492902127184792},
      {4.1135554702189978, 3.8975597704255094, 4.9879314593473270, 5.0357415748839269, 5.0043723252492276,
       4.8560287769551707, 4.6952132204975552, 4.6897327816816627, 4.4429346188393337, 4.2653006391197623,
       5.1198345922852591, 5.8245134160791494, 5.5499359523430796, 5.5363299204620269, 5.4286156891311004,
       5.4433557447296472, 5.1886203169564098, 4.9862310995897490, 5.4456233416188597, 6.2738900255139418,
       6.1958446840601367, 6.1038148522736719, 6.0253910523682599, 5.3872308069020827, 5.9006691685495731,
       5.8528590530129732, 5.7956005420136902, 5.7169879634282497, 5.6593514445231357, 5.1846519097638435,
       5.4892758207370900, 5.5824392256952731, 5.5569279722365215, 5.6336509618385806, 5.4223796850493207,
       5.2496589070140063, 5.6336509618385806, 6.4258239730947340, 6.3976671148409379, 6.3026141208993733,
       5.7670656758540630, 5.6633198517157020, 5.6232573213384311, 5.5675103914166986, 5.5081730622082308,
       5.2984138170422970, 5.2184779855135588, 5.3723019726235082, 5.6822170931868436, 5.8211117954724418,
       5.8373634321485390, 5.9390308886234902, 5.7436332406044963, 5.5805496367118916, 5.8738351126932988,
       6.6469221037982837, 6.6433317045115485, 6.5497902916475335, 6.5193653859587490, 6.4929093380083067,
       6.4709884838362894, 6.4492568588900765, 6.4309264040047944, 6.2969449034034533, 6.1291373270523959,
       6.0354066849625774, 6.3456998134317439, 6.3271805798664342, 6.3111177218703647, 6.3011020892760472,
       6.2941100693826053, 6.2651971953171461, 6.1831834466707747, 6.1087280539579289, 6.0439102859335696,
       6.0072498267087813, 5.5882976724169966, 5.3551055414557203, 5.2583512866650262, 5.3524599366606767,
       5.6795714883917992, 5.8655202904023351, 5.9120076035414124, 6.0388083055692841, 5.8581502626030613,
       5.7022479078297028, 5.9870297828401178, 6.6658193452694245, 6.9063815040299747, 6.6512685188966811,
       6.5928759841799041, 6.5528139043484090, 6.5178538048811987, 6.4296036016072726},
      {3.9015281776180757, 3.7588538468199388, 4.7405665099191383, 5.0848744928180487, 4.9573182255242898,
       4.6808511728048394, 4.5136103830396417, 4.4429346188393337, 4.3456135774627800, 4.1471927672886864,
       4.8719024057254359, 5.4012148466889673, 5.4858746506761582, 5.4186000565367820, 5.2277380528419899,
       5.1646210951209834, 5.0797725124537640, 4.8603751920535680, 5.2065727639358599, 5.8418986259269641,
       5.7833173125301593, 5.6943110934445702, 5.6151317282732709, 5.1750147356211329, 5.4862522080362144,
       5.4382533138195868, 5.3811835815003315, 5.3260034381644576, 5.2738473580752361, 5.2171551831160370,
       5.4153876651558788, 5.3866640203162239, 5.3520819287548447, 5.3479242923364749, 5.3118306196975462,
       5.2587292945708572, 5.4019704119548537, 5.9945885893194193, 5.9991237830978452, 5.9123856114472444,
       5.5546603753473089, 5.4512925591147798, 5.3843959728812347, 5.3182558530051285, 5.2536273142065726,
       5.1914551509772577, 5.1332513949405083, 5.3842071942012071, 5.6011476884863871, 5.6869410656452972,
       5.6266589419451387, 5.6449893968304199, 5.6293045467401823, 5.5939668899129167, 5.6534929978014121,
       6.2141746883996429, 6.2449771514484835, 6.1255464772198849, 6.0949332433968486, 6.0679099583147709,
       6.0456110962369225, 6.0238794712907096, 6.0053602377253998, 5.9012364056812077, 5.6869410656452972,
       5.6644534248874203, 5.9195664100207139, 5.9021812001728984, 5.8859295634968012, 5.8747799071849895,
       5.9072831805371830, 5.8853623263651667, 5.8063717398738950, 5.7313500111209654, 5.6640754169815892,
       5.6196669220516968, 5.3284602642794745, 5.2706345161485570, 5.2107304003542296, 5.3573731383449337,
       5.5856520676219530, 5.6602962390148264, 5.6969566982396147, 5.7390975962802946, 5.7436332406044963,
       5.7220903943383110, 5.7761365139566898, 6.2521584005677289, 6.5431762796599227, 6.2908976780017021,
       6.2198439058955630, 6.1765694346831639, 6.1404753114984603, 6.0238794712907096},
      {3.7590428507728544, 3.6379113662409028, 4.5712469019447557, 4.8136991002344622, 4.7231808495255470,
       4.5525388896994174, 4.3796288824382996, 4.2653006391197623, 4.1471927672886864, 4.0391000775061521,
       4.6938904181000325, 5.1249365726495437, 5.2513592667715843, 5.2101636137683709, 5.0882761134247563,
       4.9758374590895995, 4.8618867731311184, 4.7505821425134558, 5.0276159818187667, 5.5563611856506627,
       5.5155430900075046, 5.4303164994344533, 5.3515146916232101, 5.0342299938063766, 5.2196120092310530,
       5.1699123047110715, 5.1124645644859861, 5.0576624290559433, 5.0060731355525805, 4.9644990240977602,
       5.1850299176696746, 5.2498476856940339, 5.2101636137683709, 5.1559282649241878, 5.0888429000106150,
       5.0138207207119097, 5.2313284521287242, 5.7092399277231447, 5.7411764144894795, 5.6629418438098709,
       5.4155764438359064, 5.3120198489233497, 5.2228244006119571, 5.1476134426332241, 5.0782604808304379,
       5.0134431633518544, 4.9533498183317235, 5.1447790591581519, 5.3719239647176771, 5.4807722197660977,
       5.4820950221636195, 5.4494025195856226, 5.4012148466889673, 5.3426330827463868, 5.4909766310404438,
       5.9295820426150314, 5.9868410041600901, 5.8475678434228842, 5.8163873724682125, 5.7891753087061071,
       5.7661208813623723, 5.7440107979645516, 5.7251135564934108, 5.6480130095312964, 5.4008368387831354,
       5.4133088469466939, 5.6374305903511193, 5.6213677323550497, 5.6051160956789534, 5.5930220954212260,
       5.6578394128998104, 5.6438553731129257, 5.5684556364541651, 5.4934334571554597, 5.4244585032585055,
       5.3719239647176771, 5.1515818498257904, 5.0895989158222781, 5.0278047604987943, 5.1430782488547981,
       5.3654987314100939, 5.5164878844991954, 5.5493691657572199, 5.5393535331629025, 5.5113859041349098,
       5.4667881799792140, 5.6211789536750221, 5.9815493440242262, 6.3060152909603042, 6.0601619226096659,
       5.9796597550408448, 5.9346844735250919, 5.8976455558486975, 5.7606404425464799},
      {5.6499025985146778, 4.9883094672531580, 6.8088716839733934, 5.8995355953778539, 7.1091491798482425,
       6.1387749517408823, 5.5476683554538671, 5.1198345922852591, 4.8719024057254359, 4.6938904181000325,
       7.0074817233732904, 6.1548373591911751, 6.4101395735502722, 7.2053361975073011, 6.5399634377332436,
       6.0652643535197264, 5.7776480950342401, 5.5779040319168480, 7.5097708792547442, 6.6087496129501702,
       6.4320604277222886, 6.4296036016072726, 6.2956221010059314, 6.6637405270602397, 6.1397197462325730,
       6.0860511839742495, 6.0592171281179752, 6.0410759024584975, 6.0376742818517899, 6.0114070125813752,
       6.3394638093499642, 7.2707207521175201, 6.7000234289249718, 6.2850396818257543, 6.0391863134751160,
       5.8700559347265360, 7.8287563873749306, 7.0375286211562438, 6.8066040870841800, 6.9275463423903281,
       7.0891179146596066, 6.9789468572136633, 6.8190760952477385, 6.7712659797111385, 6.7606835605309614,
       6.7531247540516608, 6.7546363351292102, 6.4014467433534765, 6.7812816123054560, 7.7231214247989648,
       7.1900293553228956, 6.7895964345964206, 6.5603727108277106, 6.3929426918367085, 8.1900724733073762,
       7.9168182625146057, 7.1565812874790105, 7.0891179146596066, 7.0405517833113436, 6.9989781224022991,
       6.9600496157425225, 6.9209323304027182, 6.8935314879605851, 6.4273360047180601, 6.4728780728196709,
       7.2070370078106549, 6.8177532928502167, 6.7975332489815532, 6.7837379878746971, 6.7659747701210504,
       6.5650971338319399, 7.4867164519110094, 7.4916296535952664, 7.3364828640877953, 7.1941865411954895,
       7.0634174319750516, 7.1244560220326481, 6.9780020627219725, 6.8200208897394292, 6.4239343841113525,
       6.8572485860958521, 7.8295124031865937, 7.2805471554860341, 6.8833266261404633, 6.6644965428719027,
       6.4870513418323590, 8.2443073716057835, 7.4644175898331611, 7.2554139099331145, 6.9356723860012650,
       6.9294359313737095, 6.9061922748041713, 6.8642406059892949, 6.9936869128122101},
      {4.7486925535300752, 5.1629202848176297, 5.8679771165173511, 6.2289147439981898, 5.9207004337382081,
       6.0471231278602486, 5.5909432772120411, 5.8245134160791494, 5.4012148466889673, 5.1249365726495437,
       6.1548373591911751, 6.5836163673972496, 6.4193991903329266, 6.2863624842232761, 6.2330719298707837,
       6.7402747379822703, 6.2859844763174451, 5.9726677351474020, 6.8392961391164020, 7.1796357148227461,
       6.9464435838614698, 6.8298477436537190, 6.7990448300591035, 5.7215231572066756, 6.6348276529947805,
       6.5320266233481110, 6.5231450144712877, 6.4466112540950320, 6.4103283522302998, 6.3477786316409288,
       6.3324717894565223, 6.2995905081984978, 6.3332273547224096, 6.2704888554530109, 6.5516798806309149,
       6.2631188276537371, 7.1641400939583120, 7.4840708471159649, 7.1692425248683724, 7.0371506132504127,
       6.1028700577819812, 6.0614851755529635, 5.8985908008861632, 5.8182774119973697, 5.7765140713167451,
       5.7795376840176207, 5.7211451493008445, 6.5970331700524989, 6.6554257047692751, 6.6306704671221866,
       6.6824489898513528, 6.6486229141016366, 6.9698760191110365, 6.6816929740396898, 7.5400065557177252,
       7.9975092087634545, 7.5445417494961511, 7.6339259764875722, 7.5910295131810051, 7.5509669828037342,
       7.5180857015457088, 7.4848264123818522, 7.4589371510172686, 7.0193869449509902, 7.2106274070973901,
       7.1686752877367379, 7.3240108559242367, 7.2916963612520709, 7.2693979497199983, 7.2487998979455037,
       6.9305699550912045, 7.2622167006007530, 7.1904073632287266, 7.1673524853392152, 7.0660634873158710,
       6.9432311924805656, 6.1314049239416084, 6.0119742497130106, 5.7789708974317620, 6.5148301921803231,
       6.6675201555727783, 6.6745121754662202, 6.7278022792729368, 6.7208102593794941, 7.0825039026719958,
       6.7809036043996249, 7.6059578969138038, 7.8858261196941859, 7.6222099841356767, 7.2656183212074597,
       7.3177748518424570, 7.2903735588545482, 7.2518235106463793, 7.4551579730505058},
      {5.4359857169303734, 5.1829515500062655, 6.1937658658509518, 6.1157200738513708, 6.7895964345964206,
       6.2032142613136347, 5.8413318393411053, 5.5499359523430796, 5.4858746506761582, 5.2513592667715843,
       6.4101395735502722, 6.4193991903329266, 7.5789350623775018, 7.0118281384716887, 6.6408748783965326,
       6.3653530707145478, 6.3422981928250364, 6.0631855353105415, 6.9105386899025687, 6.9490891886565143,
       6.9137515318292477, 6.7990448300591035, 6.6624177246627179, 6.2342059535882788, 6.5061373619835274,
       6.4054151505460428, 6.3481561890009841, 6.3290701688498157, 6.2958108796859591, 6.2445995940884274,
       7.5281013341400262, 7.0598270326883164, 6.7722107742028292, 6.5586719005243568, 6.4284695778897785,
       6.3398413667100204, 7.2234774231667798, 7.2754447245759737, 7.3126724209323966, 7.1941865411954895,
       6.7355503149780418, 6.6032691741342777, 6.4262019810005651, 6.3526918333251858, 6.3179209630837789,
       6.2935432827967457, 6.2831496422965971, 6.5233337931513153, 7.8072139916545211, 7.3897734184402868,
       7.1329596230036403, 6.9407743663655497, 6.8228557237602772, 6.7410303032481576, 8.8038554890322960,
       7.8848817757482710, 7.6792797164549311, 7.4135838615956855, 7.3704981690633149, 7.3306248679118475,
       7.2903735588545482, 7.2559806965189733, 7.2304694430602217, 7.1488332517739055, 7.0407410125371470,
       7.0072929446932628, 7.1083931640365794, 7.0791022820652891, 7.0596382540082887, 7.0394182101396252,
       7.1492112596797375, 7.5551241686763282, 7.3663405326449443, 7.2308474509660527, 7.1182195674050934,
       6.9783800706278045, 6.6592053332818137, 6.5248458247746415, 6.3772582922922467, 6.4919645435166160,
       7.8444416880109440, 7.4602599534147913, 7.2057142054131331, 7.0314813957544926, 6.9258455320869752,
       6.8459101511040128, 8.8716959686659802, 7.6747436215849545, 8.0073356121319694, 7.6231543280815917,
       7.5475653621970267, 7.4903068511977438, 7.4449531112303848, 7.2688307125883629},
      {5.4582845790082217, 5.0276159818187667, 7.0889291359795790, 5.9739905375449247, 6.3504242364359733,
       6.2975116899893120, 5.8564499028454833, 5.5363299204620269, 5.4186000565367820, 5.2101636137683709,
       7.2053361975073011, 6.2863624842232761, 7.0118281384716887, 7.1652736671300303, 6.6930314090315299,
       6.3657306280746031, 6.1470893234860702, 5.9889193718234992, 7.4806692265092574, 7.3090820216456622,
       7.2427526725437525, 6.6697877524619908, 6.6102611940277196, 7.1282351999994109, 6.3957775258575564,
       6.2846616739199224, 6.2211667082930848, 6.1777030078548831, 6.1576721932120231, 6.1130744690563263,
       6.9676084222218240, 7.2312254588718847, 6.8364617556413299, 6.5664199362294617, 6.3882187193782549,
       6.2587724125553397, 7.6292015534833428, 7.5294241365375481, 7.5175189149598491, 7.0231665734635280,
       7.1446760659013115, 6.7946984149607053, 6.5749235372004540, 6.4715552704221491, 6.4154307831403603,
       6.3774470709722744, 6.3545814223085673, 6.3509910230218320, 7.2036353872039474, 7.5273453183283632,
       7.1641400939583120, 6.9156411208126292, 6.7470775286499087, 6.6233004393229136, 7.8004107504411069,
       7.9052905982969621, 7.8200635571781358, 7.6649172182164396, 7.6292015534833428, 7.5955647069594301,
       7.5681634139715213, 7.5386837533202034, 7.5160068833365239, 6.9286803661078231, 6.8279577041245618,
       7.3992218139029697, 7.4118830512923317, 7.3888286239485961, 7.3725769872724998, 7.3585929474856151,
       6.9517347934515579, 7.2062809919989919, 7.1359832357045159, 6.9931196756805756, 6.8527133923174270,
       6.7107946267851766, 6.5565930823151719, 6.4602172859760847, 6.3640302683170260, 6.3188657575754696,
       7.2299026564743620, 7.6025562763070971, 7.2469098584163465, 7.0123949250575475, 6.8527133923174270,
       6.7304478840679804, 7.8835589733507492, 7.8410400674042373, 8.1301679069672748, 7.4661184001365148,
       7.7131057922046473, 7.6686972972747540, 7.2397290598428761, 7.5349041248076647},
      {5.3486803081481380, 4.8259823297179931, 6.4122179412136813, 5.8946219431478211, 6.0218006530815247,
       5.9044487970621109, 5.7570500432597447, 5.4286156891311004, 5.2277380528419899, 5.0882761134247563,
       6.5399634377332436, 6.2330719298707837, 6.6408748783965326, 6.6930314090315299, 6.6172532139211615,
       6.2701108475471790, 6.0312494990899825, 5.8547490925421304, 6.8385401233047389, 6.7667307859327135,
       6.6843385788347343, 6.5826715729055589, 6.4961217293892100, 6.6172532139211615, 6.3445657897142498,
       6.2258911312973142, 6.1561601615886978, 6.1075944807862106, 6.0728231599990279, 6.0361627007742404,
       6.5879627824956470, 6.7525575169200254, 6.7682423670102629, 6.4777917250497037, 6.2772916461206494,
       6.1266805009373790, 7.0031353082748931, 6.9986001144964671, 6.9687424459393181, 6.8725554282802586,
       6.9789468572136633, 6.8746342464894434, 6.6418196728882233, 6.5204994096762432, 6.4500124241559638,
       6.4006907275418135, 6.3657306280746031, 6.2428987837850736, 6.8058480712725178, 7.0252453916727138,
       7.0885511280737470, 6.8185093086618798, 6.6251900283062941, 6.4787365195413944, 7.2045801816956381,
       7.3565141292764302, 7.2926416062895374, 7.1545024692698256, 7.1153851839300213, 7.0789135033852615,
       7.0494338427339427, 7.0193869449509902, 6.9948204859839285, 6.7837379878746971, 6.6701657603678228,
       6.7860060353096854, 6.8729334361860897, 6.8460993803298162, 6.8273909175387031, 6.8118952966742690,
       6.8324933484487635, 6.9262235399928063, 6.8438313328948279, 6.7646519677235277, 6.6900077963306543,
       6.5410974614507378, 6.4743901044429970, 6.3940767155542027, 6.3211333544646831, 6.1996238620268995,
       6.8154856959610042, 7.0881731201679159, 7.1800137227285772, 6.9237667138777903, 6.7383846984531139,
       6.5940095573516233, 7.3017119938463884, 7.3364828640877953, 7.6167295453197843, 7.2915075825720432,
       7.1911629284946139, 7.1424080184663232, 7.1017791520489686, 7.0069149367874317},
      {4.9491921819133537, 4.6859536037148999, 5.9207004337382081, 5.7712228617266570, 5.7921989214069827,
       5.6463121992279426, 5.4904098444545841, 5.4433557447296472, 5.1646210951209834, 4.9758374590895995,
       6.0652643535197264, 6.7402747379822703, 6.3653530707145478, 6.3657306280746031, 6.2701108475471790,
       6.2563155864403228, 5.9569828850571644, 5.7356964262193628, 6.4093835577386091, 7.2268790437734864,
       7.1176527808192338, 7.0184421504592995, 6.9379399828904775, 6.1890414428467224, 6.8115172887684370,
       6.7449987104407239, 6.0775475830032573, 6.0066830401229216, 5.9556600826596426, 5.9131416272589066,
       6.3096061407928152, 6.4173203721237417, 6.4065487237177612, 6.4594612701644225, 6.2011358936502257,
       6.0068718188029502, 6.6017575930567274, 7.4003558376204639, 7.3200424487316704, 7.2115722015890809,
       6.5616955132252324, 6.4609728512419720, 6.3884074980582835, 6.3198105520671604, 6.2580163967436766,
       6.0524143374503367, 5.9845729567251018, 6.1123184532446642, 6.5199326230903836, 6.6750789620520798,
       6.7075822354042733, 6.7892184266905895, 6.5420422559424285, 6.3534473985910731, 6.8387293525305424,
       7.6617048268355363, 7.5898950389177351, 7.5296129152175766, 7.4969208631853546, 7.4676299812140643,
       7.4440083167386941, 7.4201978735832963, 7.4005446163004915, 7.1955093435930113, 7.0989443180281206,
       6.5777579206755261, 7.3032240254697145, 7.2820587365635845, 7.2646735267157689, 7.2518235106463793,
       7.1887065529253729, 7.1781241337451958, 7.0923303060405098, 7.0186309291393272, 6.9555144219640965,
       6.4415088231849715, 6.3320937815506912, 6.0873739863717713, 5.9811717866641700, 6.0790596146265825,
       6.5246565955488371, 6.7281802871787679, 6.7877068456130392, 6.8944762824522750, 6.6563704992609658,
       6.4706104759304583, 6.9419083900830438, 7.6603820244380145, 7.8249772094081687, 7.5388725320002310,
       7.4948420449761697, 7.4574255699397192, 7.4217099052066215, 7.3856157820219170},
      {4.6783943466898235, 4.5389328578183648, 5.6293045467401823, 5.9465896951027917, 5.6253361395476160,
       5.4514813377948075, 5.2821621803661998, 5.1886203169564098, 5.0797725124537640, 4.8618867731311184,
       5.7776480950342401, 6.2859844763174451, 6.3422981928250364, 6.1470893234860702, 6.0312494990899825,
       5.9569828850571644, 5.8668430927998569, 5.6181553409741465, 6.1376409280233881, 6.7701324065394202,
       6.6726221359370630, 6.5728447189912691, 6.4904529624390657, 5.9276924536316500, 6.3604394184845150,
       6.2950548638742960, 6.2326939219649526, 6.1756241896456983, 6.1159088525313985, 6.0680991875405743,
       6.1251689198598296, 6.1935766366251483, 6.1612625924987583, 6.1504909440927777, 6.1125076824704676,
       5.8914095517669178, 6.3360621887432576, 6.9449320027839194, 6.8882402783704961, 6.7812816123054560,
       6.3035589153910641, 6.1998130912527030, 6.1143972714538490, 6.0403198866468344, 5.9745577746765592,
       5.9140864217505973, 5.7795376840176207, 6.2359067638916317, 6.3302041925673098, 6.4443436572058195,
       6.4532252660826428, 6.4673980845495551, 6.4511464478734579, 6.2362847717974637, 6.5858839642864622,
       7.2032573792981163, 7.1579040898765331, 7.0777794796677664, 7.0454654355413764, 7.0161745535700861,
       6.9923641104146883, 6.9687424459393181, 6.9487111807506823, 6.7671087938385446, 6.6319932695197084,
       6.5924979762740730, 6.8513905899199044, 6.8307925381454098, 6.8132180990717908, 6.8011236482682884,
       6.7676755804244042, 6.7569039320184237, 6.6701657603678228, 6.5940095573516233, 6.5280582161555447,
       6.4630516694511568, 6.0410759024584975, 5.9764473636599407, 5.9070944018571554, 6.1750574030598386,
       6.3268025719606031, 6.4893189387215715, 6.5269246429838264, 6.5658526990978272, 6.5690655410245062,
       6.3572270271036118, 6.6962438004124332, 7.2187530001625504, 7.4260558697592440, 7.1407076587087452,
       7.0849607287870127, 7.0439534039180511, 7.0080485099591501, 6.9458767972756101},
      {4.4833747065766607, 4.4009829500244573, 5.4348516932128792, 5.6321393807610303, 5.4977798722538580,
       5.2987913744023531, 5.1156769558668893, 4.9862310995897490, 4.8603751920535680, 4.7505821425134558,
       5.5779040319168480, 5.9726677351474020, 6.0631855353105415, 5.9889193718234992, 5.8547490925421304,
       5.7356964262193628, 5.6181553409741465, 5.5081730622082308, 5.9401644617952085, 6.4545480684801646,
       6.3683762328696476, 6.2682208080180217, 6.1837502332566343, 5.7513808257638255, 6.0495795034294888,
       5.9847621859509053, 5.9222120148157584, 5.8649535038164755, 5.8112853921039278, 5.7700892885549386,
       5.9919429845243748, 6.0310602698641791, 5.9813605653441977, 5.9248576196108020, 5.8570166894313429,
       5.7850181228335131, 6.1400977541384041, 6.6282136410071706, 6.5932539920857360, 6.4885629229099084,
       6.1281920820149294, 6.0206670799098063, 5.9214564495498712, 5.8417098472469364, 5.7727348933499822,
       5.7096179356289767, 5.6531149898955810, 5.9501800943895260, 6.1939546445309794, 6.2789924564240023,
       6.2695436104155444, 6.2355287559858006, 6.1873406325433695, 6.1323497184332991, 6.3984231306526009,
       6.8812482584770542, 6.8627285743659687, 6.7655967622152184, 6.7330934888630249, 6.7036138282117070,
       6.6796141558305049, 6.6559929419009105, 6.6357724474864712, 6.4751461202546601, 6.3056377336002489,
       6.3111177218703647, 6.5375070621640035, 6.5176645756553953, 6.4999013579017486, 6.4872401205123866,
       6.4819489109222985, 6.4641856931686510, 6.3763134978005560, 6.2977009192151163, 6.2289147439981898,
       6.1618293790846170, 5.8294266177634064, 5.7649868576448782, 5.6969566982396147, 5.9140864217505973,
       6.1818606442732529, 6.3165981606862571, 6.3381410069524424, 6.3290701688498157, 6.3007240813702161,
       6.2587724125553397, 6.5155862079919862, 6.9129955160175856, 7.1539356826839668, 6.8719881911486240,
       6.8075488815758707, 6.7637071732318370, 6.7270462634612738, 6.6423869100198578},
      {6.0749019782082128, 5.3507591263573229, 7.3177748518424570, 6.5424202638482605, 7.3364828640877953,
       6.3797146678614878, 5.8101513683864336, 5.4456233416188597, 5.2065727639358599, 5.0276159818187667,
       7.5097708792547442, 6.8392961391164020, 6.9105386899025687, 7.4806692265092574, 6.8385401233047389,
       6.4093835577386091, 6.1376409280233881, 5.9401644617952085, 8.0145164107054381, 7.3255224370017862,
       7.1408964373887729, 7.0740003017010045, 6.9653408253326106, 7.1569592953848424, 6.8238005182519679,
       6.7661635488010781, 6.7285582950845990, 6.6996454210191407, 6.6835825630230712, 6.6497564872733559,
       6.8211549134569234, 7.5358489192993554, 6.9989781224022991, 6.6359616767122747, 6.4131627357053720,
       6.2515911634360934, 8.3813129354537761, 7.7764110780599056, 7.5320697413325925, 7.5265893025167010,
       7.6118158930897515, 7.5305577097092673, 7.3971434462395607, 7.3550025481988808, 7.3432861053012095,
       7.3313808837235097, 7.3270344686251123, 7.0692758786967751, 7.3009564285805011, 8.0695073248155094,
       7.5832810269301243, 7.2168634111791690, 7.0071037154674594, 6.8425085304973052, 8.7930838406263163,
       8.7632261720691655, 7.8879049379033708, 7.8170403950230352, 7.7652618722938689, 7.7236882113848244,
       7.6830588944216940, 7.6435631506302828, 7.6167295453197843, 7.1291799944911016, 7.1601716867657457,
       7.8113716280728909, 7.5350933540334681, 7.5143060730331701, 7.4989996813945394, 7.4827480447184431,
       7.2716655466092108, 8.3263211202521550, 8.2796454789788250, 8.1348923299715032, 7.9903284101899850,
       7.8474648501660447, 7.8416068539900969, 7.6624608426471994, 7.4640395819273291, 7.0851495074670403,
       7.3769234023708972, 8.1878044258723879, 7.6824921078358344, 7.3236328480184056, 7.1314475913803150,
       6.9570260030416469, 8.8450515925812851, 8.2055676436260345, 7.9192746380838459, 7.8281896007890719,
       7.6397839726635199, 7.6146507271105994, 7.5740214101474690, 7.7155621677738875},
      {5.2150768154526279, 5.7802936998292838, 6.3256685482431081, 6.8434537755347717, 6.4917757648365884,
       6.5628290863969516, 6.1391525091009376, 6.2738900255139418, 5.8418986259269641, 5.5563611856506627,
       6.6087496129501702, 7.1796357148227461, 6.9490891886565143, 7.3090820216456622, 6.7667307859327135,
       7.2268790437734864, 6.7701324065394202, 6.4545480684801646, 7.3255224370017862, 7.7544902238878883,
       7.5218653300582474, 7.4876612464027001, 7.3670965484566073, 6.2092614867153859, 7.2265010358676554,
       7.1263456110160295, 7.1019679307289962, 7.0409297912171755, 7.0140957353609013, 6.9558924298699276,
       6.8559257836983303, 7.3351600616902726, 6.8564925702841899, 7.4515675737637714, 7.0422525936146974,
       6.7567151533383951, 7.7149953811880287, 8.1025782858451141, 7.7817022876499937, 7.6588699928146884,
       6.7253459037036958, 6.6338828585030898, 6.4596500488444502, 6.3708330589846645, 6.3719670827021586,
       6.2997792868785254, 6.2548040053627734, 7.2550359020272825, 7.2248002255643016, 7.7267113735399242,
       7.2508782656089128, 7.9113378236987133, 7.5160068833365239, 7.2374614629536635, 8.1511435161018237,
       8.7433841361063340, 8.1571907415035767, 8.2356145414089887, 8.1913952757048989, 8.1518995319134877,
       8.1173174403521084, 8.0825470206564773, 8.0566577592918946, 7.6169187745455877, 7.8041899284078697,
       7.7709306392440132, 7.9251326342597936, 7.8939521633051219, 7.8718420799073021, 7.8502996841868917,
       7.5303689310292388, 8.0912398508532739, 8.0914281789875258, 7.9778564020264264, 7.8797788942924347,
       7.7435302473476559, 6.9160191287184611, 6.6730001438428941, 6.3980451227467698, 7.1951313356871802,
       7.2491779058513348, 7.2353826447444787, 7.2949092031787499, 8.0232092409022346, 7.6218315256840690,
       7.3300580813259879, 8.2174728652037352, 8.5247423809720253, 8.1692851923070791, 7.8646612813338317,
       7.9275890098290347, 7.9035902385393841, 7.8690081469780058, 8.0842473804140553},
      {5.1704790912969312, 5.6247693529617573, 6.1368849122117251, 6.6147968383519213, 6.4600280567502812,
       6.4964997372950419, 6.0629967566305139, 6.1958446840601367, 5.7833173125301593, 5.5155430900075046,
       6.4320604277222886, 6.9464435838614698, 6.9137515318292477, 7.2427526725437525, 6.6843385788347343,
       7.1176527808192338, 6.6726221359370630, 6.3683762328696476, 7.1408964373887729, 7.5218653300582474,
       7.3516009275921732, 7.2151626008758152, 7.1654628963558338, 6.0735791758106910, 7.0089937549966166,
       6.9043026858207899, 6.8846494285379851, 6.8200208897394292, 6.7918640314856331, 6.7340382833547157,
       6.9655300545584140, 6.7671087938385446, 6.7761791813953955, 7.3374281091252618, 6.9347271409637985,
       6.6559929419009105, 7.5084480768572224, 7.8572912535345587, 7.6401624311151268, 7.4802912186034263,
       6.5452546473233326, 6.4874293497381901, 6.3058265122802766, 6.2120963207362339, 6.2204106924814218,
       6.1400977541384041, 6.0907756069784789, 7.0199537315368490, 7.2727991197809292, 7.0978107448564023,
       7.1316368206061185, 7.7694195087122395, 7.3718209714608367, 7.0945983534754991, 7.9856039871857556,
       8.4565234428867342, 8.0114932485503392, 8.0031779757135997, 7.9589587100095098, 7.9190854088580425,
       7.8811016966899565, 7.8467088343543816, 7.8206312448555462, 7.5141172943531425, 7.5948086911477679,
       7.5377389588285126, 7.6872165308400637, 7.6554683722079817, 7.6333591899017126, 7.6114383357296962,
       7.4292687116859231, 7.8380160041575859, 7.7527898641303103, 7.7074361241629514, 7.6042575371562258,
       7.4687640049315585, 6.6350168822205839, 6.4711777130620938, 6.2204106924814218, 6.9628844497633704,
       7.2903735588545482, 7.1473216706963560, 7.1756673076301798, 7.1352272198928528, 7.4772676059025507,
       7.1868169639419914, 8.0630825420537011, 8.2760546291463157, 8.0623265262420389, 7.7365377769084382,
       7.8359371859484011, 7.7945527542651591, 7.7586478603062590, 7.8444416880109440},
      {5.0705124451253338, 5.4915434176263025, 6.1629634028021121, 6.4929093380083067, 6.3377629990466113,
       6.3986119093326286, 5.9620853159672258, 6.1038148522736719, 5.6943110934445702, 5.4303164994344533,
       6.4296036016072726, 6.8298477436537190, 6.7990448300591035, 6.6697877524619908, 6.5826715729055589,
       7.0184421504592995, 6.5728447189912691, 6.2682208080180217, 7.0740003017010045, 7.4876612464027001,
       7.2151626008758152, 7.1133063657208364, 7.0622838588033332, 5.9955333838111100, 6.8967438793414884,
       6.7892184266905895, 6.7714552089369420, 6.7034250495316794, 6.6731893730686984, 6.6142296012202859,
       6.7213770459653537, 6.6828269977571848, 6.6764017644496017, 7.2382174787653266, 6.8340049295263139,
       6.5531914617084652, 7.4179302766940829, 7.7945527542651591, 7.4763228114108600, 7.3746553549359088,
       6.6129067988227641, 6.3867066877549297, 6.1988678462152365, 6.1002244529869367, 6.1123184532446642,
       6.0250130444624279, 5.9715341619756837, 6.8859722309355078, 7.0326149689262110, 7.1218099666918286,
       7.0156077669842274, 7.6609488110238733, 7.2582482934081858, 6.9768680390044784, 7.8062687466170546,
       8.3401168319047869, 7.9049130409369059, 7.9389274448208740, 7.8977313412718848, 7.8599368583296023,
       7.8280003715632684, 7.7945527542651591, 7.7697970660722948, 7.3491441014771572, 7.4782124003942414,
       7.4725431828983213, 7.6401624311151268, 7.6099263041063701, 7.5887614657460158, 7.5679741847457178,
       7.2574927281422994, 7.6753113092623648, 7.6144614978847960, 7.5400065557177252, 7.4343706920502077,
       7.2996331756372035, 6.4626736615453257, 6.3133853187595781, 6.0762247806057355, 6.8226664945344737,
       7.0507566451314654, 7.1637620860524809, 7.0596382540082887, 7.0224105576518658, 7.3652069594732259,
       7.0700318945084373, 7.8782677637606602, 8.1942301097257477, 7.9695411291896869, 7.6469647712369895,
       7.6562243880196448, 7.6263667194624949, 7.5887614657460158, 7.7970091298344002},
      {4.9849082971922272, 5.3845852021070391, 6.0168874513972677, 6.4675868632295828, 6.1928210713592611,
       6.3118737376820278, 5.8717562944841140, 6.0253910523682599, 5.6151317282732709, 5.3515146916232101,
       6.2956221010059314, 6.7990448300591035, 6.6624177246627179, 6.6102611940277196, 6.4961217293892100,
       6.9379399828904775, 6.4904529624390657, 6.1837502332566343, 6.9653408253326106, 7.3670965484566073,
       7.1654628963558338, 7.0622838588033332, 7.0318594036603237, 6.0807599743841605, 6.8659414162926478,
       6.7604947818509338, 6.7446207025348928, 6.6758349778637420, 6.6450325148149023, 6.5864507508723218,
       6.5830491302656142, 6.6253792575320984, 6.5921199683682419, 7.1582820977823642, 6.7519907303341657,
       6.4687208869470769, 7.3030347962439102, 7.6754996373966167, 7.4043242448130302, 7.2956647684446372,
       6.4834609425456238, 6.4543588392543612, 6.1754354109656697, 6.0783035988149203, 6.0989016505894149,
       6.0097066528237972, 5.9564160984713057, 6.8374065501330206, 6.8950430690381346, 6.9317039788086987,
       7.0207097473485121, 6.8644293846693225, 7.1703760980400908, 6.8854054443496482, 7.6896729064093048,
       8.2123708848394497, 7.7680967063147168, 7.8238436362364494, 7.7817022876499937, 7.7427742315359929,
       7.7100817289579959, 7.6772008982457463, 7.6513116368811627, 7.2733663569125646, 7.4147178853131797,
       7.3542465323872177, 7.5175189149598491, 7.4859604360993464, 7.4640395819273291, 7.4436303088328630,
       7.1864389560361603, 7.5549353899963005, 7.4908736377836034, 7.4373943047510833, 7.3153180257274411,
       7.1813365251260999, 6.3889742846441422, 6.2001906486127591, 5.9722897272415709, 6.7230778562687075,
       6.9116727136200629, 6.9744116634352373, 7.0660634873158710, 6.9339715756979112, 7.2794131317685400,
       6.9804588888369894, 7.7594038761179212, 8.0787678426897145, 7.8389612491950524, 7.5477545914228301,
       7.5730770662015541, 7.5449197574019822, 7.5071252744597006, 7.6668068071998210},
      {5.5236691336184407, 4.9554281859951335, 6.4554928629718553, 5.4800162039544347, 5.7744357036533360,
       6.2942992986084088, 5.7740576957475049, 5.3872308069020827, 5.1750147356211329, 5.0342299938063766,
       6.6637405270602397, 5.7215231572066756, 6.2342059535882788, 7.1282351999994109, 6.6172532139211615,
       6.1890414428467224, 5.9276924536316500, 5.7513808257638255, 7.1569592953848424, 6.2092614867153859,
       6.0735791758106910, 5.9955333838111100, 6.0807599743841605, 6.5033029785084553, 5.8384974558660332,
       5.7468456319853995, 5.6996023030346592, 5.6725794684983573, 5.6619970493181802, 5.6421545628095728,
       6.2111510756987673, 7.1992889721055500, 6.7631403866459783, 6.3797146678614878, 6.1457665210885484,
       5.9874077907459498, 7.4294574903659507, 6.5902303793848604, 6.4254459651889029, 6.3632742525053629,
       6.8364617556413299, 6.8092496918792245, 6.5896631422532250, 6.4980117689183672, 6.4573824519552367,
       6.4345168032915296, 6.4294143723814692, 6.0004465854953670, 6.5618842919052609, 6.8795474481737005,
       7.1244560220326481, 6.7457547262523869, 6.5129406031969417, 6.3491014340384506, 7.7372937927201013,
       7.2909407959861836, 6.7759904027153679, 6.6860393891380880, 6.6403080918106729, 6.5987339803558518,
       6.5605614895077382, 6.5252238326804726, 6.4942321404058285, 6.1805378418757311, 6.1750574030598386,
       6.7829824226088098, 6.4035255615626614, 6.3785810946897694, 6.3640302683170260, 6.3439990031283902,
       6.2289147439981898, 6.8959878635298253, 6.8423197518172776, 6.8655634083868167, 6.7064482116867792,
       6.5684983038928708, 6.6862281678181157, 6.5728447189912691, 6.4579492385410964, 6.0569495312287627,
       6.6473001117041148, 6.9662856198243013, 7.2106274070973901, 6.8323041192229601, 6.6006235693392332,
       6.4269579968122281, 7.7941742958135523, 7.0114501305658568, 6.9585380346649721, 6.6945429901090803,
       6.6178204510527969, 6.5813487705080362, 6.5511130940450553, 6.5885295690815067},
      {4.8169114916153664, 5.2199900171368849, 5.8564499028454833, 6.2837168794282316, 6.0161314355856046,
       6.1280033033349017, 5.7181215365999689, 5.9006691685495731, 5.4862522080362144, 5.2196120092310530,
       6.1397197462325730, 6.6348276529947805, 6.5061373619835274, 6.3957775258575564, 6.3445657897142498,
       6.8115172887684370, 6.3604394184845150, 6.0495795034294888, 6.8238005182519679, 7.2265010358676554,
       7.0089937549966166, 6.8967438793414884, 6.8659414162926478, 5.8384974558660332, 6.7013462313224945,
       6.5923091975940453, 6.5785139364871892, 6.5031137492826518, 6.4675868632295828, 6.4056039292260705,
       6.4239343841113525, 6.4108951388161595, 6.4428316255824933, 7.0331822060578464, 6.6236784472287447,
       6.3358729595174541, 7.1501560541714282, 7.5301801523492111, 7.2395402811628484, 7.1163299784217120,
       6.2258911312973142, 6.1863958380516788, 5.9930770082418690, 5.8891419548777053, 5.9135196351647386,
       5.8120409573698151, 5.7496800154604717, 6.6537248944659222, 6.7325267022771653, 6.7215662751911571,
       6.7676755804244042, 6.7200542435678319, 7.0363945974387496, 6.7465107420640500, 7.5305577097092673,
       8.0470196840576325, 7.6099263041063701, 7.6823028786100309, 7.6394064153034638, 7.5997214422862491,
       7.5670298407998029, 7.5335813224101429, 7.5082592981771947, 7.1044247568440131, 7.2716655466092108,
       7.2057142054131331, 7.3714429635550056, 7.3389396902028112, 7.3166408281249629, 7.2958539976704406,
       7.0116393597916611, 7.3383729036169525, 7.2686419339083352, 7.2438866962612467, 7.1427860263721543,
       7.0112613518858291, 6.1552153670970071, 6.0291706808807977, 5.8080725501772488, 6.5768131261838354,
       6.7449987104407239, 6.7638959519118655, 6.8134068777518184, 6.7918640314856331, 7.1488332517739055,
       6.8455321431981808, 7.5978318533028677, 7.9317466462474044, 7.6932637562418149, 7.3608605443748285,
       7.3997890510346052, 7.3712541848749771, 7.3325144568952290, 7.5105268950664072},
      {4.7103308334561573, 5.1279601853504193, 5.8001357357921162, 6.1792150394782084, 5.9095507774263965,
       6.0269030839915851, 5.6045493090930938, 5.8528590530129732, 5.4382533138195868, 5.1699123047110715,
       6.0860511839742495, 6.5320266233481110, 6.4054151505460428, 6.2846616739199224, 6.2258911312973142,
       6.7449987104407239, 6.2950548638742960, 5.9847621859509053, 6.7661635488010781, 7.1263456110160295,
       6.9043026858207899, 6.7892184266905895, 6.7604947818509338, 5.7468456319853995, 6.5923091975940453,
       6.5046257809059780, 6.4891301600415439, 6.4074939687552277, 6.3681874541896200, 6.3046924885627824,
       6.3260465561489401, 6.3018581050877103, 6.3254797695630804, 6.3131965400795504, 6.5547034933317905,
       6.2667092269404723, 7.0798582978769522, 7.4228434783783399, 7.1272904055077202, 6.9986001144964671,
       6.1181768999663877, 6.0745239703023817, 5.9566053276971092, 5.8997243740578824, 5.7763252926367175,
       5.7595068693747615, 5.6878863106827628, 6.5458218844549672, 6.6304812378963831, 6.6121512335568768,
       6.6493784793675239, 6.6629845112485766, 6.9557032006441242, 6.6643073136460993, 7.4485439610628958,
       7.9175733772347172, 7.4976768789970176, 7.5764786868082616, 7.5337705516359463, 7.4940860291645066,
       7.4615827558123131, 7.4285126958742600, 7.4031902210955360, 6.9889624898079807, 7.2106274070973901,
       7.1119835633233146, 7.2684527046825318, 7.2359494313303383, 7.2119502094949119, 7.1915409364004450,
       6.9003342786282236, 7.1860609481303284, 7.1144403894383306, 7.0518902183031837, 6.9919861025088563,
       6.8910746618455683, 6.0660203693313894, 5.9515028967870487, 5.7338063866902065, 6.4696656814387676,
       6.6437097124173796, 6.6558037126751071, 6.6968110375440686, 6.7355503149780418, 7.0692758786967751,
       6.7646519677235277, 7.5144953022589736, 7.8206312448555462, 7.5929191021643865, 7.2427526725437525,
       7.2841375547727694, 7.2550359020272825, 7.2164854032733379, 7.3982770194112790},
      {4.6613871447478390, 5.0472692391015705, 5.7778368737142678, 6.1145860501338767, 5.8407650527552457,
       5.9596289403979847, 5.5246139281101314, 5.7956005420136902, 5.3811835815003315, 5.1124645644859861,
       6.0592171281179752, 6.5231450144712877, 6.3481561890009841, 6.2211667082930848, 6.1561601615886978,
       6.0775475830032573, 6.2326939219649526, 5.9222120148157584, 6.7285582950845990, 7.1019679307289962,
       6.8846494285379851, 6.7714552089369420, 6.7446207025348928, 5.6996023030346592, 6.5785139364871892,
       6.4891301600415439, 6.4868621126065555, 6.3481561890009841, 6.3060152909603042, 6.2411979734817207,
       6.2682208080180217, 6.2396859418583945, 6.2583944046495077, 6.2472451988834719, 6.4921533221966436,
       6.2034034905394382, 7.0318594036603237, 7.3848602167560298, 7.0945983534754991, 6.9666636277301333,
       6.0565715233229307, 6.0106514473154879, 5.8772367333000064, 5.8129862024072816, 5.7759472847308864,
       5.7300272087234436, 5.6540602349330475, 6.4751461202546601, 6.5681202959870397, 6.5496010624217300,
       6.5828603515855866, 6.5953323597491451, 6.8886178357305523, 6.5958995968807796, 7.3910962208378095,
       7.8540788621536546, 7.4574255699397192, 7.5367941643368219, 7.4948420449761697, 7.4557247596363654,
       7.4239775020958341, 7.3916630074236682, 7.3667185405507762, 6.9489004099764857, 7.1159519705158809,
       7.0711654676801565, 7.2353826447444787, 7.2038241658839759, 7.1819033117119586, 7.1618724970690995,
       6.8744450172636400, 7.1167079863275431, 7.0450874276355453, 6.9838600588979203, 6.9266015478986374,
       6.8230445024403048, 6.0314382777700111, 5.8999136032836859, 5.6873190735511283, 6.3957775258575564,
       6.5788919443930203, 6.5919311896882133, 6.6308592458022142, 6.6692209658761321, 7.0037020948607527,
       6.6977558320357593, 7.4564807754480285, 7.7722534416415359, 7.5621161885697701, 7.2047689603756666,
       7.2455870560188247, 7.2170521898591966, 7.1788796990110830, 7.3567029079564579},
      {4.6326630493624084, 4.9771602614871222, 5.7657428734565404, 6.0903975990726469, 5.7882305142144164,
       5.9082279750288738, 5.4630085514666753, 5.7169879634282497, 5.3260034381644576, 5.0576624290559433,
       6.0410759024584975, 6.4466112540950320, 6.3290701688498157, 6.1777030078548831, 6.1075944807862106,
       6.0066830401229216, 6.1756241896456983, 5.8649535038164755, 6.6996454210191407, 7.0409297912171755,
       6.8200208897394292, 6.7034250495316794, 6.6758349778637420, 5.6725794684983573, 6.5031137492826518,
       6.4074939687552277, 6.3481561890009841, 6.3234009513538956, 6.2803152588215250, 6.2153087121171371,
       6.2266471471089773, 6.1975450438177146, 6.2126631073220926, 6.1790258102524049, 6.4409420365991128,
       6.1467117661260149, 6.9942536993980697, 7.3228768322067426, 7.0303473720369984, 6.8982554604190378,
       6.0182102537947895, 5.9707781461640215, 5.8284818232717157, 5.7598848772805926, 5.7209563706208160,
       5.7302159874034713, 5.6330841752527210, 6.4392412262957590, 6.5227670065654557, 6.5068933777951905,
       6.5375070621640035, 6.5293810185530665, 6.8347609453379761, 6.5365622676723127, 7.3459317100962531,
       7.7839703350849820, 7.3982770194112790, 7.4776456138083818, 7.4351267078618708, 7.3952534067104034,
       7.3633173704898445, 7.3306248679118475, 7.3054916223589270, 6.8838938632720987, 7.0518902183031837,
       7.0054029051641056, 7.1715101217575858, 7.1393848563112234, 7.1170859942333751, 7.0978107448564023,
       6.8067928657642085, 7.0265681940702356, 6.9523015800374175, 6.8899406381280741, 6.8321153405429325,
       6.7236450934003420, 5.9562273197912781, 5.8823387136642911, 5.6948783305762056, 6.3570382484235832,
       6.5307038209505892, 6.5477114734383486, 6.5862619721922941, 6.6045919765317995, 6.9515460147715302,
       6.6403080918106729, 7.4109382568006410, 7.7100817289579959, 7.5110936816522669, 7.1424080184663232,
       7.1824705488435940, 7.1528016589664718, 7.1136843736266675, 7.2901847801745205},
      {4.6133877999854356, 4.9144213116719468, 5.7712228617266570, 6.0446663017452318, 5.7617744662639740,
       5.8691106896890695, 5.4127420603608343, 5.6593514445231357, 5.2738473580752361, 5.0060731355525805,
       6.0376742818517899, 6.4103283522302998, 6.2958108796859591, 6.1576721932120231, 6.0728231599990279,
       5.9556600826596426, 6.1159088525313985, 5.8112853921039278, 6.6835825630230712, 7.0140957353609013,
       6.7918640314856331, 6.6731893730686984, 6.6450325148149023, 5.6619970493181802, 6.4675868632295828,
       6.3681874541896200, 6.3060152909603042, 6.2803152588215250, 6.2376075741949855, 6.1709002171872447,
       6.2122850994162615, 6.1722230195847665, 6.1807266205557587, 6.1314049239416084, 6.3774470709722744,
       6.1053264333512214, 6.9717660586401937, 7.2950979818587776, 7.0020017351031747, 6.8670749894643661,
       5.9932657869218975, 5.9452668927052690, 5.7957893206937179, 5.7249247778133823, 5.6856182632477745,
       5.6999803109404903, 5.6313833649493681, 6.3925651344766532, 6.5050037888118091, 6.4874293497381901,
       6.5080274015126847, 6.4845945157173421, 6.7735335766003519, 6.4938541324999974, 7.3200424487316704,
       7.7609159077412473, 7.3738997896700216, 7.4510003366321360, 7.4077258654197378, 7.3672857776824108,
       7.3349712830102449, 7.3020900017522195, 7.2767675269734964, 6.8489337638048884, 7.0157965456642550,
       6.9693092325251769, 7.1412744452946049, 7.1089599506224381, 7.0864723098645621, 7.0675750683934213,
       6.7744783710920427, 6.9832932723120615, 6.9054367095382840, 6.8402409336080927, 6.7807148257195973,
       6.6673309263469749, 5.8955671881852876, 5.8303714122550971, 5.6417770054495167, 6.3050704964686135,
       6.5091609746844030, 6.5222002199795970, 6.5569710902210039, 6.5611282760935978, 6.8920194563372590,
       6.5994899961675149, 7.3841042009443676, 7.6811693054383126, 7.4897396140661092, 7.1119835633233146,
       7.1516676352489776, 7.1218099666918286, 7.0815591081803051, 7.2569254910106640},
      {4.5954353530059846, 4.8620760023569218, 5.7477904264770903, 5.9789037392291817, 5.7062163150222691,
       5.8284818232717157, 5.3692783599226326, 5.1846519097638435, 5.2171551831160370, 4.9644990240977602,
       6.0114070125813752, 6.3477786316409288, 6.2445995940884274, 6.1130744690563263, 6.0361627007742404,
       5.9131416272589066, 6.0680991875405743, 5.7700892885549386, 6.6497564872733559, 6.9558924298699276,
       6.7340382833547157, 6.6142296012202859, 6.5864507508723218, 5.6421545628095728, 6.4056039292260705,
       6.3046924885627824, 6.2411979734817207, 6.2153087121171371, 6.1709002171872447, 6.1055156625770257,
       6.1616406004045894, 6.1348065445483160, 6.1455777424085207, 6.0903975990726469, 6.3298261846614787,
       6.0526031161303644, 6.9307587337712322, 7.2333038265352938, 6.9432311924805656, 6.8075488815758707,
       5.9713449327498802, 5.9218340069099265, 5.7710340830466293, 5.6984687298629400, 5.6576506342197819,
       5.6727682471783849, 5.6022817122038813, 6.3256685482431081, 6.4513352265534856, 6.4424536176766622,
       6.4707997051562618, 6.4422648389966346, 6.7228890775886798, 6.4396192342015901, 7.2703427442116890,
       7.6796572738149873, 7.3147512391415814, 7.3910962208378095, 7.3478212990796345, 7.3073812113423084,
       7.2754447245759737, 7.2425638938637240, 7.2170521898591966, 6.7937536204690145, 6.9568372243616183,
       6.9114834843942594, 7.0811811002744740, 7.0486778269222805, 7.0261901861644045, 7.0072929446932628,
       6.7179758759044228, 6.9022238676116041, 6.8251233206494897, 6.7612503471168210, 6.7032358203058751,
       6.5909859446507468, 5.8590955076405278, 5.7417432010753391, 5.6087064949656877, 6.2379855821008166,
       6.4553036337460519, 6.4819489109222985, 6.5210661962621028, 6.5201214017704121, 6.8423197518172776,
       6.5463886710408268, 7.3345932751044138, 7.6171071026798396, 7.4423075064353403, 7.0592602461024567,
       7.0957319266472174, 7.0651182422784053, 7.0246781545410784, 7.1947537783271240},
      {5.3860967831845885, 5.0897876945023057, 6.1353733311341747, 6.0321942935816733, 6.7393294929448038,
       6.1493573709210594, 5.7770813084483805, 5.4892758207370900, 5.4153876651558788, 5.1850299176696746,
       6.3394638093499642, 6.3324717894565223, 7.5281013341400262, 6.9676084222218240, 6.5879627824956470,
       6.3096061407928152, 6.1251689198598296, 5.9919429845243748, 6.8211549134569234, 6.8559257836983303,
       6.9655300545584140, 6.7213770459653537, 6.5830491302656142, 6.2111510756987673, 6.4239343841113525,
       6.3260465561489401, 6.2682208080180217, 6.2266471471089773, 6.2122850994162615, 6.1616406004045894,
       7.4793464241117356, 7.0176861346476365, 6.7217550538711848, 6.5048145595860056, 6.3668646517920982,
       6.2687880451496572, 7.1310700340202589, 7.1813365251260999, 7.2410518622403988, 7.1257788244301707,
       6.7189206703961135, 6.5785139364871892, 6.4042811268285487, 6.3296369554356744, 6.2914644645875608,
       6.2634968355595690, 6.2489455586410498, 6.4384852104840959, 7.7531674214903665, 7.3447976863787590,
       7.0819371160861371, 6.8857834522554802, 6.7599275447192992, 6.6682761713844414, 8.6978420680047233,
       7.7964423432485406, 7.6042575371562258, 7.3185308676541201, 7.2750671672159184, 7.2355718739702821,
       7.1947537783271240, 7.1605496946715768, 7.1350384412128252, 7.0859055232787034, 6.9528688171690520,
       6.9061922748041713, 7.0140957353609013, 6.9853720905212464, 6.9660968411442736, 6.9462548051814421,
       7.0853382861470680, 7.4963540765994958, 7.3037908120555732, 7.1637620860524809, 7.0465990087130956,
       6.9063815040299747, 6.6274580757412833, 6.4942321404058285, 6.3502350072101699, 6.4133519649311754,
       7.7915286910185078, 7.4183082845999149, 7.1586596551424195, 6.9798916517053540, 6.8666969815585350,
       6.7765571893012275, 7.5488881645945485, 7.5770454733941204, 7.9355258242141673, 7.5677858566114660,
       7.4836928392101338, 7.4247330673617213, 7.3791909992601097, 7.1766121021218705},
      {5.5229131178067776, 5.0716464688428280, 7.1558252716673474, 5.9964781783028007, 6.4008799567676169,
       6.3576050350094429, 5.9086059829347057, 5.5824392256952731, 5.3866640203162239, 5.2498476856940339,
       7.2707207521175201, 6.2995905081984978, 7.0598270326883164, 7.2312254588718847, 6.7525575169200254,
       6.4173203721237417, 6.1935766366251483, 6.0310602698641791, 7.5358489192993554, 7.3351600616902726,
       6.7671087938385446, 6.6828269977571848, 6.6253792575320984, 7.1992889721055500, 6.4108951388161595,
       6.3018581050877103, 6.2396859418583945, 6.1975450438177146, 6.1722230195847665, 6.1348065445483160,
       7.0176861346476365, 7.2979328158796255, 6.8967438793414884, 6.6183872376386557, 6.4348948111973607,
       6.3009133105960196, 7.6836256810075536, 7.5553133979021316, 7.1112275475116515, 7.0454654355413764,
       7.2019345769005945, 6.8504453448824378, 6.6302924592163555, 6.5276802082497127, 6.4721225075537836,
       6.4345168032915296, 6.4118399333078502, 6.3785810946897694, 7.2542798862156195, 7.5948086911477679,
       7.2261230279618243, 6.9691204538451492, 6.7952656520923407, 6.6669533689869187, 7.8495436683752295,
       7.9319358754732079, 7.8520000439444697, 7.6877833174259234, 7.6526344392786854, 7.6199419367006875,
       7.5929191021643865, 7.5636282201930962, 7.5413293581152478, 6.9494671965623454, 6.8442093408006590,
       7.4407959253577909, 7.4400399095461278, 7.4177410474682794, 7.4020566479238177, 7.3878833789111296,
       6.9717660586401937, 7.2510674948347162, 7.1828485567494251, 7.0405517833113436, 6.8993894841365329,
       6.7582267344159455, 6.6302924592163555, 6.5312706075364479, 6.4320604277222886, 6.3540146357227076,
       7.2835707681869097, 7.6715312302040504, 7.3098380374573244, 7.0660634873158710, 6.9010902944398858,
       6.7744783710920427, 7.9347707094940558, 7.8648505105596351, 8.1645607693028488, 7.4937084718044513,
       7.3457429314162255, 7.2954759897646095, 7.2610831274290337, 7.5657070384022811},
      {5.5066614811306813, 4.9644990240977602, 6.5707659007820842, 6.0044154432337091, 6.1529477702077937,
       6.0414539103643286, 5.8972675479428656, 5.5569279722365215, 5.3520819287548447, 5.2101636137683709,
       6.7000234289249718, 6.3332273547224096, 6.7722107742028292, 6.8364617556413299, 6.7682423670102629,
       6.4065487237177612, 6.1612625924987583, 5.9813605653441977, 6.9989781224022991, 6.8564925702841899,
       6.7761791813953955, 6.6764017644496017, 6.5921199683682419, 6.7631403866459783, 6.4428316255824933,
       6.3254797695630804, 6.2583944046495077, 6.2126631073220926, 6.1807266205557587, 6.1455777424085207,
       6.7217550538711848, 6.8967438793414884, 6.9211211090827458, 6.6157416328436121, 6.4080607553410873,
       6.2534812029652507, 7.1669744774333841, 7.0947871321555267, 7.0660634873158710, 6.9723328452260525,
       7.1229439904093228, 7.0207097473485121, 6.7831712012888374, 6.6631737404743800, 6.5955215889749486,
       6.5490342758358704, 6.5169090103895080, 6.3581718215953025, 6.9437979790664253, 7.1735884894209949,
       7.2463430718304869, 6.9602383944225501, 6.7591719794534120, 6.6081828263643105, 7.3687973587599611,
       7.4704648152349122, 7.3903402050261464, 7.2472878663221776, 7.2083598102081767, 7.1722656870234722,
       7.1424080184663232, 7.1121727925491180, 7.0876063335820563, 6.8717994124685955, 6.7629511574201748,
       6.9124287294317259, 6.9679864301276551, 6.9417191608572404, 6.9233891565177341, 6.9073262985216655,
       6.9288691447878508, 7.0534022499265090, 6.9719548373202214, 6.8929642508289497, 6.8177532928502167,
       6.6686537287444967, 6.6278356331013386, 6.5456326552291637, 6.4709884838362894, 6.3192437654813016,
       6.9572152322674503, 7.2389730440312139, 7.3391284688828389, 7.0653074715042088, 6.8716106337885678,
       6.7217550538711848, 7.4657403922306829, 7.4379615418827187, 7.7140501361505622, 7.3918522366494717,
       7.2875391753794769, 7.2387842653511862, 7.1989109641997189, 7.1061255671473660},
      {5.1589518776250634, 4.8887208289873918, 6.1361293469458378, 5.9501800943895260, 5.9841949488192698,
       5.8418986259269641, 5.6893978917603132, 5.6336509618385806, 5.3479242923364749, 5.1559282649241878,
       6.2850396818257543, 6.2704888554530109, 6.5586719005243568, 6.5664199362294617, 6.4777917250497037,
       6.4594612701644225, 6.1504909440927777, 5.9248576196108020, 6.6359616767122747, 7.4515675737637714,
       7.3374281091252618, 7.2382174787653266, 7.1582820977823642, 6.3797146678614878, 7.0331822060578464,
       6.3131965400795504, 6.2472451988834719, 6.1790258102524049, 6.1314049239416084, 6.0903975990726469,
       6.5048145595860056, 6.6183872376386557, 6.6157416328436121, 6.6646853215519304, 6.3965335416692195,
       6.1973562651376870, 6.8328709058088188, 7.6290123242575394, 7.5384949746401757, 7.4294574903659507,
       6.7497231334449532, 6.6520245347083442, 6.5730339482170725, 6.5031137492826518, 6.4422648389966346,
       6.2400639497642265, 6.1760021975515293, 6.2942992986084088, 6.7202434727936353, 6.8808702505712231,
       6.9214991169885769, 7.0001116955740175, 6.7421643269656517, 6.5479002521183762, 7.0711654676801565,
       7.8979205704976883, 7.8115599562071436, 7.7580810737203993, 7.7253885711424015, 7.6959093610368594,
       7.6719096886556581, 7.6479100162744560, 7.6278787510858201, 7.4082926520055965, 7.3251448796417309,
       6.7784467782846090, 7.5298021444433800, 7.5084480768572224, 7.4910624164636310, 7.4780236217142138,
       7.4020566479238177, 7.4001666083946605, 7.3156960336332721, 6.7518019516541381, 6.6811261874538310,
       6.6302924592163555, 6.5212554254879063, 6.2759688437231267, 6.1667425807688740, 6.2610400094445522,
       6.7278022792729368, 6.9356723860012650, 7.0023792924632300, 7.1057475592415349, 6.8555477757924992,
       6.6626069538885213, 7.1733997107409664, 7.8926293609076001, 8.0402173439357707, 7.7497658008836590,
       7.7087589265604732, 7.6724764752415169, 7.6367608105084193, 7.6133279247130776},
      {4.9182009401844855, 4.7914002381566139, 5.8825279428900945, 6.2102062812070766, 5.8636307014189537,
       5.6895866704403408, 5.5231018964868062, 5.4223796850493207, 5.3118306196975462, 5.0888429000106150,
       6.0391863134751160, 6.5516798806309149, 6.4284695778897785, 6.3882187193782549, 6.2772916461206494,
       6.2011358936502257, 6.1125076824704676, 5.8570166894313429, 6.4131627357053720, 7.0422525936146974,
       6.9347271409637985, 6.8340049295263139, 6.7519907303341657, 6.1457665210885484, 6.6236784472287447,
       6.5547034933317905, 6.4921533221966436, 6.4409420365991128, 6.3774470709722744, 6.3298261846614787,
       6.3668646517920982, 6.4348948111973607, 6.4080607553410873, 6.3965335416692195, 6.3611954342961781,
       6.1325384971133268, 6.6172532139211615, 7.2234774231667798, 7.1492112596797375, 7.0392289809138218,
       6.5191766072787214, 6.4186431745212644, 6.3275585877722653, 6.2521584005677289, 6.1877186404492006,
       6.1289480978265924, 5.9959113917169411, 6.4993345713158890, 6.5773803633154699, 6.6905745829165140,
       6.7045586227033978, 6.7192986783019446, 6.7058814251009196, 6.4825156975081573, 6.8702878313910460,
       7.4946532662961420, 7.4245442886816937, 7.3563253505964026, 7.3238220772442091, 7.2939644086870592,
       7.2697755070800536, 7.2457762852446281, 7.2253670121501612, 7.0216545418402028, 6.9111059270342041,
       6.8712326258827368, 7.1259676031101984, 7.1048027647498442, 7.0870390964504217, 7.0743778590610598,
       7.0203317394426810, 7.0214657631601751, 6.9351051488696296, 6.8604609774767562, 6.7960212173582280,
       6.7261014689695831, 6.2587724125553397, 6.1911202610559073, 6.0112182339013476, 6.4316824198164575,
       6.5758683316921447, 6.7370618960555912, 6.7784467782846090, 6.8177532928502167, 6.8238005182519679,
       6.6019463717367559, 6.9778132840419449, 7.5031568672671343, 7.6834364517817493, 7.3892066318544281,
       7.3387509115227836, 7.2990663890513439, 7.2633507243182471, 7.2206430396917076},
      {4.7407552885991660, 4.6891659950958040, 5.7143423586332061, 5.9171100344514729, 5.7702780672349663,
       5.5665655969250079, 5.3849632100128701, 5.2496589070140063, 5.2587292945708572, 5.0138207207119097,
       5.8700559347265360, 6.2631188276537371, 6.3398413667100204, 6.2587724125553397, 6.1266805009373790,
       6.0068718188029502, 5.8914095517669178, 5.7850181228335131, 6.2515911634360934, 6.7567151533383951,
       6.6559929419009105, 6.5531914617084652, 6.4687208869470769, 5.9874077907459498, 6.3358729595174541,
       6.2667092269404723, 6.2034034905394382, 6.1467117661260149, 6.1053264333512214, 6.0526031161303644,
       6.2687880451496572, 6.3009133105960196, 6.2534812029652507, 6.1973562651376870, 6.1325384971133268,
       6.0652643535197264, 6.4564376574635460, 6.9386955481563648, 6.8797362268537281, 6.7688091535961226,
       6.3611954342961781, 6.2563155864403228, 6.1527585409819903, 6.0724451520931968, 6.0047930005937644,
       5.9437548610819437, 5.8898979706893684, 6.2347727401741375, 6.4764689226521819, 6.5537586988400998,
       6.5458218844549672, 6.5131293818769702, 6.4681541003612173, 6.4180763879354048, 6.7175978679985908,
       7.2055249761873288, 7.1569592953848424, 7.0760786693644135, 7.0431973881063881, 7.0133397195492382,
       6.9887732605821773, 6.9643960308409198, 6.9437979790664253, 6.7550143430350422, 6.6181980084128522,
       6.6214108503395321, 6.8426977597231087, 6.8217217000427830, 6.8037692530633329, 6.7903524504080837,
       6.7576599478300858, 6.7489671176332902, 6.6601501277735045, 6.5826715729055589, 6.5157749866720138,
       6.4437764200741841, 6.0626187487246828, 5.9961006209427445, 5.9244796117049709, 6.1896082294325820,
       6.4643744718486795, 6.5921199683682419, 6.6140408225402583, 6.6064820160609568, 6.5809707626022051,
       6.5437430662457823, 6.8313593247312694, 7.2304694430602217, 7.4377723126569153, 7.1431640342779863,
       7.0855275153728714, 7.0428198307463328, 7.0065369288815997, 6.9460655759556378},
      {6.2158754987029967, 5.4785046228768852, 7.6322247156384426, 6.8219104787228106, 7.4432527514728068,
       6.5148301921803231, 5.9696441224465264, 5.6336509618385806, 5.4019704119548537, 5.2313284521287242,
       7.8287563873749306, 7.1641400939583120, 7.2234774231667798, 7.6292015534833428, 7.0031353082748931,
       6.6017575930567274, 6.3360621887432576, 6.1400977541384041, 8.3813129354537761, 7.7149953811880287,
       7.5084480768572224, 7.4179302766940829, 7.3030347962439102, 7.4294574903659507, 7.1501560541714282,
       7.0798582978769522, 7.0318594036603237, 6.9942536993980697, 6.9717660586401937, 6.9307587337712322,
       7.1310700340202589, 7.6836256810075536, 7.1669744774333841, 6.8328709058088188, 6.6172532139211615,
       6.4564376574635460, 8.7259984757127427, 8.1432067017166911, 7.8803465819698451, 7.8342368261908231,
       7.8627716923504503, 7.7720651135072831, 7.6386503994918007, 7.5910295131810051, 7.5736438527874137,
       7.5572029868855131, 7.5498329590862392, 7.3260896741334216, 7.5932966595244418, 8.2274884977980527,
       7.7527898641303103, 7.4082926520055965, 7.2011785610889314, 7.0352605737212555, 9.1373918235252258,
       9.1009201429804651, 8.2516773994050574, 8.2042448412285136, 8.1500099429301063, 8.1061682345860717,
       8.0649721310370825, 8.0245320432997573, 7.9961864063659327, 7.4632835661156669, 7.5092040926688854,
       8.1539783501226726, 7.9013221911043958, 7.8769449613631384, 7.8593700717437436, 7.8438749014250853,
       7.6095487467463148, 8.6995424277623012, 8.5323016379971026, 8.3843360976088768, 8.2367481145807062,
       8.0857594120373815, 8.0523108936477215, 7.8673068861288762, 7.6641612024047774, 7.3134284367440596,
       7.6613272694754810, 8.3491872194616388, 7.8561576803628395, 7.5211093142465844, 7.3311916544977063,
       7.1571480740648701, 9.1717846858608016, 8.5695293343535255, 8.3057230684776613, 8.1190187012012380,
       7.9771003862147634, 7.9485655200551362, 7.9060466141086252, 8.0589249056353314},
      {5.4397648948971362, 5.9824945890616918, 6.7695651694077847, 7.1221879745976597, 6.7580379557359178,
       6.7631403866459783, 6.3404086038416549, 6.4258239730947340, 5.9945885893194193, 5.7092399277231447,
       7.0375286211562438, 7.4840708471159649, 7.2754447245759737, 7.5294241365375481, 6.9986001144964671,
       7.4003558376204639, 6.9449320027839194, 6.6282136410071706, 7.7764110780599056, 8.1025782858451141,
       7.8572912535345587, 7.7945527542651591, 7.6754996373966167, 6.5902303793848604, 7.5301801523492111,
       7.4228434783783399, 7.3848602167560298, 7.3228768322067426, 7.2950979818587776, 7.2333038265352938,
       7.1813365251260999, 7.5553133979021316, 7.0947871321555267, 7.6290123242575394, 7.2234774231667798,
       6.9386955481563648, 8.1432067017166911, 8.4591690476817778, 8.1288451045697521, 7.9897616236041253,
       7.0707874597743245, 6.9728996318119121, 6.7967772331698901, 6.7017242392283256, 6.6847165867405653,
       6.6172532139211615, 6.5754903237863136, 7.5383057454143723, 7.5572029868855131, 7.9625495598420208,
       7.5084480768572224, 8.1105151002302467, 7.7225537371215545, 7.4458983562678513, 8.5818121132912797,
       9.1220849813408194, 8.5188843847960776, 8.5976857420615449, 8.5494980691648905, 8.5077351790300408,
       8.4716410558453372, 8.4347918179405212, 8.4073905249526124, 7.9512111248501807, 8.1740096153113075,
       8.1216643059962816, 8.2675510281753226, 8.2339141816514108, 8.2104812958560682, 8.1895056867215175,
       7.8605036449154619, 8.4319569839196724, 8.4124934064084478, 8.2998650723017136, 8.1995213193158349,
       8.0547681703085132, 7.2030686006180886, 6.9589155920250283, 6.6835825630230712, 7.4687640049315585,
       7.5851706159135057, 8.0343584466682714, 7.5598490422263334, 8.2237093198312898, 7.8325364664332451,
       7.5430301684186007, 8.6250870350494537, 8.8864360242645262, 8.5370260610013311, 8.1881828843239948,
       8.2588581979785278, 8.2327797073881399, 8.1949852244458583, 8.4202409915677769},
      {5.4796381960486036, 5.8492686537262379, 6.5240898089629784, 6.8232332811203333, 6.8147296801493411,
       6.7620063629284841, 6.3235897300339232, 6.3976671148409379, 5.9991237830978452, 5.7411764144894795,
       6.8066040870841800, 7.1692425248683724, 7.3126724209323966, 7.5175189149598491, 6.9687424459393181,
       7.3200424487316704, 6.8882402783704961, 6.5932539920857360, 7.5320697413325925, 7.7817022876499937,
       7.6401624311151268, 7.4763228114108600, 7.4043242448130302, 6.4254459651889029, 7.2395402811628484,
       7.1272904055077202, 7.0945983534754991, 7.0303473720369984, 7.0020017351031747, 6.9432311924805656,
       7.2410518622403988, 7.1112275475116515, 7.0660634873158710, 7.5384949746401757, 7.1492112596797375,
       6.8797362268537281, 7.8803465819698451, 8.1288451045697521, 7.9738879948338601, 7.7724426708673393,
       6.9043026858207899, 6.8185093086618798, 6.6353948901264159, 6.5359950305406773, 6.5265466350779944,
       6.4517132344593175, 6.4063599450377335, 7.2395402811628484, 7.5602265995863887, 7.4419299490752850,
       7.4251110752675533, 7.9738879948338601, 7.5908402839552007, 7.3232548401125737, 8.2941963053515693,
       8.7482968872448161, 8.3586360654700975, 8.2832354277197844, 8.2350477548231282, 8.1872376392865291,
       8.1507659587417685, 8.1142942781970095, 8.0865154278490436, 7.8542680913794589, 7.9043453532594956,
       7.8085367940520438, 7.9451638994484295, 7.9107710371128537, 7.8875273805433155, 7.8657948545055500,
       7.7680967063147168, 8.1432067017166911, 8.0385160830866411, 7.9759659119514934, 7.8678736727147358,
       7.7238765395190763, 6.9897185056196438, 6.7693763907277571, 6.5244678168688095, 7.1843601378269755,
       7.5900842681435385, 7.4997552466604267, 7.4772676059025507, 8.0884050168324251, 7.6979872781544927,
       7.4169854822023922, 8.3401168319047869, 8.5544117213949225, 8.2936295187657088, 8.0543897118569063,
       8.1825132162823007, 8.1299786777414713, 8.0910506216274705, 8.0997434518242653},
      {5.3923327872663673, 5.7007363267521534, 6.4209107714104769, 6.6826377685313805, 6.6883069860273006,
       6.6709217761794850, 6.2285367360923587, 6.3026141208993733, 5.9123856114472444, 5.6629418438098709,
       6.9275463423903281, 7.0371506132504127, 7.1941865411954895, 7.0231665734635280, 6.8725554282802586,
       7.2115722015890809, 6.7812816123054560, 6.4885629229099084, 7.5265893025167010, 7.6588699928146884,
       7.4802912186034263, 7.3746553549359088, 7.2956647684446372, 6.3632742525053629, 7.1163299784217120,
       6.9986001144964671, 6.9666636277301333, 6.8982554604190378, 6.8670749894643661, 6.8075488815758707,
       7.1257788244301707, 7.0454654355413764, 6.9723328452260525, 7.4294574903659507, 7.0392289809138218,
       6.7688091535961226, 7.8342368261908231, 7.9897616236041253, 7.7724426708673393, 7.6717204594298538,
       6.8321153405429325, 6.7281802871787679, 6.5339166628772682, 6.4262019810005651, 6.4167535855378830,
       6.3343613784399038, 6.2842836660140913, 7.0832599184836589, 7.4232214862841719, 7.3533017378955270,
       7.3070032034364765, 7.8502996841868917, 7.4598819455089593, 7.1875725292078787, 8.2044340704543171,
       8.6625039606316818, 8.1551119232943918, 8.1488763697583870, 8.1012554834475914, 8.0587365775010795,
       8.0207528653329945, 7.9844704140140372, 7.9566906625745215, 7.6549015856221221, 7.7777338804574274,
       7.8041899284078697, 7.8121267427930023, 7.7767895365115125, 7.7533566507161700, 7.7320025831300123,
       7.5579590026971761, 7.9599039550469763, 7.8391504784208559, 7.7730094574531980, 7.6601927952122111,
       7.5154400967506643, 6.7361171015639005, 6.5894743635731974, 6.3681874541896200, 7.0201429607626524,
       7.4477879452512328, 7.4050802606246933, 7.3578369316739529, 7.2905627880803525, 7.5674073981598582,
       7.2811139420718938, 8.2558350358234271, 8.5181283689844154, 8.2405281936390207, 7.9848479713740925,
       7.9597147258211729, 7.9234322745022157, 7.8775117479489971, 7.9599039550469763},
      {5.8401982661693870, 5.2908545600172197, 6.8876730412388616, 5.8088285659889110, 6.2353395267599971,
       6.6635517483802120, 6.1527585409819903, 5.7670656758540630, 5.5546603753473089, 5.4155764438359064,
       7.0891179146596066, 6.1028700577819812, 6.7355503149780418, 7.1446760659013115, 6.9789468572136633,
       6.5616955132252324, 6.3035589153910641, 6.1281920820149294, 7.6118158930897515, 6.7253459037036958,
       6.5452546473233326, 6.6129067988227641, 6.4834609425456238, 6.8364617556413299, 6.2258911312973142,
       6.1181768999663877, 6.0565715233229307, 6.0182102537947895, 5.9932657869218975, 5.9713449327498802,
       6.7189206703961135, 7.2019345769005945, 7.1229439904093228, 6.7497231334449532, 6.5191766072787214,
       6.3611954342961781, 7.8627716923504503, 7.0707874597743245, 6.9043026858207899, 6.8321153405429325,
       7.2935864007812281, 7.1495892675855686, 6.9152631129067981, 6.8081156681617303, 6.7552035722608457,
       6.7242118799862016, 6.7141962473918841, 6.3148973503829033, 7.0369613840246092, 7.4986216734887083,
       7.4668739654024012, 7.1002671204256433, 6.8706653887511013, 6.7083382512159364, 8.1717415678763192,
       7.7044129620078516, 7.2680751473224765, 7.1905961419087543, 7.1424080184663232, 7.0998895630655872,
       7.0619058508975012, 7.0239225892751911, 6.9925528890947160, 6.6762129857695740, 6.6287808781388051,
       7.2119502094949119, 6.8897518594480465, 6.8619730091000815, 6.8453433645181532, 6.8268241309528435,
       6.7225110696828478, 7.2712875387033797, 7.3020900017522195, 7.1936197546096299, 7.0284577830536170,
       6.8799254560795324, 6.9778132840419449, 6.8644293846693225, 6.7476443152357684, 6.3632742525053629,
       7.1238887849010135, 7.4615827558123131, 7.5570142082054854, 7.1915409364004450, 6.9623172126317350,
       6.7901632211822802, 8.2006548924875542, 7.4761335821850565, 7.4801024399233986, 7.2178082056708597,
       7.1134955949466399, 7.0677638470734490, 7.0473550245247578, 7.0086157470907855},
      {5.7260588015308773, 5.1612194745142768, 6.7691871615019537, 5.7585620748830708, 6.0730123892248313,
       6.5533806909342687, 6.0423987048560193, 5.6633198517157020, 5.4512925591147798, 5.3120198489233497,
       6.9789468572136633, 6.0614851755529635, 6.6032691741342777, 6.7946984149607053, 6.8746342464894434,
       6.4609728512419720, 6.1998130912527030, 6.0206670799098063, 7.5305577097092673, 6.6338828585030898,
       6.4874293497381901, 6.3867066877549297, 6.4543588392543612, 6.8092496918792245, 6.1863958380516788,
       6.0745239703023817, 6.0106514473154879, 5.9707781461640215, 5.9452668927052690, 5.9218340069099265,
       6.5785139364871892, 6.8504453448824378, 7.0207097473485121, 6.6520245347083442, 6.4186431745212644,
       6.2563155864403228, 7.7720651135072831, 6.9728996318119121, 6.8185093086618798, 6.7281802871787679,
       7.1495892675855686, 7.0945983534754991, 6.8621617877801091, 6.7519907303341657, 6.6952990059207425,
       6.6614729301710272, 6.6486229141016366, 6.2396859418583945, 6.8971218872473194, 7.1637620860524809,
       7.3621833467723503, 6.9993561303081302, 6.7652187543093874, 6.5976004071841334, 8.0791454000497716,
       7.5755334417707951, 7.1822813196177906, 7.1021571599547997, 7.0520794475289872, 7.0076709525990948,
       6.9685532167135147, 6.9305699550912045, 6.8993894841365329, 6.5724667110854371, 6.5375070621640035,
       7.1204871642943059, 6.7895964345964206, 6.7599275447192992, 6.7419750977398483, 6.7240226507603982,
       6.6335048505972587, 7.1000783417456148, 7.0380954077421034, 7.0715434755859876, 6.9029798834232672,
       6.7544475564491826, 6.8587601671734024, 6.7538803193175472, 6.6488116927816652, 6.2737012468339142,
       6.9759232445127877, 7.2508782656089128, 7.4517563524437991, 7.0915747407746235, 6.8580046019075152,
       6.6811261874538310, 8.1029558432051694, 7.3822146119609862, 7.4228434783783399, 7.0978107448564023,
       7.0165525614759181, 6.9708212641485030, 6.9369951883987868, 6.9407743663655497},
      {5.5159210979133357, 5.0601192551709602, 6.5998680040733460, 5.6026592695639366, 5.8989683582462193,
       6.2247571075798200, 5.8534262901446077, 5.6232573213384311, 5.3843959728812347, 5.2228244006119571,
       6.8190760952477385, 5.8985908008861632, 6.4262019810005651, 6.5749235372004540, 6.6418196728882233,
       6.3884074980582835, 6.1143972714538490, 5.9214564495498712, 7.3971434462395607, 6.4596500488444502,
       6.3058265122802766, 6.1988678462152365, 6.1754354109656697, 6.5896631422532250, 5.9930770082418690,
       5.9566053276971092, 5.8772367333000064, 5.8284818232717157, 5.7957893206937179, 5.7710340830466293,
       6.4042811268285487, 6.6302924592163555, 6.7831712012888374, 6.5730339482170725, 6.3275585877722653,
       6.1527585409819903, 7.6386503994918007, 6.7967772331698901, 6.6353948901264159, 6.5339166628772682,
       6.9152631129067981, 6.8621617877801091, 6.7444319238548651, 6.6036471820401088, 6.5237118010571464,
       6.4709884838362894, 6.4433984121683530, 6.0943660062652132, 6.7283690658587956, 6.9492784178823177,
       7.1238887849010135, 6.9061922748041713, 6.6631737404743800, 6.4853505315290052, 7.9468642592060075,
       7.3893954105344557, 6.9902852922055034, 6.9205543224968862, 6.8717994124685955, 6.8287137199362249,
       6.7907300077681390, 6.7536915406375195, 6.7232670854945109, 6.3742346795913711, 6.3494789913985068,
       7.0210877552543431, 6.6208436132078967, 6.5915531817823823, 6.5734115055771278, 6.5543254854259594,
       6.4558708708776873, 6.9158298994926568, 6.8489337638048884, 6.8294697357478880, 6.7107946267851766,
       6.6476781196099459, 6.6864173970439191, 6.5635851022086138, 6.4411308152791404, 6.1383969438350503,
       6.8111392808626059, 7.0401742259512883, 7.2134617905724623, 6.9955765017955915, 6.7540695485433515,
       6.5681202959870397, 7.9682183267921642, 7.2034461579781439, 7.2546578941214515, 6.8999562707223916,
       6.8302253010137752, 6.7841159957805282, 6.7463215128382465, 6.7652187543093874},
      {5.4004588308773043, 4.9764042456754591, 6.5522466672167745, 5.5265035170935128, 5.8075057635913891,
       6.1293261057324235, 5.7330503708785434, 5.5675103914166986, 5.3182558530051285, 5.1476134426332241,
       6.7712659797111385, 5.8182774119973697, 6.3526918333251858, 6.4715552704221491, 6.5204994096762432,
       6.3198105520671604, 6.0403198866468344, 5.8417098472469364, 7.3550025481988808, 6.3708330589846645,
       6.2120963207362339, 6.1002244529869367, 6.0783035988149203, 6.4980117689183672, 5.8891419548777053,
       5.8997243740578824, 5.8129862024072816, 5.7598848772805926, 5.7249247778133823, 5.6984687298629400,
       6.3296369554356744, 6.5276802082497127, 6.6631737404743800, 6.5031137492826518, 6.2521584005677289,
       6.0724451520931968, 7.5910295131810051, 6.7017242392283256, 6.5359950305406773, 6.4262019810005651,
       6.8081156681617303, 6.7519907303341657, 6.6036471820401088, 6.5427982717540916, 6.4553036337460519,
       6.3969110990292748, 6.3651638414887444, 6.0153754197739415, 6.6539141236917256, 6.8525241630916236,
       7.0088045257708131, 6.8315481034112970, 6.5834271381714462, 6.4010687354476454, 7.8971645546860261,
       7.2822479657893879, 6.8874842625588331, 6.8234225103461368, 6.7754236161295083, 6.7327154809571939,
       6.6949209980149114, 6.6584493174701516, 6.6282136410071706, 6.2599059857270580, 6.2527251871535885,
       6.9651520466525829, 6.5341054415572959, 6.5061373619835274, 6.4844057370373145, 6.4651304876603417,
       6.3598726318986554, 6.7971552410757221, 6.7253459037036958, 6.7068262195926103, 6.5832383594914177,
       6.5673647307211525, 6.5970331700524989, 6.4717444996479525, 6.3485341969068161, 6.0601619226096659,
       6.7347942991663787, 6.9436092003863976, 7.0989443180281206, 6.9192315200993644, 6.6730001438428941,
       6.4828937054139892, 7.9156837882513358, 7.1063143458273945, 7.1754785289501521, 6.7918640314856331,
       6.7272354926870772, 6.6801813929621403, 6.6429536966057174, 6.6661973531752556},
      {5.3365858573446348, 4.9057284814751521, 6.5450658686433041, 5.4862522080362144, 5.7506252604979382,
       6.0688547528064616, 5.6487690253429585, 5.5081730622082308, 5.2536273142065726, 5.0782604808304379,
       6.7606835605309614, 5.7765140713167451, 6.3179209630837789, 6.4154307831403603, 6.4500124241559638,
       6.2580163967436766, 5.9745577746765592, 5.7727348933499822, 7.3432861053012095, 6.3719670827021586,
       6.2204106924814218, 6.1123184532446642, 6.0989016505894149, 6.4573824519552367, 5.9135196351647386,
       5.7763252926367175, 5.7759472847308864, 5.7209563706208160, 5.6856182632477745, 5.6576506342197819,
       6.2914644645875608, 6.4721225075537836, 6.5955215889749486, 6.4422648389966346, 6.1877186404492006,
       6.0047930005937644, 7.5736438527874137, 6.6847165867405653, 6.5265466350779944, 6.4167535855378830,
       6.7552035722608457, 6.6952990059207425, 6.5237118010571464, 6.4553036337460519, 6.4203439848246173,
       6.3611954342961781, 6.3290701688498157, 5.9673765255573139, 6.6157416328436121, 6.8028244585716422,
       6.9475776075789639, 6.6694097445561598, 6.5197433938645801, 6.3343613784399038, 7.8773225187231937,
       7.2421858859578929, 6.8612169932884184, 6.7975332489815532, 6.7512347145225036, 6.7098498322934859,
       6.6733781517487261, 6.6378512656956561, 6.6085608342701416, 6.2069938898261734, 6.1557821536828659,
       6.9369951883987868, 6.5248458247746415, 6.4997125792217210, 6.4842165078115110, 6.4641856931686510,
       6.3640302683170260, 6.7226998483628755, 6.6465440958924518, 6.6285916489130017, 6.5006573737134117,
       6.4120291625336536, 6.5431762796599227, 6.4192099611071232, 6.2990237216126381, 6.0080058425204443,
       6.6926534011256988, 6.8929642508289497, 7.0384734156479345, 6.7591719794534120, 6.6091276208560013,
       6.4161863484062476, 7.8930069182676554, 7.0764566772702446, 7.1724549162492766, 6.7671087938385446,
       6.7036138282117070, 6.6567485071667978, 6.6214108503395321, 6.6452212934949300},
      {5.2919885837347147, 4.8416667292624549, 6.5416642480365974, 5.5093070859257249, 5.7075391174197918,
       6.0235014633848785, 5.5837620280927958, 5.2984138170422970, 5.1914551509772577, 5.0134431633518544,
       6.7531247540516608, 5.7795376840176207, 6.2935432827967457, 6.3774470709722744, 6.4006907275418135,
       6.0524143374503367, 5.9140864217505973, 5.7096179356289767, 7.3313808837235097, 6.2997792868785254,
       6.1400977541384041, 6.0250130444624279, 6.0097066528237972, 6.4345168032915296, 5.8120409573698151,
       5.7595068693747615, 5.7300272087234436, 5.7302159874034713, 5.6999803109404903, 5.6727682471783849,
       6.2634968355595690, 6.4345168032915296, 6.5490342758358704, 6.2400639497642265, 6.1289480978265924,
       5.9437548610819437, 7.5572029868855131, 6.6172532139211615, 6.4517132344593175, 6.3343613784399038,
       6.7242118799862016, 6.6614729301710272, 6.4709884838362894, 6.3969110990292748, 6.3611954342961781,
       6.3353061729315945, 6.3041257019769228, 5.9647309207622694, 6.5873955453640125, 6.7695651694077847,
       6.9065702827100024, 6.5855059563806311, 6.4636184560370165, 6.2755908358172956, 7.8584257277978287,
       7.1800137227285772, 6.7952656520923407, 6.7327154809571939, 6.6854726025522284, 6.6433317045115485,
       6.6061040081551257, 6.5701991141962246, 6.5401526669590471, 6.1276252954290698, 6.0781143695891169,
       6.9288691447878508, 6.4526580289510083, 6.4267687675864247, 6.4105171309103275, 6.3927539131566808,
       6.2820160691248788, 6.6869841836297788, 6.6089383916301978, 6.5926867549541006, 6.4611620804677754,
       6.3672426596979292, 6.4690988948529080, 6.3804706836731508, 6.2631188276537371, 5.9635973475905510,
       6.6607169143593641, 6.8587601671734024, 6.9982221065906360, 6.6764017644496017, 6.5142634055944644,
       6.3581718215953025, 7.8725980957189643, 7.0144737432667332, 7.1193535911225876, 6.6934094169373610,
       6.6316152616138773, 6.5834271381714462, 6.5543254854259594, 6.5730339482170725},
      {5.2621309151775648, 4.7855417914348903, 6.5482782600242082, 5.4346629145328516, 5.6788154725801361,
       5.9889193718234992, 5.5317947266836009, 5.2184779855135588, 5.1332513949405083, 4.9533498183317235,
       6.7546363351292102, 5.7211451493008445, 6.2831496422965971, 6.3545814223085673, 6.3657306280746031,
       5.9845729567251018, 5.7795376840176207, 5.6531149898955810, 7.3270344686251123, 6.2548040053627734,
       6.0907756069784789, 5.9715341619756837, 5.9564160984713057, 6.4294143723814692, 5.7496800154604717,
       5.6878863106827628, 5.6540602349330475, 5.6330841752527210, 5.6313833649493681, 5.6022817122038813,
       6.2489455586410498, 6.4118399333078502, 6.5169090103895080, 6.1760021975515293, 5.9959113917169411,
       5.8898979706893684, 7.5498329590862392, 6.5754903237863136, 6.4063599450377335, 6.2842836660140913,
       6.7141962473918841, 6.6486229141016366, 6.4433984121683530, 6.3651638414887444, 6.3290701688498157,
       6.3041257019769228, 6.2942992986084088, 5.9004803898695446, 6.5734115055771278, 6.7514239437483070,
       6.8803030134395877, 6.5293810185530665, 6.3405973825216835, 6.2262691392031453, 7.8491661110151734,
       7.1480772359622433, 6.7555811296209010, 6.6911418200481485, 6.6433317045115485, 6.6006235693392332,
       6.5630183156227551, 6.5265466350779944, 6.4963109586150134, 6.0726343813190002, 6.0199110640981432,
       6.9007122865340547, 6.4069267316235932, 6.3808486915789819, 6.3645970549028847, 6.3468333866034623,
       6.2313711195674308, 6.6429536966057174, 6.5601834816019071, 6.5427982717540916, 6.4044703560543521,
       6.3001572947843565, 6.4263912102263694, 6.3353061729315945, 6.2438435782767643, 5.9327944339959355,
       6.6429536966057174, 6.8394849177964296, 6.9730888610377155, 6.6212216211137278, 6.4314931905906541,
       6.3097949194728429, 7.8627716923504503, 6.9761124737385911, 7.0904407170571284, 6.6476781196099459,
       6.5862619721922941, 6.5373178329381991, 6.5076493936068536, 6.5269246429838264},
      {4.8044390329060320, 5.0268599660071036, 6.1520029757161030, 6.1996238620268995, 5.9307160663325256,
       5.9974234233402672, 5.5552271619331686, 5.3723019726235082, 5.3842071942012071, 5.1447790591581519,
       6.4014467433534765, 6.5970331700524989, 6.5233337931513153, 6.3509910230218320, 6.2428987837850736,
       6.1123184532446642, 6.2359067638916317, 5.9501800943895260, 7.0692758786967751, 7.2550359020272825,
       7.0199537315368490, 6.8859722309355078, 6.8374065501330206, 6.0004465854953670, 6.6537248944659222,
       6.5458218844549672, 6.4751461202546601, 6.4392412262957590, 6.3925651344766532, 6.3256685482431081,
       6.4384852104840959, 6.3785810946897694, 6.3581718215953025, 6.2942992986084088, 6.4993345713158890,
       6.2347727401741375, 7.3260896741334216, 7.5383057454143723, 7.2395402811628484, 7.0832599184836589,
       6.3148973503829033, 6.2396859418583945, 6.0943660062652132, 6.0153754197739415, 5.9673765255573139,
       5.9647309207622694, 5.9004803898695446, 6.5422310346224561, 6.7313926785596712, 6.6932201877115576,
       6.6892517805189913, 6.6503237244049904, 6.8950430690381346, 6.6238672259087723, 7.6683188388231471,
       7.9924072283991698, 7.6267451779141018, 7.7063025509912331, 7.6584924354546331, 7.6150282844706547,
       7.5810138805866867, 7.5460537811194772, 7.5188417173573718, 7.0847714995612092, 7.2796023609943434,
       7.2138397984782934, 7.3725769872724998, 7.3372388798994574, 7.3128616501582000, 7.2950979818587776,
       7.0054029051641056, 7.1465656548846930, 7.0562366334015811, 6.9836712802178926, 6.9205543224968862,
       6.8033912451575009, 6.1092948405437886, 6.0440995151593730, 5.8759139309024837, 6.4433984121683530,
       6.7376286826414509, 6.7385739276789174, 6.7457547262523869, 6.7347942991663787, 7.0167413401559457,
       6.7344162912605467, 7.6996885390036223, 7.9274006816947820, 7.7595931053437246, 7.3436641132070406,
       7.3841042009443676, 7.3487660935713253, 7.3062476381705901, 7.4744332224274785},
      {5.6206117165433875, 5.3021929950090598, 6.5819155570938959, 6.3256685482431081, 6.9660968411442736,
       6.3526918333251858, 5.9730457430532340, 5.6822170931868436, 5.6011476884863871, 5.3719239647176771,
       6.7812816123054560, 6.6554257047692751, 7.8072139916545211, 7.2036353872039474, 6.8058480712725178,
       6.5199326230903836, 6.3302041925673098, 6.1939546445309794, 7.3009564285805011, 7.2248002255643016,
       7.2727991197809292, 7.0326149689262110, 6.8950430690381346, 6.5618842919052609, 6.7325267022771653,
       6.6304812378963831, 6.5681202959870397, 6.5227670065654557, 6.5050037888118091, 6.4513352265534856,
       7.7531674214903665, 7.2542798862156195, 6.9437979790664253, 6.7202434727936353, 6.5773803633154699,
       6.4764689226521819, 7.5932966595244418, 7.5572029868855131, 7.5602265995863887, 7.4232214862841719,
       7.0369613840246092, 6.8971218872473194, 6.7283690658587956, 6.6539141236917256, 6.6157416328436121,
       6.5873955453640125, 6.5734115055771278, 6.7313926785596712, 8.0545789410827098, 7.6052018811021407,
       7.3243888638300678, 7.1191643618967841, 6.9868836715987959, 6.8908858831655406, 7.9330694486449262,
       8.1976308292409019, 8.1492539271184441, 7.6962869183969156, 7.6492328186719778, 7.6031230628929558,
       7.5670298407998029, 7.5313137255209295, 7.5044796696646561, 7.3710649556491736, 7.3066251955306454,
       7.2958539976704406, 7.3757893786534030, 7.3449869156045624, 7.3241996346042644, 7.3053023931331236,
       7.3788129913542786, 7.8111823988470874, 7.6186191343031657, 7.4772676059025507, 7.3574589237681209,
       7.2119502094949119, 6.9396407931938313, 6.8009348695882608, 6.6505125030850181, 6.6915193774082038,
       8.0997434518242653, 7.6841924675934123, 7.4045130234930578, 7.2172414190850001, 7.0972439582705427,
       7.0020017351031747, 7.9914619833617033, 7.9576359076119880, 8.2454409447775028, 7.8346143835508784,
       7.7696078368464914, 7.7125381045272361, 7.6675628230114841, 7.5245109348532910},
      {5.8156318072023252, 5.3432003198780214, 7.5931074302986383, 6.3088501249811522, 6.6856613812322561,
       6.6208436132078967, 6.1584277584779104, 5.8211117954724418, 5.6869410656452972, 5.4807722197660977,
       7.7231214247989648, 6.6306704671221866, 7.3897734184402868, 7.5273453183283632, 7.0252453916727138,
       6.6750789620520798, 6.4443436572058195, 6.2789924564240023, 8.0695073248155094, 7.7267113735399242,
       7.0978107448564023, 7.1218099666918286, 6.9317039788086987, 6.8795474481737005, 6.7215662751911571,
       6.6121512335568768, 6.5496010624217300, 6.5068933777951905, 6.4874293497381901, 6.4424536176766622,
       7.3447976863787590, 7.5948086911477679, 7.1735884894209949, 6.8808702505712231, 6.6905745829165140,
       6.5537586988400998, 8.2274884977980527, 7.9625495598420208, 7.4419299490752850, 7.3533017378955270,
       7.4986216734887083, 7.1637620860524809, 6.9492784178823177, 6.8525241630916236, 6.8028244585716422,
       6.7695651694077847, 6.7514239437483070, 6.6932201877115576, 7.6052018811021407, 7.9130381834562913,
       7.5218653300582474, 7.2487998979455037, 7.0671970604875902, 6.9345383622837709, 8.4213745647394962,
       8.3813129354537761, 8.2276777270238561, 8.0948297995942333, 8.0559017434802325, 8.0203753079729374,
       7.9910844260016480, 7.9600922831812282, 7.9362818400258304, 7.2436974670354433, 7.1951313356871802,
       7.8999993887068731, 7.8315912213957786, 7.8075915490145773, 7.7911511336584525, 7.7773563230973721,
       7.2780903293710182, 7.5878162207085493, 7.5509669828037342, 7.3653957381532535, 7.2253670121501612,
       7.0823151239919682, 6.9626952205375661, 6.8595161829850655, 6.7548255643550137, 6.6571265150726289,
       7.6414852335126486, 7.9952411613284662, 7.6084142724830448, 7.3478212990796345, 7.1745337344584614,
       7.0422525936146974, 8.4778775104728918, 8.2851259177947174, 8.1241206815655236, 7.7756550622482425,
       7.6482875736345113, 7.5983986398887264, 7.5626829751556297, 7.9444078836367664},
      {5.8379302187343978, 5.2755477178328141, 7.0414970283488101, 6.3339833705340727, 6.4706104759304583,
       6.6958657925066021, 6.1931986287193173, 5.8373634321485390, 5.6266589419451387, 5.4820950221636195,
       7.1900293553228956, 6.6824489898513528, 7.1329596230036403, 7.1641400939583120, 7.0885511280737470,
       6.7075822354042733, 6.4532252660826428, 6.2695436104155444, 7.5832810269301243, 7.2508782656089128,
       7.1316368206061185, 7.0156077669842274, 7.0207097473485121, 7.1244560220326481, 6.7676755804244042,
       6.6493784793675239, 6.5828603515855866, 6.5375070621640035, 6.5080274015126847, 6.4707997051562618,
       7.0819371160861371, 7.2261230279618243, 7.2463430718304869, 6.9214991169885769, 6.7045586227033978,
       6.5458218844549672, 7.7527898641303103, 7.5084480768572224, 7.4251110752675533, 7.3070032034364765,
       7.4668739654024012, 7.3621833467723503, 7.1238887849010135, 7.0088045257708131, 6.9475776075789639,
       6.9065702827100024, 6.8803030134395877, 6.6892517805189913, 7.3243888638300678, 7.5218653300582474,
       7.5872494341226906, 7.2801691475802031, 7.0685198628851120, 6.9120507215258948, 7.9731319790221971,
       7.9241873892223271, 7.7709306392440132, 7.6634060876846659, 7.6197527074748841, 7.5800686355492202,
       7.5473765835169990, 7.5144953022589736, 7.4878500250827278, 7.1864389560361603, 7.1378728246878973,
       7.3892066318544281, 7.3625613546781814, 7.3334597019326955, 7.3139952233299192, 7.2981215945596531,
       7.2607051195232026, 7.4033794503213395, 7.3107828319490151, 7.2263118066418519, 7.1475104493763837,
       6.9972773120989453, 6.9685532167135147, 6.8848386577637894, 6.8077376602558992, 6.6359616767122747,
       7.3455537021904220, 7.5934858887502452, 7.6834364517817493, 7.3873165923252708, 7.1819033117119586,
       7.0254341703527414, 8.0337916600824109, 7.8671176569030727, 8.0817910048448152, 7.7047905193679069,
       7.6189966916632210, 7.5700530029549027, 7.5290461286317170, 7.4840708471159649},
      {5.4974018643480260, 5.2228244006119571, 6.6265128307038168, 6.3073380933578269, 6.3269913506406308,
       6.1667425807688740, 6.0112182339013476, 5.9390308886234902, 5.6449893968304199, 5.4494025195856226,
       6.7895964345964206, 6.6486229141016366, 6.9407743663655497, 6.9156411208126292, 6.8185093086618798,
       6.7892184266905895, 6.4673980845495551, 6.2355287559858006, 7.2168634111791690, 7.9113378236987133,
       7.7694195087122395, 7.6609488110238733, 6.8644293846693225, 6.7457547262523869, 6.7200542435678319,
       6.6629845112485766, 6.5953323597491451, 6.5293810185530665, 6.4845945157173421, 6.4422648389966346,
       6.8857834522554802, 6.9691204538451492, 6.9602383944225501, 7.0001116955740175, 6.7192986783019446,
       6.5131293818769702, 7.4082926520055965, 8.1105151002302467, 7.9738879948338601, 7.8502996841868917,
       7.1002671204256433, 6.9993561303081302, 6.9061922748041713, 6.8315481034112970, 6.6694097445561598,
       6.5855059563806311, 6.5293810185530665, 6.6503237244049904, 7.1191643618967841, 7.2487998979455037,
       7.2801691475802031, 7.3512229196863421, 7.0787242741594572, 6.8763346062470214, 7.6586816646804365,
       8.4243986279861485, 8.2717086645936924, 8.2480865495725464, 8.2106705250818717, 8.1776004651438186,
       8.1511435161018237, 8.1243099107913270, 8.1025782858451141, 7.8302684189982568, 7.8217648180272645,
       7.2554139099331145, 7.9925955565334217, 7.9678407694321089, 7.9487538481893880, 7.9357150534399707,
       7.1851161536386376, 7.2949092031787499, 7.2026905927122566, 7.1212431801059690, 7.0471657952989544,
       6.9829152644062296, 6.8402409336080927, 6.6021356009625594, 6.4889409308157404, 6.5991119882616838,
       7.1320148285119496, 7.3094600295514933, 7.3640729357557317, 7.4602599534147913, 7.1938085332896575,
       6.9912300866971941, 7.7284126343890529, 8.3954853033749135, 8.4767439373011726, 8.1484979113067801,
       8.1256327131888479, 8.0899170484557512, 8.0525001228735249, 8.0725313880621599},
      {5.2545721086982633, 5.1417554464572763, 6.3895415217757776, 6.6061040081551257, 6.2196551272155354,
       6.0259582894998944, 5.8562606736196798, 5.7436332406044963, 5.6293045467401823, 5.4012148466889673,
       6.5603727108277106, 6.9698760191110365, 6.8228557237602772, 6.7470775286499087, 6.6251900283062941,
       6.5420422559424285, 6.4511464478734579, 6.1873406325433695, 7.0071037154674594, 7.5160068833365239,
       7.3718209714608367, 7.2582482934081858, 7.1703760980400908, 6.5129406031969417, 7.0363945974387496,
       6.9557032006441242, 6.8886178357305523, 6.8347609453379761, 6.7735335766003519, 6.7228890775886798,
       6.7599275447192992, 6.7952656520923407, 6.7591719794534120, 6.7421643269656517, 6.7058814251009196,
       6.4681541003612173, 7.2011785610889314, 7.7225537371215545, 7.5908402839552007, 7.4598819455089593,
       6.8706653887511013, 6.7652187543093874, 6.6631737404743800, 6.5834271381714462, 6.5197433938645801,
       6.4636184560370165, 6.3405973825216835, 6.8950430690381346, 6.9868836715987959, 7.0671970604875902,
       7.0685198628851120, 7.0787242741594572, 7.0637954398808827, 6.8298477436537190, 7.4623387716239762,
       8.0415401463332916, 7.8962202107401103, 7.8654172971454948, 7.8274335849774088, 7.7932299518676373,
       7.7660178881055320, 7.7386165951176231, 7.7159397251339437, 7.4468431507595421, 7.4258670910792155,
       7.3638841570757041, 7.6018002604954340, 7.5762894575824582, 7.5564469710738500, 7.5432189470986293,
       7.4309695219892768, 7.4389063363744095, 7.3438528918870682, 7.2661851077933193, 7.2009897824089037,
       7.1184087966308969, 6.5760571103721723, 6.5038697650943149, 6.3207553465588511, 6.5726559403112415,
       6.9891512684880093, 7.1189755832167565, 7.1448648445813392, 7.1800137227285772, 7.1837933512411158,
       6.9494671965623454, 7.5354713619393001, 8.0243437151655037, 8.1331910691223737, 7.7869934972400827,
       7.7573250579087363, 7.7182077725689320, 7.6804132896266495, 7.6891061198234452},
      {5.0735360578262094, 5.0531272352775183, 6.2219227241047479, 6.3090389036611798, 6.1365069043058931,
       5.9118188248613848, 5.7260588015308773, 5.5805496367118916, 5.5939668899129167, 5.3426330827463868,
       6.3929426918367085, 6.6816929740396898, 6.7410303032481576, 6.6233004393229136, 6.4787365195413944,
       6.3534473985910731, 6.2362847717974637, 6.1323497184332991, 6.8425085304973052, 7.2374614629536635,
       7.0945983534754991, 6.9768680390044784, 6.8854054443496482, 6.3491014340384506, 6.7465107420640500,
       6.6643073136460993, 6.5958995968807796, 6.5365622676723127, 6.4938541324999974, 6.4396192342015901,
       6.6682761713844414, 6.6669533689869187, 6.6081828263643105, 6.5479002521183762, 6.4825156975081573,
       6.4180763879354048, 7.0352605737212555, 7.4458983562678513, 7.3232548401125737, 7.1875725292078787,
       6.7083382512159364, 6.5976004071841334, 6.4853505315290052, 6.4010687354476454, 6.3343613784399038,
       6.2755908358172956, 6.2262691392031453, 6.6238672259087723, 6.8908858831655406, 6.9345383622837709,
       6.9120507215258948, 6.8763346062470214, 6.8298477436537190, 6.7820376281171191, 7.3022792309780238,
       7.7614826943271060, 7.6318471582783873, 7.5925406437127787, 7.5541793741846375, 7.5197865118490625,
       7.4920076615010975, 7.4644175898331611, 7.4413627119436496, 7.1805805093144368, 7.1378728246878973,
       7.1191643618967841, 7.3251448796417309, 7.2994443969571758, 7.2790355744084847, 7.2652403133016286,
       7.1681085011508783, 7.1631952994666213, 7.0636066612008550, 6.9823484778203708, 6.9143183184151074,
       6.8289024986162525, 6.3753682527630895, 6.3050704964686135, 6.2270251550148084, 6.5554595091434535,
       6.8810590292512508, 6.9774352761361138, 6.9823484778203708, 6.9723328452260525, 6.9441759869722564,
       6.9084603222391596, 7.3814585961493231, 7.7580810737203993, 7.8918733450959371, 7.5400065557177252,
       7.5041016617588250, 7.4617719850381166, 7.4237882728700306, 7.4177410474682794},
      {6.4190211824270955, 5.6922327257811611, 7.9823915958048524, 7.1596049001798860, 7.5681634139715213,
       6.6866061757239468, 6.1790258102524049, 5.8738351126932988, 5.6534929978014121, 5.4909766310404438,
       8.1900724733073762, 7.5400065557177252, 8.8038554890322960, 7.8004107504411069, 7.2045801816956381,
       6.8387293525305424, 6.5858839642864622, 6.3984231306526009, 8.7930838406263163, 8.1511435161018237,
       7.9856039871857556, 7.8062687466170546, 7.6896729064093048, 7.7372937927201013, 7.5305577097092673,
       7.4485439610628958, 7.3910962208378095, 7.3459317100962531, 7.3200424487316704, 7.2703427442116890,
       8.6978420680047233, 7.8495436683752295, 7.3687973587599611, 7.0711654676801565, 6.8702878313910460,
       6.7175978679985908, 9.1373918235252258, 8.5818121132912797, 8.2941963053515693, 8.2044340704543171,
       8.1717415678763192, 8.0791454000497716, 7.9468642592060075, 7.8971645546860261, 7.8773225187231937,
       7.8584257277978287, 7.8491661110151734, 7.6683188388231471, 7.9330694486449262, 8.4213745647394962,
       7.9731319790221971, 7.6586816646804365, 7.4623387716239762, 7.3022792309780238, 9.5557776417769258,
       9.5344235741907699, 8.8824676170719599, 8.6553231620582114, 8.5980642005131536, 8.5523329031857376,
       8.5098139972392257, 8.4680511071043778, 8.4385709959072841, 7.8731648823048239, 7.9621711013904131,
       8.5555452945666417, 8.3310455432563852, 8.3026999063225606, 8.2830470995855325, 8.2675510281753226,
       8.0016668451818251, 9.0585904662597585, 8.8989080324280856, 8.7528329821147928, 8.6048674417265669,
       8.4466970395182202, 8.3792341172445912, 8.1832692320939628, 7.9667062951688390, 7.6382719410401938,
       8.0020444025418804, 8.5491205118048335, 8.0814134474847599, 7.7790566828549492, 7.6008559165495191,
       7.4324811030668263, 9.5790203972549133, 9.0126699397065391, 8.7299668829053090, 8.4533110515058301,
       8.3752657100520249, 8.3454080414948759, 8.3011878746992345, 8.4790110836446111},
      {5.7882305142144164, 6.2873072787149669, 7.6405399884751821, 7.5675966273856625, 7.8527560597561328,
       7.0724882700776783, 6.6386072815073192, 6.6469221037982837, 6.2141746883996429, 5.9295820426150314,
       7.9168182625146057, 7.9975092087634545, 7.8848817757482710, 7.9052905982969621, 7.3565141292764302,
       7.6617048268355363, 7.2032573792981163, 6.8812482584770542, 8.7632261720691655, 8.7433841361063340,
       8.4565234428867342, 8.3401168319047869, 8.2123708848394497, 7.2909407959861836, 8.0470196840576325,
       7.9175733772347172, 7.8540788621536546, 7.7839703350849820, 7.7609159077412473, 7.6796572738149873,
       7.7964423432485406, 7.9319358754732079, 7.4704648152349122, 7.8979205704976883, 7.4946532662961420,
       7.2055249761873288, 9.1009201429804651, 9.1220849813408194, 8.7482968872448161, 8.6625039606316818,
       7.7044129620078516, 7.5755334417707951, 7.3893954105344557, 7.2822479657893879, 7.2421858859578929,
       7.1800137227285772, 7.1480772359622433, 7.9924072283991698, 8.1976308292409019, 8.3813129354537761,
       7.9241873892223271, 8.4243986279861485, 8.0415401463332916, 7.7614826943271060, 9.5344235741907699,
       9.7914256977616674, 9.1848243817017696, 9.2832785457042668, 9.2265872718366193, 9.1780211404883563,
       9.1379586101110846, 9.0961957199762367, 9.0657712648332271, 8.5621593065542516, 8.9399158078428229,
       8.8216187067859426, 8.9000425066913547, 8.8573343715190394, 8.8303115369827392, 8.8093350267566368,
       8.4249654145720072, 8.9591910572197957, 8.9111917124573914, 8.7983750502164035, 8.6951955621181263,
       8.5959853823039669, 7.6581139770030262, 7.4156626798048704, 7.1435420421838174, 7.8913065585100783,
       8.2511106128191987, 8.4723970716569994, 7.9958088490058774, 8.5476084801815091, 8.1649392277544557,
       7.8716537517730494, 9.5329115425674438, 9.5712728120955841, 9.1942723266186768, 8.7224076258802317,
       8.8261539005643677, 8.7951626588355012, 8.7509424920398597, 9.0313784024976531},
      {5.7886085221202483, 6.1722230195847665, 6.8846494285379851, 7.1975881618021962, 7.1639508647325085,
       7.0639846691066861, 6.6316152616138773, 6.6433317045115485, 6.2449771514484835, 5.9868410041600901,
       7.1565812874790105, 7.5445417494961511, 7.6792797164549311, 7.8200635571781358, 7.2926416062895374,
       7.5898950389177351, 7.1579040898765331, 6.8627285743659687, 7.8879049379033708, 8.1571907415035767,
       8.0114932485503392, 7.9049130409369059, 7.7680967063147168, 6.7759904027153679, 7.6099263041063701,
       7.4976768789970176, 7.4574255699397192, 7.3982770194112790, 7.3738997896700216, 7.3147512391415814,
       7.6042575371562258, 7.8520000439444697, 7.3903402050261464, 7.8115599562071436, 7.4245442886816937,
       7.1569592953848424, 8.2516773994050574, 8.5188843847960776, 8.3586360654700975, 8.1551119232943918,
       7.2680751473224765, 7.1822813196177906, 6.9902852922055034, 6.8874842625588331, 6.8612169932884184,
       6.7952656520923407, 6.7555811296209010, 7.6267451779141018, 8.1492539271184441, 8.2276777270238561,
       7.7709306392440132, 8.2717086645936924, 7.8962202107401103, 7.6318471582783873, 8.8824676170719599,
       9.1848243817017696, 8.7420613337088113, 8.6664723678242481, 8.6173394498901246, 8.5704736782994395,
       8.5332468830345682, 8.4960191866781454, 8.4684286644644331, 8.2167177504836229, 8.2745425975229896,
       8.1908284891190402, 8.3268888079295671, 8.2924950445024397, 8.2692513879329006, 8.2475197629866877,
       8.1475535673608661, 8.5908834019396831, 8.5413720255539545, 8.4155165685635485, 8.3074243293267909,
       8.1575691999551836, 7.3720097501408643, 7.1418412318804636, 6.8844606498579575, 7.5736438527874137,
       8.1743871726713628, 7.8555908937769807, 7.8236544070106460, 8.3822572793996919, 8.0024228609934873,
       7.7250110137823462, 8.7278880646961241, 8.9506865551572510, 8.8123590900032873, 8.4243986279861485,
       8.5553560653408383, 8.5047120168749419, 8.4657830596693895, 8.4882707004272646},
      {5.5661875890191768, 6.1359401177200343, 6.8234225103461368, 7.2796023609943434, 6.9211211090827458,
       6.9160191287184611, 6.5027357413768208, 6.5497902916475335, 6.1255464772198849, 5.8475678434228842,
       7.0891179146596066, 7.6339259764875722, 7.4135838615956855, 7.6649172182164396, 7.1545024692698256,
       7.5296129152175766, 7.0777794796677664, 6.7655967622152184, 7.8170403950230352, 8.2356145414089887,
       8.0031779757135997, 7.9389274448208740, 7.8238436362364494, 6.6860393891380880, 7.6823028786100309,
       7.5764786868082616, 7.5367941643368219, 7.4776456138083818, 7.4510003366321360, 7.3910962208378095,
       7.3185308676541201, 7.6877833174259234, 7.2472878663221776, 7.7580810737203993, 7.3563253505964026,
       7.0760786693644135, 8.2042448412285136, 8.5976857420615449, 8.2832354277197844, 8.1488763697583870,
       7.1905961419087543, 7.1021571599547997, 6.9205543224968862, 6.8234225103461368, 6.7975332489815532,
       6.7327154809571939, 6.6911418200481485, 7.7063025509912331, 7.6962869183969156, 8.0948297995942333,
       7.6634060876846659, 8.2480865495725464, 7.8654172971454948, 7.5925406437127787, 8.6553231620582114,
       9.2832785457042668, 8.6664723678242481, 8.7307228987169729, 8.6832903405404274, 8.6422834662172416,
       8.6065678014841449, 8.5700961209393842, 8.5432616145373359, 8.1090030686069206, 8.2947630919374280,
       8.2465754190407718, 8.4070129675925571, 8.3745096942403627, 8.3516435950308789, 8.3304787566705247,
       8.0232092409022346, 8.6288662130162166, 8.6061902441240896, 8.4954524000922866, 8.3947292875632495,
       8.2478982214382945, 7.3595377419773058, 7.1097159664341012, 6.8273909175387031, 7.6439416090818897,
       7.7231214247989648, 8.1643724411685969, 7.7129165629788439, 8.3575015912068267, 7.9729427497963936,
       7.6875940882001199, 8.7074792421474339, 9.0255204063217054, 8.6696847592051505, 8.3514552668966271,
       8.4245869561204003, 8.3986976947558158, 8.3609032118135342, 8.5683957611818062},
      {5.5263147384134852, 6.0926656465076352, 6.7742895924120141, 7.2363274392361694, 6.8776574086445432,
       6.8786026536820097, 6.4636184560370165, 6.5193653859587490, 6.0949332433968486, 5.8163873724682125,
       7.0405517833113436, 7.5910295131810051, 7.3704981690633149, 7.6292015534833428, 7.1153851839300213,
       7.4969208631853546, 7.0454654355413764, 6.7330934888630249, 7.7652618722938689, 8.1913952757048989,
       7.9589587100095098, 7.8977313412718848, 7.7817022876499937, 6.6403080918106729, 7.6394064153034638,
       7.5337705516359463, 7.4948420449761697, 7.4351267078618708, 7.4077258654197378, 7.3478212990796345,
       7.2750671672159184, 7.6526344392786854, 7.2083598102081767, 7.7253885711424015, 7.3238220772442091,
       7.0431973881063881, 8.1500099429301063, 8.5494980691648905, 8.2350477548231282, 8.1012554834475914,
       7.1424080184663232, 7.0520794475289872, 6.8717994124685955, 6.7754236161295083, 6.7512347145225036,
       6.6854726025522284, 6.6433317045115485, 7.6584924354546331, 7.6492328186719778, 8.0559017434802325,
       7.6197527074748841, 8.2106705250818717, 7.8274335849774088, 7.5541793741846375, 8.5980642005131536,
       9.2265872718366193, 8.6173394498901246, 8.6832903405404274, 8.6364254700412939, 8.5954185957181082,
       8.5598912591192633, 8.5236088078003061, 8.4969635306240612, 8.0636493286395616, 8.2484650080241533,
       8.1983868450525659, 8.3607148836792824, 8.3285891676871433, 8.3057230684776613, 8.2845582301173071,
       7.9771003862147634, 8.5680173027301993, 8.5472300217299022, 8.4359253911122387, 8.3353924089005584,
       8.1896940148557693, 7.3009564285805011, 7.0537802578323410, 6.7752343869037048, 7.5946194619219636,
       7.6745552934507018, 8.1244991400171305, 7.6685080680489506, 8.3206523533020107, 7.9349590376283077,
       7.6494220478977812, 8.6511655256398416, 8.9758207018017231, 8.6247094776893984, 8.3061015269292682,
       8.3775328563954634, 8.3518328242566824, 8.3140383413144008, 8.5207739737794590},
      {5.4888978128312580, 6.0554379501512123, 6.7313926785596712, 7.1956985728188148, 6.8379733367188802,
       6.8445873487064901, 6.4286588071155819, 6.4929093380083067, 6.0679099583147709, 5.7891753087061071,
       6.9989781224022991, 7.5509669828037342, 7.3306248679118475, 7.5955647069594301, 7.0789135033852615,
       7.4676299812140643, 7.0161745535700861, 6.7036138282117070, 7.7236882113848244, 8.1518995319134877,
       7.9190854088580425, 7.8599368583296023, 7.7427742315359929, 6.5987339803558518, 7.5997214422862491,
       7.4940860291645066, 7.4557247596363654, 7.3952534067104034, 7.3672857776824108, 7.3073812113423084,
       7.2355718739702821, 7.6199419367006875, 7.1722656870234722, 7.6959093610368594, 7.2939644086870592,
       7.0133397195492382, 8.1061682345860717, 8.5077351790300408, 8.1872376392865291, 8.0587365775010795,
       7.0998895630655872, 7.0076709525990948, 6.8287137199362249, 6.7327154809571939, 6.7098498322934859,
       6.6433317045115485, 6.6006235693392332, 7.6150282844706547, 7.6031230628929558, 8.0203753079729374,
       7.5800686355492202, 8.1776004651438186, 7.7932299518676373, 7.5197865118490625, 8.5523329031857376,
       9.1780211404883563, 8.5704736782994395, 8.6422834662172416, 8.5954185957181082, 8.5549785079807812,
       8.5194511713819363, 8.4831687200629808, 8.4565234428867342, 8.0152724265171020, 8.2091584934585455,
       8.1564347256919127, 8.3202747959419554, 8.2879598507240146, 8.2650946526060824, 8.2439298142457282,
       7.9264554366573163, 8.5149159776035113, 8.4960191866781454, 8.3839585402488215, 8.2836138861713913,
       8.1384831798040143, 7.2505007082488575, 7.0054029051641056, 6.7293143108962621, 7.5498329590862392,
       7.6276895218600167, 8.0887825741924821, 7.6286347668974832, 8.2883383091756215, 7.9011329618785924,
       7.6154067429222625, 8.6054342283124257, 8.9331125666294078, 8.5835133741404093, 8.2654722099661377,
       8.3268888079295671, 8.3021331197367019, 8.2650946526060824, 8.4788227555103575},
      {5.4571505552907276, 6.0263358468599506, 6.6918973853140358, 7.1635733073724523, 6.7981000355674128,
       6.8177532928502167, 6.4006907275418135, 6.4709884838362894, 6.0456110962369225, 5.7661208813623723,
       6.9600496157425225, 7.5180857015457088, 7.2903735588545482, 7.5681634139715213, 7.0494338427339427,
       7.4440083167386941, 6.9923641104146883, 6.6796141558305049, 7.6830588944216940, 8.1173174403521084,
       7.8811016966899565, 7.8280003715632684, 7.7100817289579959, 6.5605614895077382, 7.5670298407998029,
       7.4615827558123131, 7.4239775020958341, 7.3633173704898445, 7.3349712830102449, 7.2754447245759737,
       7.1947537783271240, 7.5929191021643865, 7.1424080184663232, 7.6719096886556581, 7.2697755070800536,
       6.9887732605821773, 8.0649721310370825, 8.4716410558453372, 8.1507659587417685, 8.0207528653329945,
       7.0619058508975012, 6.9685532167135147, 6.7907300077681390, 6.6949209980149114, 6.6733781517487261,
       6.6061040081551257, 6.5630183156227551, 7.5810138805866867, 7.5670298407998029, 7.9910844260016480,
       7.5473765835169990, 8.1511435161018237, 7.7660178881055320, 7.4920076615010975, 8.5098139972392257,
       9.1379586101110846, 8.5332468830345682, 8.6065678014841449, 8.5598912591192633, 8.5194511713819363,
       8.4841139651004465, 8.4480198419157428, 8.4215637939652996, 7.9803127775956675, 8.1743871726713628,
       8.1205307328245642, 8.2849366885689140, 8.2528109725767766, 8.2299457744588445, 8.2087809360984902,
       7.8907397719242187, 8.4720195142969441, 8.4565234428867342, 8.3440852390973532, 8.2439298142457282,
       8.0993649933726584, 7.2098713912857271, 6.9655300545584140, 6.6907638121423174, 7.5154400967506643,
       7.5910295131810051, 8.0591141348611348, 7.5955647069594301, 8.2620705893594302, 7.8737325699822343,
       7.5876278925742975, 8.5640488955376330, 8.8960740994987884, 8.5474192509557056, 8.2261656954005300,
       8.2907946847448617, 8.2658506684177446, 8.2291897586471805, 8.4432954189115126},
      {5.4252140685243928, 5.9974234233402672, 6.6516465268025131, 7.1306920261144278, 6.7655967622152184,
       6.7905412290881113, 6.3729118771938493, 6.4492568588900765, 6.0238794712907096, 5.7440107979645516,
       6.9209323304027182, 7.4848264123818522, 7.2559806965189733, 7.5386837533202034, 7.0193869449509902,
       7.4201978735832963, 6.9687424459393181, 6.6559929419009105, 7.6435631506302828, 8.0825470206564773,
       7.8467088343543816, 7.7945527542651591, 7.6772008982457463, 6.5252238326804726, 7.5335813224101429,
       7.4285126958742600, 7.3916630074236682, 7.3306248679118475, 7.3020900017522195, 7.2425638938637240,
       7.1605496946715768, 7.5636282201930962, 7.1121727925491180, 7.6479100162744560, 7.2457762852446281,
       6.9643960308409198, 8.0245320432997573, 8.4347918179405212, 8.1142942781970095, 7.9844704140140372,
       7.0239225892751911, 6.9305699550912045, 6.7536915406375195, 6.6584493174701516, 6.6378512656956561,
       6.5701991141962246, 6.5265466350779944, 7.5460537811194772, 7.5313137255209295, 7.9600922831812282,
       7.5144953022589736, 8.1243099107913270, 7.7386165951176231, 7.4644175898331611, 8.4680511071043778,
       9.0961957199762367, 8.4960191866781454, 8.5700961209393842, 8.5236088078003061, 8.4831687200629808,
       8.4480198419157428, 8.4121149479568409, 8.3856589000063995, 7.9457306860342882, 8.1396167529757335,
       8.0840590522798035, 8.2495985811958725, 8.2174728652037352, 8.1946076670858030, 8.1736320579512522,
       7.8554016645511773, 8.4289338217645735, 8.4147605527518863, 8.3026999063225606, 8.2027337106967391,
       8.0587365775010795, 7.1688645169625413, 6.9258455320869752, 6.6525913212942038, 7.4801024399233986,
       7.5547466113162720, 8.0277444346806597, 7.5621161885697701, 8.2354253121831853, 7.8463312769943254,
       7.5600373703605852, 8.5228527919886439, 8.8584679446907586, 8.5117035862226071, 8.1906392598932367,
       8.2550790200117650, 8.2301341025930963, 8.1934740939140838, 8.4066345091409485},
      {5.3985692418939228, 5.9730457430532340, 6.6234896685487170, 7.1049915434298718, 6.7397075008506357,
       6.7657859914410228, 6.3496682206243102, 6.4309264040047944, 6.0053602377253998, 5.7251135564934108,
       6.8935314879605851, 7.4589371510172686, 7.2304694430602217, 7.5160068833365239, 6.9948204859839285,
       7.4005446163004915, 6.9487111807506823, 6.6357724474864712, 7.6167295453197843, 8.0566577592918946,
       7.8206312448555462, 7.7697970660722948, 7.6513116368811627, 6.4942321404058285, 7.5082592981771947,
       7.4031902210955360, 7.3667185405507762, 7.3054916223589270, 7.2767675269734964, 7.2170521898591966,
       7.1350384412128252, 7.5413293581152478, 7.0876063335820563, 7.6278787510858201, 7.2253670121501612,
       6.9437979790664253, 7.9961864063659327, 8.4073905249526124, 8.0865154278490436, 7.9566906625745215,
       6.9925528890947160, 6.8993894841365329, 6.7232670854945109, 6.6282136410071706, 6.6085608342701416,
       6.5401526669590471, 6.4963109586150134, 7.5188417173573718, 7.5044796696646561, 7.9362818400258304,
       7.4878500250827278, 8.1025782858451141, 7.7159397251339437, 7.4413627119436496, 8.4385709959072841,
       9.0657712648332271, 8.4684286644644331, 8.5432616145373359, 8.4969635306240612, 8.4565234428867342,
       8.4215637939652996, 8.3856589000063995, 8.3590136228301528, 7.9194638673096494, 8.1146718355570648,
       8.0570353166519499, 8.2229533040196259, 8.1908284891190402, 8.1679623899095564, 8.1469867807750056,
       7.8280003715632684, 8.3958637618265204, 8.3830132952113541, 8.2701966329703662, 8.1706079947046000,
       8.0271776480948009, 7.1352272198928528, 6.8929642508289497, 6.6208436132078967, 7.4519451311238267,
       7.5275345475541675, 8.0037456633910100, 7.5350933540334681, 8.2138829164627758, 7.8236544070106460,
       7.5367941643368219, 8.4933735818831018, 8.8306890943427945, 8.4850583090463623, 8.1628604095452708,
       8.2273001696637991, 8.2025444814709356, 8.1658835717003715, 8.3792341172445912},
      {5.0640876623635265, 5.6977127140512778, 6.1578609718920507, 6.6692209658761321, 6.6879289781214695,
       6.5955215889749486, 6.2147419255312784, 6.2969449034034533, 5.9012364056812077, 5.6480130095312964,
       6.4273360047180601, 7.0193869449509902, 7.1488332517739055, 6.9286803661078231, 6.7837379878746971,
       7.1955093435930113, 6.7671087938385446, 6.4751461202546601, 7.1291799944911016, 7.6169187745455877,
       7.5141172943531425, 7.3491441014771572, 7.2733663569125646, 6.1805378418757311, 7.1044247568440131,
       6.9889624898079807, 6.9489004099764857, 6.8838938632720987, 6.8489337638048884, 6.7937536204690145,
       7.0859055232787034, 6.9494671965623454, 6.8717994124685955, 7.4082926520055965, 7.0216545418402028,
       6.7550143430350422, 7.4632835661156669, 7.9512111248501807, 7.8542680913794589, 7.6549015856221221,
       6.6762129857695740, 6.5724667110854371, 6.3742346795913711, 6.2599059857270580, 6.2069938898261734,
       6.1276252954290698, 6.0726343813190002, 7.0847714995612092, 7.3710649556491736, 7.2436974670354433,
       7.1864389560361603, 7.8302684189982568, 7.4468431507595421, 7.1805805093144368, 7.8731648823048239,
       8.5621593065542516, 8.2167177504836229, 8.1090030686069206, 8.0636493286395616, 8.0152724265171020,
       7.9803127775956675, 7.9457306860342882, 7.9194638673096494, 7.7937967384534970, 7.7954970982110741,
       7.6067139127254668, 7.7784898962690905, 7.7448530497451777, 7.7216093931756395, 7.6996885390036223,
       7.6760664239824763, 8.0022336317676839, 7.8941413925309254, 7.8049459442195328, 7.7219869505356948,
       7.5715650345782288, 6.6034584033600812, 6.3965335416692195, 6.1754354109656697, 7.0329929768320429,
       7.3958206438420380, 7.2992556182771473, 7.2399178385229046, 7.9436518678251034, 7.5532345796929468,
       7.2754447245759737, 7.9202189820297608, 8.3665728798552301, 8.1791115956755931, 7.9633055756536830,
       8.0744209770455413, 8.0283121223580700, 7.9888163785666597, 7.9411954922558632},
      {4.7600305379761387, 5.6130529100640860, 6.2480007641493591, 6.8474221827273380, 6.5227670065654557,
       6.4229895896196618, 6.0673431717289112, 6.1291373270523959, 5.6869410656452972, 5.4008368387831354,
       6.4728780728196709, 7.2106274070973901, 7.0407410125371470, 6.8279577041245618, 6.6701657603678228,
       7.0989443180281206, 6.6319932695197084, 6.3056377336002489, 7.1601716867657457, 7.8041899284078697,
       7.5948086911477679, 7.4782124003942414, 7.4147178853131797, 6.1750574030598386, 7.2716655466092108,
       7.2106274070973901, 7.1159519705158809, 7.0518902183031837, 7.0157965456642550, 6.9568372243616183,
       6.9528688171690520, 6.8442093408006590, 6.7629511574201748, 7.3251448796417309, 6.9111059270342041,
       6.6181980084128522, 7.5092040926688854, 8.1740096153113075, 7.9043453532594956, 7.7777338804574274,
       6.6287808781388051, 6.5375070621640035, 6.3494789913985068, 6.2527251871535885, 6.1557821536828659,
       6.0781143695891169, 6.0199110640981432, 7.2796023609943434, 7.3066251955306454, 7.1951313356871802,
       7.1378728246878973, 7.8217648180272645, 7.4258670910792155, 7.1378728246878973, 7.9621711013904131,
       8.9399158078428229, 8.2745425975229896, 8.2947630919374280, 8.2484650080241533, 8.2091584934585455,
       8.1743871726713628, 8.1396167529757335, 8.1146718355570648, 7.7954970982110741, 8.0175404739520904,
       7.7796234694408088, 7.9755883545914381, 7.9436518678251034, 7.9200306538955090, 7.9003769460669293,
       7.6522559808270776, 8.3533448558800085, 8.2720862219537477, 8.2010324498476095, 8.0946414714599797,
       8.0028004183535426, 6.5900411501590561, 6.3370069832349483, 6.0777368122290607, 7.2240442097526385,
       7.3368608719936264, 7.2656183212074597, 7.2036353872039474, 7.9438410970509068, 7.5481321487828863,
       7.2487998979455037, 7.9963756355917361, 8.5931505482831199, 8.3106367207076932, 8.0009108293701612,
       8.0405949012958260, 8.0148948691570467, 7.9776671728006230, 8.1488763697583870},
      {5.7600736559606203, 5.5905652693062100, 6.9766792603244507, 6.8011236482682884, 7.3474432911738035,
       6.4645637010744830, 5.9726677351474020, 6.0354066849625774, 5.6644534248874203, 5.4133088469466939,
       7.2070370078106549, 7.1686752877367379, 7.0072929446932628, 7.3992218139029697, 6.7860060353096854,
       6.5777579206755261, 6.5924979762740730, 6.3111177218703647, 7.8113716280728909, 7.7709306392440132,
       7.5377389588285126, 7.4725431828983213, 7.3542465323872177, 6.7829824226088098, 7.2057142054131331,
       7.1119835633233146, 7.0711654676801565, 7.0054029051641056, 6.9693092325251769, 6.9114834843942594,
       6.9061922748041713, 7.4407959253577909, 6.9124287294317259, 6.7784467782846090, 6.8712326258827368,
       6.6214108503395321, 8.1539783501226726, 8.1216643059962816, 7.8085367940520438, 7.8041899284078697,
       7.2119502094949119, 7.1204871642943059, 7.0210877552543431, 6.9651520466525829, 6.9369951883987868,
       6.9288691447878508, 6.9007122865340547, 7.2138397984782934, 7.2958539976704406, 7.8999993887068731,
       7.3892066318544281, 7.2554139099331145, 7.3638841570757041, 7.1191643618967841, 8.5555452945666417,
       8.8216187067859426, 8.1908284891190402, 8.2465754190407718, 8.1983868450525659, 8.1564347256919127,
       8.1205307328245642, 8.0840590522798035, 8.0570353166519499, 7.6067139127254668, 7.7796234694408088,
       7.8501104549610883, 7.9279674682806416, 7.8973537839118295, 7.8756221589656157, 7.8571029254003060,
       7.5500221883120435, 8.1483095831725283, 8.1188294719754346, 7.9903284101899850, 7.8754329297398122,
       7.7399393975151449, 7.3774901889567568, 7.2217766128634260, 7.0048361185782468, 7.1480772359622433,
       7.3241996346042644, 7.9873043469433336, 7.4557247596363654, 7.3447976863787590, 7.4818027996809766,
       7.2280126169452057, 8.6046782125007635, 8.5357032586038102, 8.2170953078436781, 7.8873381513175111,
       7.9428958520134403, 7.9192746380838459, 7.8756221589656157, 8.0710193564388337},
      {5.3169330506076067, 5.8488906458204069, 6.5450658686433041, 6.9683644380334862, 6.6115839964252423,
       6.6486229141016366, 6.2241903209939604, 6.3456998134317439, 5.9195664100207139, 5.6374305903511193,
       6.8177532928502167, 7.3240108559242367, 7.1083931640365794, 7.4118830512923317, 6.8729334361860897,
       7.3032240254697145, 6.8513905899199044, 6.5375070621640035, 7.5350933540334681, 7.9251326342597936,
       7.6872165308400637, 7.6401624311151268, 7.5175189149598491, 6.4035255615626614, 7.3714429635550056,
       7.2684527046825318, 7.2353826447444787, 7.1715101217575858, 7.1412744452946049, 7.0811811002744740,
       7.0140957353609013, 7.4400399095461278, 6.9679864301276551, 7.5298021444433800, 7.1259676031101984,
       6.8426977597231087, 7.9013221911043958, 8.2675510281753226, 7.9451638994484295, 7.8121267427930023,
       6.8897518594480465, 6.7895964345964206, 6.6208436132078967, 6.5341054415572959, 6.5248458247746415,
       6.4526580289510083, 6.4069267316235932, 7.3725769872724998, 7.3757893786534030, 7.8315912213957786,
       7.3625613546781814, 7.9925955565334217, 7.6018002604954340, 7.3251448796417309, 8.3310455432563852,
       8.9000425066913547, 8.3268888079295671, 8.4070129675925571, 8.3607148836792824, 8.3202747959419554,
       8.2849366885689140, 8.2495985811958725, 8.2229533040196259, 7.7784898962690905, 7.9755883545914381,
       7.9279674682806416, 8.0868929852091007, 8.0543897118569063, 8.0315245137389741, 8.0105480035128718,
       7.6900504637693601, 8.2108588532161235, 8.2025444814709356, 8.0872714436607076, 7.9882495919808001,
       7.8504889134126961, 7.0127729329633794, 6.7784467782846090, 6.5163417732578734, 7.3019012230721918,
       7.3969542170137572, 7.8988658155351548, 7.4096154544031192, 8.1059799064518199, 7.7100817289579959,
       7.4207646601691550, 8.3852813426463424, 8.6880147635446576, 8.3529663974284016, 8.0233984701280381,
       8.0868929852091007, 8.0608153957102644, 8.0247212725255608, 8.2367481145807062},
      {5.2972797933248028, 5.8213005741524695, 6.5231450144712877, 6.9352943780954330, 6.5813487705080362,
       6.6234896685487170, 6.1981122809493501, 6.3271805798664342, 5.9021812001728984, 5.6213677323550497,
       6.7975332489815532, 7.2916963612520709, 7.0791022820652891, 7.3888286239485961, 6.8460993803298162,
       7.2820587365635845, 6.8307925381454098, 6.5176645756553953, 7.5143060730331701, 7.8939521633051219,
       7.6554683722079817, 7.6099263041063701, 7.4859604360993464, 6.3785810946897694, 7.3389396902028112,
       7.2359494313303383, 7.2038241658839759, 7.1393848563112234, 7.1089599506224381, 7.0486778269222805,
       6.9853720905212464, 7.4177410474682794, 6.9417191608572404, 7.5084480768572224, 7.1048027647498442,
       6.8217217000427830, 7.8769449613631384, 8.2339141816514108, 7.9107710371128537, 7.7767895365115125,
       6.8619730091000815, 6.7599275447192992, 6.5915531817823823, 6.5061373619835274, 6.4997125792217210,
       6.4267687675864247, 6.3808486915789819, 7.3372388798994574, 7.3449869156045624, 7.8075915490145773,
       7.3334597019326955, 7.9678407694321089, 7.5762894575824582, 7.2994443969571758, 8.3026999063225606,
       8.8573343715190394, 8.2924950445024397, 8.3745096942403627, 8.3285891676871433, 8.2879598507240146,
       8.2528109725767766, 8.2174728652037352, 8.1908284891190402, 7.7448530497451777, 7.9436518678251034,
       7.8973537839118295, 8.0543897118569063, 8.0218864385047119, 7.9990212403867806, 7.9780447301606792,
       7.6564136172454482, 8.1622936229594121, 8.1558679391060540, 8.0402173439357707, 7.9415730496159185,
       7.8049459442195328, 6.9772460469103095, 6.7453767183465558, 6.4861065473406683, 7.2652403133016286,
       7.3653957381532535, 7.8744876847023457, 7.3803245724318289, 8.0816017756190117, 7.6843816968192158,
       7.3950646280303758, 8.3571240338467714, 8.6536219012090818, 8.3221643849253368, 7.9893831651525185,
       8.0526893520993283, 8.0264216322831388, 7.9903284101899850, 8.2027337106967391},
      {5.2849965638412719, 5.7993801705262289, 6.5091609746844030, 6.9128062867917812, 6.5605614895077382,
       6.6047812057576039, 6.1782697944407419, 6.3111177218703647, 5.8859295634968012, 5.6051160956789534,
       6.7837379878746971, 7.2693979497199983, 7.0596382540082887, 7.3725769872724998, 6.8273909175387031,
       7.2646735267157689, 6.8132180990717908, 6.4999013579017486, 7.4989996813945394, 7.8718420799073021,
       7.6333591899017126, 7.5887614657460158, 7.4640395819273291, 6.3640302683170260, 7.3166408281249629,
       7.2119502094949119, 7.1819033117119586, 7.1170859942333751, 7.0864723098645621, 7.0261901861644045,
       6.9660968411442736, 7.4020566479238177, 6.9233891565177341, 7.4910624164636310, 7.0870390964504217,
       6.8037692530633329, 7.8593700717437436, 8.2104812958560682, 7.8875273805433155, 7.7533566507161700,
       6.8453433645181532, 6.7419750977398483, 6.5734115055771278, 6.4844057370373145, 6.4842165078115110,
       6.4105171309103275, 6.3645970549028847, 7.3128616501582000, 7.3241996346042644, 7.7911511336584525,
       7.3139952233299192, 7.9487538481893880, 7.5564469710738500, 7.2790355744084847, 8.2830470995855325,
       8.8303115369827392, 8.2692513879329006, 8.3516435950308789, 8.3057230684776613, 8.2650946526060824,
       8.2299457744588445, 8.1946076670858030, 8.1679623899095564, 7.7216093931756395, 7.9200306538955090,
       7.8756221589656157, 8.0315245137389741, 7.9990212403867806, 7.9761551411772977, 7.9551795320427470,
       7.6337367472617679, 8.1324359544022631, 8.1269555155863706, 8.0107372327386752, 7.9120938395103764,
       7.7745214890765242, 6.9540023903407713, 6.7244006586662293, 6.4668312979636955, 7.2402958464287357,
       7.3442308997929002, 7.3204204566375015, 7.3606717656948000, 8.0619489688819836, 7.6645396608563843,
       7.3746553549359088, 8.3374712271097433, 8.6298114580536840, 8.3008103173391792, 7.9667062951688390,
       8.0296349247555927, 8.0033672049394031, 7.9672739828462493, 8.1789232675413412},
      {5.2717685398660512, 5.7893645379319105, 6.4900749545332346, 6.8937202666406128, 6.5433650583399512,
       6.5936319999915671, 6.1654197783713522, 6.3011020892760472, 5.8747799071849895, 5.5930220954212260,
       6.7659747701210504, 7.2487998979455037, 7.0394182101396252, 7.3585929474856151, 6.8118952966742690,
       7.2518235106463793, 6.8011236482682884, 6.4872401205123866, 7.4827480447184431, 7.8502996841868917,
       7.6114383357296962, 7.5679741847457178, 7.4436303088328630, 6.3439990031283902, 7.2958539976704406,
       7.1915409364004450, 7.1618724970690995, 7.0978107448564023, 7.0675750683934213, 7.0072929446932628,
       6.9462548051814421, 7.3878833789111296, 6.9073262985216655, 7.4780236217142138, 7.0743778590610598,
       6.7903524504080837, 7.8438749014250853, 8.1895056867215175, 7.8657948545055500, 7.7320025831300123,
       6.8268241309528435, 6.7240226507603982, 6.5543254854259594, 6.4651304876603417, 6.4641856931686510,
       6.3927539131566808, 6.3468333866034623, 7.2950979818587776, 7.3053023931331236, 7.7773563230973721,
       7.2981215945596531, 7.9357150534399707, 7.5432189470986293, 7.2652403133016286, 8.2675510281753226,
       8.8093350267566368, 8.2475197629866877, 8.3304787566705247, 8.2845582301173071, 8.2439298142457282,
       8.2087809360984902, 8.1736320579512522, 8.1469867807750056, 7.6996885390036223, 7.9003769460669293,
       7.8571029254003060, 8.0105480035128718, 7.9780447301606792, 7.9551795320427470, 7.9342030218166455,
       7.6118158930897515, 8.1131607050252903, 8.1088138393811171, 7.9927847857592260, 7.8941413925309254,
       7.7569475005486810, 6.9398295718738590, 6.7100390615192893, 6.4511464478734579, 7.2234774231667798,
       7.3260896741334216, 7.3024680096580514, 7.3446089076987313, 8.0498545180784813, 7.6511224076553592,
       7.3601045285631654, 8.3223527130595887, 8.6094026355049920, 8.2783226765813041, 7.9442186544109630,
       8.0077140705835763, 7.9814463507673858, 7.9453531286742329, 8.1587027731269011},
      {5.3658767393159250, 5.6844846900760562, 6.2684100372438261, 6.5909859446507468, 6.6820709819455217,
       6.6414416649823913, 6.2000018699327315, 6.2941100693826053, 5.9072831805371830, 5.6578394128998104,
       6.5650971338319399, 6.9305699550912045, 7.1492112596797375, 6.9517347934515579, 6.8324933484487635,
       7.1887065529253729, 6.7676755804244042, 6.4819489109222985, 7.2716655466092108, 7.5303689310292388,
       7.4292687116859231, 7.2574927281422994, 7.1864389560361603, 6.2289147439981898, 7.0116393597916611,
       6.9003342786282236, 6.8744450172636400, 6.8067928657642085, 6.7744783710920427, 6.7179758759044228,
       7.0853382861470680, 6.9717660586401937, 6.9288691447878508, 7.4020566479238177, 7.0203317394426810,
       6.7576599478300858, 7.6095487467463148, 7.8605036449154619, 7.7680967063147168, 7.5579590026971761,
       6.7225110696828478, 6.6335048505972587, 6.4558708708776873, 6.3598726318986554, 6.3640302683170260,
       6.2820160691248788, 6.2313711195674308, 7.0054029051641056, 7.3788129913542786, 7.2780903293710182,
       7.2607051195232026, 7.1851161536386376, 7.4309695219892768, 7.1681085011508783, 8.0016668451818251,
       8.4249654145720072, 8.1475535673608661, 8.0232092409022346, 7.9771003862147634, 7.9264554366573163,
       7.8907397719242187, 7.8554016645511773, 7.8280003715632684, 7.6760664239824763, 7.6522559808270776,
       7.5500221883120435, 7.6900504637693601, 7.6564136172454482, 7.6337367472617679, 7.6118158930897515,
       7.5834702561559277, 7.8578580401204174, 7.7493882435236037, 7.6909957088068266, 7.5827142403442647,
       7.4474099373454008, 6.7315819077854746, 6.5868287587781529, 6.3593058453127966, 6.9574040109474780,
       7.4058358258905805, 7.3313808837235097, 7.3111608398548471, 7.2580595147281581, 7.5366049351110185,
       7.2607051195232026, 8.0566577592918946, 8.2777558899954435, 8.0594925933127417, 7.8722205383589090,
       7.9976984379892588, 7.9496990932268545, 7.9068026299202874, 7.8421736405759557},
      {5.4815282355777608, 5.6075729217939694, 7.2138397984782934, 6.7758011734895645, 6.8948542903581069,
       6.6781025747529554, 6.2043482850311289, 6.2651971953171461, 5.8853623263651667, 5.6438553731129257,
       7.4867164519110094, 7.2622167006007530, 7.5551241686763282, 7.2062809919989919, 6.9262235399928063,
       7.1781241337451958, 6.7569039320184237, 6.4641856931686510, 8.3263211202521550, 8.0912398508532739,
       7.8380160041575859, 7.6753113092623648, 7.5549353899963005, 6.8959878635298253, 7.3383729036169525,
       7.1860609481303284, 7.1167079863275431, 7.0265681940702356, 6.9832932723120615, 6.9022238676116041,
       7.4963540765994958, 7.2510674948347162, 7.0534022499265090, 7.4001666083946605, 7.0214657631601751,
       6.7489671176332902, 8.6995424277623012, 8.4319569839196724, 8.1432067017166911, 7.9599039550469763,
       7.2712875387033797, 7.1000783417456148, 6.9158298994926568, 6.7971552410757221, 6.7226998483628755,
       6.6869841836297788, 6.6429536966057174, 7.1465656548846930, 7.8111823988470874, 7.5878162207085493,
       7.4033794503213395, 7.2949092031787499, 7.4389063363744095, 7.1631952994666213, 9.0585904662597585,
       8.9591910572197957, 8.5908834019396831, 8.6288662130162166, 8.5680173027301993, 8.5149159776035113,
       8.4720195142969441, 8.4289338217645735, 8.3958637618265204, 8.0022336317676839, 8.3533448558800085,
       8.1483095831725283, 8.2108588532161235, 8.1622936229594121, 8.1324359544022631, 8.1131607050252903,
       7.8578580401204174, 8.0218864385047119, 7.8552124353253738, 7.7635615125362909, 7.6458311980652711,
       7.4823700368126111, 6.8183200794360754, 6.7128734449943614, 6.5339166628772682, 7.0048361185782468,
       7.8637169373879168, 7.6675628230114841, 7.4838816178901615, 7.4018674186980142, 7.5655178091764776,
       7.2771455348793275, 9.0272207660792834, 8.8665939883016947, 8.7182508905534135, 8.2382601462040324,
       8.2514881701792540, 8.2033004972825978, 8.1513327453276290, 8.3034559221342246},
      {5.3968684315905691, 5.4896538286429211, 7.2423746646379206, 6.6975666028099559, 6.6285916489130017,
       6.5960883755608082, 6.1219560779331506, 6.1831834466707747, 5.8063717398738950, 5.5684556364541651,
       7.4916296535952664, 7.1904073632287266, 7.3663405326449443, 7.1359832357045159, 6.8438313328948279,
       7.0923303060405098, 6.6701657603678228, 6.3763134978005560, 8.2796454789788250, 8.0914281789875258,
       7.7527898641303103, 7.6144614978847960, 7.4908736377836034, 6.8423197518172776, 7.2686419339083352,
       7.1144403894383306, 7.0450874276355453, 6.9523015800374175, 6.9054367095382840, 6.8251233206494897,
       7.3037908120555732, 7.1828485567494251, 6.9719548373202214, 7.3156960336332721, 6.9351051488696296,
       6.6601501277735045, 8.5323016379971026, 8.4124934064084478, 8.0385160830866411, 7.8391504784208559,
       7.3020900017522195, 7.0380954077421034, 6.8489337638048884, 6.7253459037036958, 6.6465440958924518,
       6.6089383916301978, 6.5601834816019071, 7.0562366334015811, 7.6186191343031657, 7.5509669828037342,
       7.3107828319490151, 7.2026905927122566, 7.3438528918870682, 7.0636066612008550, 8.8989080324280856,
       8.9111917124573914, 8.5413720255539545, 8.6061902441240896, 8.5472300217299022, 8.4960191866781454,
       8.4565234428867342, 8.4147605527518863, 8.3830132952113541, 7.8941413925309254, 8.2720862219537477,
       8.1188294719754346, 8.2025444814709356, 8.1558679391060540, 8.1269555155863706, 8.1088138393811171,
       7.7493882435236037, 7.8552124353253738, 7.7440970339335156, 7.6191859208890254, 7.5239436977216565,
       7.3587817261656436, 6.6945429901090803, 6.6015688143766997, 6.4350835898773893, 6.9078930851075242,
       7.6658624632539061, 7.6257999328766353, 7.3905294342519499, 7.3096488082315210, 7.4717876176324340,
       7.1790689282368865, 8.8609252213515504, 8.8308783235685979, 8.6180954657017885, 8.0987982067867996,
       8.1373487055407434, 8.0912398508532739, 8.0430512768650662, 8.2949523211632314},
      {5.3120198489233497, 5.3894984037912961, 7.0826926813520243, 6.6338828585030898, 6.4904529624390657,
       6.3232121726738679, 6.0416426890443562, 6.1087280539579289, 5.7313500111209654, 5.4934334571554597,
       7.3364828640877953, 7.1673524853392152, 7.2308474509660527, 6.9931196756805756, 6.7646519677235277,
       7.0186309291393272, 6.5940095573516233, 6.2977009192151163, 8.1348923299715032, 7.9778564020264264,
       7.7074361241629514, 7.5400065557177252, 7.4373943047510833, 6.8655634083868167, 7.2438866962612467,
       7.0518902183031837, 6.9838600588979203, 6.8899406381280741, 6.8402409336080927, 6.7612503471168210,
       7.1637620860524809, 7.0405517833113436, 6.8929642508289497, 6.7518019516541381, 6.8604609774767562,
       6.5826715729055589, 8.3843360976088768, 8.2998650723017136, 7.9759659119514934, 7.7730094574531980,
       7.1936197546096299, 7.0715434755859876, 6.8294697357478880, 6.7068262195926103, 6.6285916489130017,
       6.5926867549541006, 6.5427982717540916, 6.9836712802178926, 7.4772676059025507, 7.3653957381532535,
       7.2263118066418519, 7.1212431801059690, 7.2661851077933193, 6.9823484778203708, 8.7528329821147928,
       8.7983750502164035, 8.4155165685635485, 8.4954524000922866, 8.4359253911122387, 8.3839585402488215,
       8.3440852390973532, 8.3026999063225606, 8.2701966329703662, 7.8049459442195328, 8.2010324498476095,
       7.9903284101899850, 8.0872714436607076, 8.0402173439357707, 8.0107372327386752, 7.9927847857592260,
       7.6909957088068266, 7.7635615125362909, 7.6191859208890254, 7.5315029547467338, 7.4341819133701801,
       7.2688307125883629, 6.5993007669417114, 6.5127513739711382, 6.3528806120052135, 6.8330601350346232,
       7.5180857015457088, 7.4389063363744095, 7.3929858098211909, 7.2282018461710091, 7.3960094225220665,
       7.0997003338397837, 8.7129596809633245, 8.7195736929509362, 8.5476084801815091, 8.0207528653329945,
       8.0717753722504977, 8.0273668773206044, 7.9795567617840044, 8.1749548603487749},
      {5.2300056497312024, 5.3059726235215985, 6.9400183505538866, 6.5770023554096388, 6.3825495018823357,
       6.1247909119539976, 5.9679433121431735, 6.0439102859335696, 5.6640754169815892, 5.4244585032585055,
       7.1941865411954895, 7.0660634873158710, 7.1182195674050934, 6.8527133923174270, 6.6900077963306543,
       6.9555144219640965, 6.5280582161555447, 6.2289147439981898, 7.9903284101899850, 7.8797788942924347,
       7.6042575371562258, 7.4343706920502077, 7.3153180257274411, 6.7064482116867792, 7.1427860263721543,
       6.9919861025088563, 6.9266015478986374, 6.8321153405429325, 6.7807148257195973, 6.7032358203058751,
       7.0465990087130956, 6.8993894841365329, 6.8177532928502167, 6.6811261874538310, 6.7960212173582280,
       6.5157749866720138, 8.2367481145807062, 8.1995213193158349, 7.8678736727147358, 7.6601927952122111,
       7.0284577830536170, 6.9029798834232672, 6.7107946267851766, 6.5832383594914177, 6.5006573737134117,
       6.4611620804677754, 6.4044703560543521, 6.9205543224968862, 7.3574589237681209, 7.2253670121501612,
       7.1475104493763837, 7.0471657952989544, 7.2009897824089037, 6.9143183184151074, 8.6048674417265669,
       8.6951955621181263, 8.3074243293267909, 8.3947292875632495, 8.3353924089005584, 8.2836138861713913,
       8.2439298142457282, 8.2027337106967391, 8.1706079947046000, 7.7219869505356948, 8.0946414714599797,
       7.8754329297398122, 7.9882495919808001, 7.9415730496159185, 7.9120938395103764, 7.8941413925309254,
       7.5827142403442647, 7.6458311980652711, 7.5239436977216565, 7.4341819133701801, 7.3614273309606872,
       7.1977773910279996, 6.5191766072787214, 6.4354615977832204, 6.2784256698381435, 6.7710772010311109,
       7.3918522366494717, 7.2964207842563003, 7.2249890042443292, 7.1535576747781349, 7.3317588916293417,
       7.0333709847378740, 8.5663169429726214, 8.6171502206643211, 8.4425394030998504, 7.9094482347153319,
       7.9661395085829794, 7.9219202428788904, 7.8739208981164870, 8.0738541904596826},
      {5.0856305086297118, 5.2163996178501497, 6.8100052571451117, 6.4617288670536350, 6.2538592108710827,
       5.9923209924302068, 5.8320722225584500, 6.0072498267087813, 5.6196669220516968, 5.3719239647176771,
       7.0634174319750516, 6.9432311924805656, 6.9783800706278045, 6.7107946267851766, 6.5410974614507378,
       6.4415088231849715, 6.4630516694511568, 6.1618293790846170, 7.8474648501660447, 7.7435302473476559,
       7.4687640049315585, 7.2996331756372035, 7.1813365251260999, 6.5684983038928708, 7.0112613518858291,
       6.8910746618455683, 6.8230445024403048, 6.7236450934003420, 6.6673309263469749, 6.5909859446507468,
       6.9063815040299747, 6.7582267344159455, 6.6686537287444967, 6.6302924592163555, 6.7261014689695831,
       6.4437764200741841, 8.0857594120373815, 8.0547681703085132, 7.7238765395190763, 7.5154400967506643,
       6.8799254560795324, 6.7544475564491826, 6.6476781196099459, 6.5673647307211525, 6.4120291625336536,
       6.3672426596979292, 6.3001572947843565, 6.8033912451575009, 7.2119502094949119, 7.0823151239919682,
       6.9972773120989453, 6.9829152644062296, 7.1184087966308969, 6.8289024986162525, 8.4466970395182202,
       8.5959853823039669, 8.1575691999551836, 8.2478982214382945, 8.1896940148557693, 8.1384831798040143,
       8.0993649933726584, 8.0587365775010795, 8.0271776480948009, 7.5715650345782288, 8.0028004183535426,
       7.7399393975151449, 7.8504889134126961, 7.8049459442195328, 7.7745214890765242, 7.7569475005486810,
       7.4474099373454008, 7.4823700368126111, 7.3587817261656436, 7.2688307125883629, 7.1977773910279996,
       7.0845827208811807, 6.4220447951279711, 6.3402193746158515, 6.1814826363674218, 6.6626069538885213,
       7.2425638938637240, 7.1520456431548096, 7.0747558669668908, 7.0866610885445906, 7.2480438821338407,
       6.9470103704473285, 8.4085249992158815, 8.4661615181209964, 8.3049679537575489, 7.7709306392440132,
       7.8264892410314939, 7.7822690742358533, 7.7350266463766637, 7.9294794999039677},
      {5.2713905319602192, 4.8847524217948255, 6.8944762824522750, 5.6997915322604626, 5.9086059829347057,
       6.0707447923356188, 5.6773038915025866, 5.5882976724169966, 5.3284602642794745, 5.1515818498257904,
       7.1244560220326481, 6.1314049239416084, 6.6592053332818137, 6.5565930823151719, 6.4743901044429970,
       6.3320937815506912, 6.0410759024584975, 5.8294266177634064, 7.8416068539900969, 6.9160191287184611,
       6.6350168822205839, 6.4626736615453257, 6.3889742846441422, 6.6862281678181157, 6.1552153670970071,
       6.0660203693313894, 6.0314382777700111, 5.9562273197912781, 5.8955671881852876, 5.8590955076405278,
       6.6274580757412833, 6.6302924592163555, 6.6278356331013386, 6.5212554254879063, 6.2587724125553397,
       6.0626187487246828, 8.0523108936477215, 7.2030686006180886, 6.9897185056196438, 6.7361171015639005,
       6.9778132840419449, 6.8587601671734024, 6.6864173970439191, 6.5970331700524989, 6.5431762796599227,
       6.4690988948529080, 6.4263912102263694, 6.1092948405437886, 6.9396407931938313, 6.9626952205375661,
       6.9685532167135147, 6.8402409336080927, 6.5760571103721723, 6.3753682527630895, 8.3792341172445912,
       7.6581139770030262, 7.3720097501408643, 7.3595377419773058, 7.3009564285805011, 7.2505007082488575,
       7.2098713912857271, 7.1688645169625413, 7.1352272198928528, 6.6034584033600812, 6.5900411501590561,
       7.3774901889567568, 7.0127729329633794, 6.9772460469103095, 6.9540023903407713, 6.9398295718738590,
       6.7315819077854746, 6.8183200794360754, 6.6945429901090803, 6.5993007669417114, 6.5191766072787214,
       6.4220447951279711, 6.5112397928935888, 6.4039031189227176, 6.2977009192151163, 6.0722563734131692,
       7.0150405298525920, 7.0653074715042088, 7.0743778590610598, 6.9453095601439756, 6.6818822032654941,
       6.4742013257629694, 8.3302895274447213, 7.5870611059884387, 7.7457973936910927, 7.1505340620772593,
       7.1167079863275431, 7.0573706571190762, 7.0173081267418045, 7.0602050405941474},
      {5.2009035464399398, 4.8233367249229486, 6.7542583272233792, 5.5762032216134942, 5.8110961628781244,
       5.6990355164487996, 5.6019037042980493, 5.3551055414557203, 5.2706345161485570, 5.0895989158222781,
       6.9780020627219725, 6.0119742497130106, 6.5248458247746415, 6.4602172859760847, 6.3940767155542027,
       6.0873739863717713, 5.9764473636599407, 5.7649868576448782, 7.6624608426471994, 6.6730001438428941,
       6.4711777130620938, 6.3133853187595781, 6.2001906486127591, 6.5728447189912691, 6.0291706808807977,
       5.9515028967870487, 5.8999136032836859, 5.8823387136642911, 5.8303714122550971, 5.7417432010753391,
       6.4942321404058285, 6.5312706075364479, 6.5456326552291637, 6.2759688437231267, 6.1911202610559073,
       5.9961006209427445, 7.8673068861288762, 6.9589155920250283, 6.7693763907277571, 6.5894743635731974,
       6.8644293846693225, 6.7538803193175472, 6.5635851022086138, 6.4717444996479525, 6.4192099611071232,
       6.3804706836731508, 6.3353061729315945, 6.0440995151593730, 6.8009348695882608, 6.8595161829850655,
       6.8848386577637894, 6.6021356009625594, 6.5038697650943149, 6.3050704964686135, 8.1832692320939628,
       7.4156626798048704, 7.1418412318804636, 7.1097159664341012, 7.0537802578323410, 7.0054029051641056,
       6.9655300545584140, 6.9258455320869752, 6.8929642508289497, 6.3965335416692195, 6.3370069832349483,
       7.2217766128634260, 6.7784467782846090, 6.7453767183465558, 6.7244006586662293, 6.7100390615192893,
       6.5868287587781529, 6.7128734449943614, 6.6015688143766997, 6.5127513739711382, 6.4354615977832204,
       6.3402193746158515, 6.4039031189227176, 6.3353061729315945, 6.2266471471089773, 5.9883525852376405,
       6.8744450172636400, 6.9591048212508317, 6.9883957032221220, 6.7083382512159364, 6.6053479923434626,
       6.3997459330501227, 8.1437743893941033, 7.3427193187153499, 7.5332033145043109, 6.9691204538451492,
       6.9216883462143803, 6.8623510170059134, 6.8236112890261644, 6.8394849177964296},
      {5.1404326440597536, 4.7588969648044204, 6.6040251899459408, 5.4293717049427626, 5.7065943229281011,
       5.9445108768936068, 5.5323615132694606, 5.2583512866650262, 5.2107304003542296, 5.0278047604987943,
       6.8200208897394292, 5.7789708974317620, 6.3772582922922467, 6.3640302683170260, 6.3211333544646831,
       5.9811717866641700, 5.9070944018571554, 5.6969566982396147, 7.4640395819273291, 6.3980451227467698,
       6.2204106924814218, 6.0762247806057355, 5.9722897272415709, 6.4579492385410964, 5.8080725501772488,
       5.7338063866902065, 5.6873190735511283, 5.6948783305762056, 5.6417770054495167, 5.6087064949656877,
       6.3502350072101699, 6.4320604277222886, 6.4709884838362894, 6.1667425807688740, 6.0112182339013476,
       5.9244796117049709, 7.6641612024047774, 6.6835825630230712, 6.5244678168688095, 6.3681874541896200,
       6.7476443152357684, 6.6488116927816652, 6.4411308152791404, 6.3485341969068161, 6.2990237216126381,
       6.2631188276537371, 6.2438435782767643, 5.8759139309024837, 6.6505125030850181, 6.7548255643550137,
       6.8077376602558992, 6.4889409308157404, 6.3207553465588511, 6.2270251550148084, 7.9667062951688390,
       7.1435420421838174, 6.8844606498579575, 6.8273909175387031, 6.7752343869037048, 6.7293143108962621,
       6.6907638121423174, 6.6525913212942038, 6.6208436132078967, 6.1754354109656697, 6.0777368122290607,
       7.0048361185782468, 6.5163417732578734, 6.4861065473406683, 6.4668312979636955, 6.4511464478734579,
       6.3593058453127966, 6.5339166628772682, 6.4350835898773893, 6.3528806120052135, 6.2784256698381435,
       6.1814826363674218, 6.2977009192151163, 6.2266471471089773, 6.1537037860194568, 5.8951891802794565,
       6.7228890775886798, 6.8504453448824378, 6.9078930851075242, 6.5898523714790285, 6.4178871587096014,
       6.3164089314604537, 7.9411954922558632, 7.0688978707909431, 7.2935864007812281, 6.7682423670102629,
       6.7039918361175381, 6.6457880800807896, 6.6068600239667887, 6.5981671937699931},
      {4.7838414316773132, 4.9329405452372566, 6.1858290514658192, 6.1036256230478685, 5.8751579150908206,
       5.6582174208056415, 5.5102518804174156, 5.3524599366606767, 5.3573731383449337, 5.1430782488547981,
       6.4239343841113525, 6.5148301921803231, 6.4919645435166160, 6.3188657575754696, 6.1996238620268995,
       6.0790596146265825, 6.1750574030598386, 5.9140864217505973, 7.0851495074670403, 7.1951313356871802,
       6.9628844497633704, 6.8226664945344737, 6.7230778562687075, 6.0569495312287627, 6.5768131261838354,
       6.4696656814387676, 6.3957775258575564, 6.3570382484235832, 6.3050704964686135, 6.2379855821008166,
       6.4133519649311754, 6.3540146357227076, 6.3192437654813016, 6.2610400094445522, 6.4316824198164575,
       6.1896082294325820, 7.3134284367440596, 7.4687640049315585, 7.1843601378269755, 7.0201429607626524,
       6.3632742525053629, 6.2737012468339142, 6.1383969438350503, 6.0601619226096659, 6.0080058425204443,
       5.9635973475905510, 5.9327944339959355, 6.4433984121683530, 6.6915193774082038, 6.6571265150726289,
       6.6359616767122747, 6.5991119882616838, 6.5726559403112415, 6.5554595091434535, 7.6382719410401938,
       7.8913065585100783, 7.5736438527874137, 7.6439416090818897, 7.5946194619219636, 7.5498329590862392,
       7.5154400967506643, 7.4801024399233986, 7.4519451311238267, 7.0329929768320429, 7.2240442097526385,
       7.1480772359622433, 7.3019012230721918, 7.2652403133016286, 7.2402958464287357, 7.2234774231667798,
       6.9574040109474780, 7.0048361185782468, 6.9078930851075242, 6.8330601350346232, 6.7710772010311109,
       6.6626069538885213, 6.0722563734131692, 5.9883525852376405, 5.8951891802794565, 6.3492902127184792,
       6.7004014368308038, 6.7070154488184137, 6.6994566423391131, 6.6898190176506267, 6.6781025747529554,
       6.6695989737819632, 7.6535787832246003, 7.8508664707727513, 7.7369162353600451, 7.3005784206746700,
       7.3302468600060156, 7.2909407959861836, 7.2474770955479810, 7.3916630074236682},
      {5.6412097683178821, 5.2687449271651756, 6.6677089342528060, 6.3292589475298433, 6.9840492881237237,
       6.3574162563294152, 5.9654869365739325, 5.6795714883917992, 5.5856520676219530, 5.3654987314100939,
       6.8572485860958521, 6.6675201555727783, 7.8444416880109440, 7.2299026564743620, 6.8154856959610042,
       6.5246565955488371, 6.3268025719606031, 6.1818606442732529, 7.3769234023708972, 7.2491779058513348,
       7.2903735588545482, 7.0507566451314654, 6.9116727136200629, 6.6473001117041148, 6.7449987104407239,
       6.6437097124173796, 6.5788919443930203, 6.5307038209505892, 6.5091609746844030, 6.4553036337460519,
       7.7915286910185078, 7.2835707681869097, 6.9572152322674503, 6.7278022792729368, 6.5758683316921447,
       6.4643744718486795, 7.6613272694754810, 7.5851706159135057, 7.5900842681435385, 7.4477879452512328,
       7.1238887849010135, 6.9759232445127877, 6.8111392808626059, 6.7347942991663787, 6.6926534011256988,
       6.6607169143593641, 6.6429536966057174, 6.7376286826414509, 8.0997434518242653, 7.6414852335126486,
       7.3455537021904220, 7.1320148285119496, 6.9891512684880093, 6.8810590292512508, 8.0020444025418804,
       8.2511106128191987, 8.1743871726713628, 7.7231214247989648, 7.6745552934507018, 7.6276895218600167,
       7.5910295131810051, 7.5547466113162720, 7.5275345475541675, 7.3958206438420380, 7.3368608719936264,
       7.3241996346042644, 7.3969542170137572, 7.3653957381532535, 7.3442308997929002, 7.3260896741334216,
       7.4058358258905805, 7.8637169373879168, 7.6658624632539061, 7.5180857015457088, 7.3918522366494717,
       7.2425638938637240, 7.0150405298525920, 6.8744450172636400, 6.7228890775886798, 6.7004014368308038,
       8.1498207137043028, 7.7269006027657277, 7.4317250872551632, 7.2353826447444787, 7.1044247568440131,
       6.9961432883814512, 8.0492877314926208, 7.9846587421482891, 8.0729089454222152, 7.8584257277978287,
       7.7919071494701155, 7.7340814013391972, 7.6889168905976417, 7.5426521605127697},
      {5.8883859390660422, 5.3823176052178257, 7.6985549658319039, 6.3492902127184792, 6.7459435049324146,
       6.6832050056630159, 6.2085054709037228, 5.8655202904023351, 5.6602962390148264, 5.5164878844991954,
       7.8295124031865937, 6.6745121754662202, 7.4602599534147913, 7.6025562763070971, 7.0881731201679159,
       6.7281802871787679, 6.4893189387215715, 6.3165981606862571, 8.1878044258723879, 7.2353826447444787,
       7.1473216706963560, 7.1637620860524809, 6.9744116634352373, 6.9662856198243013, 6.7638959519118655,
       6.6558037126751071, 6.5919311896882133, 6.5477114734383486, 6.5222002199795970, 6.4819489109222985,
       7.4183082845999149, 7.6715312302040504, 7.2389730440312139, 6.9356723860012650, 6.7370618960555912,
       6.5921199683682419, 8.3491872194616388, 8.0343584466682714, 7.4997552466604267, 7.4050802606246933,
       7.4615827558123131, 7.2508782656089128, 7.0401742259512883, 6.9436092003863976, 6.8929642508289497,
       6.8587601671734024, 6.8394849177964296, 6.7385739276789174, 7.6841924675934123, 7.9952411613284662,
       7.5934858887502452, 7.3094600295514933, 7.1189755832167565, 6.9774352761361138, 8.5491205118048335,
       8.4723970716569994, 7.8555908937769807, 8.1643724411685969, 8.1244991400171305, 8.0887825741924821,
       8.0591141348611348, 8.0277444346806597, 8.0037456633910100, 7.2992556182771473, 7.2656183212074597,
       7.9873043469433336, 7.8988658155351548, 7.8744876847023457, 7.3204204566375015, 7.3024680096580514,
       7.3313808837235097, 7.6675628230114841, 7.6257999328766353, 7.4389063363744095, 7.2964207842563003,
       7.1520456431548096, 7.0653074715042088, 6.9591048212508317, 6.8504453448824378, 6.7070154488184137,
       7.7269006027657277, 8.0812242182589564, 7.6832481236474974, 7.4109382568006410, 7.2282018461710091,
       7.0859055232787034, 8.5976857420615449, 8.3580692788842370, 8.1768444493321564, 7.8238436362364494,
       7.6983657366061005, 7.6475315578228491, 7.6120051223155549, 7.5228101245499381},
      {5.9367628411885018, 5.3526487153407043, 7.1308808047944554, 6.3848170987715482, 6.5444990820574453,
       6.7884624108789264, 6.2765356303089863, 5.9120076035414124, 5.6969566982396147, 5.5493691657572199,
       7.2805471554860341, 6.7278022792729368, 7.2057142054131331, 7.2469098584163465, 7.1800137227285772,
       6.7877068456130392, 6.5269246429838264, 6.3381410069524424, 7.6824921078358344, 7.2949092031787499,
       7.1756673076301798, 7.0596382540082887, 7.0660634873158710, 7.2106274070973901, 6.8134068777518184,
       6.6968110375440686, 6.6308592458022142, 6.5862619721922941, 6.5569710902210039, 6.5210661962621028,
       7.1586596551424195, 7.3098380374573244, 7.3391284688828389, 7.0023792924632300, 6.7784467782846090,
       6.6140408225402583, 7.8561576803628395, 7.5598490422263334, 7.4772676059025507, 7.3578369316739529,
       7.5570142082054854, 7.4517563524437991, 7.2134617905724623, 7.0989443180281206, 7.0384734156479345,
       6.9982221065906360, 6.9730888610377155, 6.7457547262523869, 7.4045130234930578, 7.6084142724830448,
       7.6834364517817493, 7.3640729357557317, 7.1448648445813392, 6.9823484778203708, 8.0814134474847599,
       7.9958088490058774, 7.8236544070106460, 7.7129165629788439, 7.6685080680489506, 7.6286347668974832,
       7.5955647069594301, 7.5621161885697701, 7.5350933540334681, 7.2399178385229046, 7.2036353872039474,
       7.4557247596363654, 7.4096154544031192, 7.3803245724318289, 7.3606717656948000, 7.3446089076987313,
       7.3111608398548471, 7.4838816178901615, 7.3905294342519499, 7.3929858098211909, 7.2249890042443292,
       7.0747558669668908, 7.0743778590610598, 6.9883957032221220, 6.9078930851075242, 6.6994566423391131,
       7.4317250872551632, 7.6832481236474974, 7.7809462718383307, 7.4715983884066306, 7.2580595147281581,
       7.0944091242496956, 8.1364043615948294, 7.9222978002389457, 8.1299786777414713, 7.7546794531136918,
       7.6675628230114841, 7.6180523477173061, 7.5776122599799800, 7.5335813224101429},
      {5.6056828822648122, 5.3307278611686870, 6.7174086387727874, 6.3865179090749020, 6.4222335738079988,
       6.2670872348463034, 6.1159088525313985, 6.0388083055692841, 5.7390975962802946, 5.5393535331629025,
       6.8833266261404633, 6.7208102593794941, 7.0314813957544926, 7.0123949250575475, 6.9237667138777903,
       6.8944762824522750, 6.5658526990978272, 6.3290701688498157, 7.3236328480184056, 8.0232092409022346,
       7.1352272198928528, 7.0224105576518658, 6.9339715756979112, 6.8323041192229601, 6.7918640314856331,
       6.7355503149780418, 6.6692209658761321, 6.6045919765317995, 6.5611282760935978, 6.5201214017704121,
       6.9798916517053540, 7.0660634873158710, 7.0653074715042088, 7.1057475592415349, 6.8177532928502167,
       6.6064820160609568, 7.5211093142465844, 8.2237093198312898, 8.0884050168324251, 7.2905627880803525,
       7.1915409364004450, 7.0915747407746235, 6.9955765017955915, 6.9192315200993644, 6.7591719794534120,
       6.6764017644496017, 6.6212216211137278, 6.7347942991663787, 7.2172414190850001, 7.3478212990796345,
       7.3873165923252708, 7.4602599534147913, 7.1800137227285772, 6.9723328452260525, 7.7790566828549492,
       8.5476084801815091, 8.3822572793996919, 8.3575015912068267, 8.3206523533020107, 8.2883383091756215,
       8.2620705893594302, 8.2354253121831853, 8.2138829164627758, 7.9436518678251034, 7.9438410970509068,
       7.3447976863787590, 8.1059799064518199, 8.0816017756190117, 8.0619489688819836, 8.0498545180784813,
       7.2580595147281581, 7.4018674186980142, 7.3096488082315210, 7.2282018461710091, 7.1535576747781349,
       7.0866610885445906, 6.9453095601439756, 6.7083382512159364, 6.5898523714790285, 6.6898190176506267,
       7.2353826447444787, 7.4109382568006410, 7.4715983884066306, 7.5691086590089878, 7.2943419660471154,
       7.0853382861470680, 7.8434964429734784, 8.5079244082558443, 8.5812453267054210, 8.2569686089951464,
       8.2344809682372695, 8.1997096474500868, 8.1634271961311313, 8.1879936550981913},
      {5.3645539369184032, 5.2704457374685285, 6.4872401205123866, 6.7261014689695831, 6.3283141530381526,
       6.1389637304209099, 5.9743685454507558, 5.8581502626030613, 5.7436332406044963, 5.5113859041349098,
       6.6644965428719027, 7.0825039026719958, 6.9258455320869752, 6.8527133923174270, 6.7383846984531139,
       6.6563704992609658, 6.5690655410245062, 6.3007240813702161, 7.1314475913803150, 7.6218315256840690,
       7.4772676059025507, 7.3652069594732259, 7.2794131317685400, 6.6006235693392332, 7.1488332517739055,
       7.0692758786967751, 7.0037020948607527, 6.9515460147715302, 6.8920194563372590, 6.8423197518172776,
       6.8666969815585350, 6.9010902944398858, 6.8716106337885678, 6.8555477757924992, 6.8238005182519679,
       6.5809707626022051, 7.3311916544977063, 7.8325364664332451, 7.6979872781544927, 7.5674073981598582,
       6.9623172126317350, 6.8580046019075152, 6.7540695485433515, 6.6730001438428941, 6.6091276208560013,
       6.5142634055944644, 6.4314931905906541, 7.0167413401559457, 7.0972439582705427, 7.1745337344584614,
       7.1819033117119586, 7.1938085332896575, 7.1837933512411158, 6.9441759869722564, 7.6008559165495191,
       8.1649392277544557, 8.0024228609934873, 7.9729427497963936, 7.9349590376283077, 7.9011329618785924,
       7.8737325699822343, 7.8463312769943254, 7.8236544070106460, 7.5532345796929468, 7.5481321487828863,
       7.4818027996809766, 7.7100817289579959, 7.6843816968192158, 7.6645396608563843, 7.6511224076553592,
       7.5366049351110185, 7.5655178091764776, 7.4717876176324340, 7.3960094225220665, 7.3317588916293417,
       7.2480438821338407, 6.6818822032654941, 6.6053479923434626, 6.4178871587096014, 6.6781025747529554,
       7.1044247568440131, 7.2282018461710091, 7.2580595147281581, 7.2943419660471154, 7.3026567883380791,
       7.0619058508975012, 7.6673744948772322, 8.1369711481806881, 8.2346701974630729, 7.8867713647316524,
       7.8586140559320805, 7.8198752290438831, 7.7834035484991233, 7.8000331930810516},
      {5.1801167159854176, 5.1946675423581610, 6.3067713067719673, 6.4156195618203888, 6.2504575902643751,
       6.0301154753724884, 5.8507806853495632, 5.7022479078297028, 5.7220903943383110, 5.4667881799792140,
       6.4870513418323590, 6.7809036043996249, 6.8459101511040128, 6.7304478840679804, 6.5940095573516233,
       6.4706104759304583, 6.3572270271036118, 6.2587724125553397, 6.9570260030416469, 7.3300580813259879,
       7.1868169639419914, 7.0700318945084373, 6.9804588888369894, 6.4269579968122281, 6.8455321431981808,
       6.7646519677235277, 6.6977558320357593, 6.6403080918106729, 6.5994899961675149, 6.5463886710408268,
       6.7765571893012275, 6.7744783710920427, 6.7217550538711848, 6.6626069538885213, 6.6019463717367559,
       6.5437430662457823, 7.1571480740648701, 7.5430301684186007, 7.4169854822023922, 7.2811139420718938,
       6.7901632211822802, 6.6811261874538310, 6.5681202959870397, 6.4828937054139892, 6.4161863484062476,
       6.3581718215953025, 6.3097949194728429, 6.7344162912605467, 7.0020017351031747, 7.0422525936146974,
       7.0254341703527414, 6.9912300866971941, 6.9494671965623454, 6.9084603222391596, 7.4324811030668263,
       7.8716537517730494, 7.7250110137823462, 7.6875940882001199, 7.6494220478977812, 7.6154067429222625,
       7.5876278925742975, 7.5600373703605852, 7.5367941643368219, 7.2754447245759737, 7.2487998979455037,
       7.2280126169452057, 7.4207646601691550, 7.3950646280303758, 7.3746553549359088, 7.3601045285631654,
       7.2607051195232026, 7.2771455348793275, 7.1790689282368865, 7.0997003338397837, 7.0333709847378740,
       6.9470103704473285, 6.4742013257629694, 6.3997459330501227, 6.3164089314604537, 6.6695989737819632,
       6.9961432883814512, 7.0859055232787034, 7.0944091242496956, 7.0853382861470680, 7.0619058508975012,
       7.0328041981520144, 7.5069364957796720, 7.8595593009695470, 7.9812571215415824, 7.6284455376716798,
       7.5934858887502452, 7.5517229986153964, 7.5150620888448332, 7.5177076936398777},
      {6.4543588392543612, 5.7678216916657252, 8.0390828696724999, 7.2221546207692571, 7.6416735616469014,
       6.7793920233220755, 6.2901416621900390, 5.9870297828401178, 5.7761365139566898, 5.6211789536750221,
       8.2443073716057835, 7.6059578969138038, 8.8716959686659802, 7.8835589733507492, 7.3017119938463884,
       6.9419083900830438, 6.6962438004124332, 6.5155862079919862, 8.8450515925812851, 8.2174728652037352,
       8.0630825420537011, 7.8782677637606602, 7.7594038761179212, 7.7941742958135523, 7.5978318533028677,
       7.5144953022589736, 7.4564807754480285, 7.4109382568006410, 7.3841042009443676, 7.3345932751044138,
       7.5488881645945485, 7.9347707094940558, 7.4657403922306829, 7.1733997107409664, 6.9778132840419449,
       6.8313593247312694, 9.1717846858608016, 8.6250870350494537, 8.3401168319047869, 8.2558350358234271,
       8.2006548924875542, 8.1029558432051694, 7.9682183267921642, 7.9156837882513358, 7.8930069182676554,
       7.8725980957189643, 7.8627716923504503, 7.6996885390036223, 7.9914619833617033, 8.4778775104728918,
       8.0337916600824109, 7.7284126343890529, 7.5354713619393001, 7.3814585961493231, 9.5790203972549133,
       9.5329115425674438, 8.7278880646961241, 8.7074792421474339, 8.6511655256398416, 8.6054342283124257,
       8.5640488955376330, 8.5228527919886439, 8.4933735818831018, 7.9202189820297608, 7.9963756355917361,
       8.6046782125007635, 8.3852813426463424, 8.3571240338467714, 8.3374712271097433, 8.3223527130595887,
       8.0566577592918946, 9.0272207660792834, 8.8609252213515504, 8.7129596809633245, 8.5663169429726214,
       8.4085249992158815, 8.3302895274447213, 8.1437743893941033, 7.9411954922558632, 7.6535787832246003,
       8.0492877314926208, 8.5976857420615449, 8.1364043615948294, 7.8434964429734784, 7.6673744948772322,
       7.5069364957796720, 9.5990516624435500, 9.0485748336654392, 8.7957294454213599, 8.5217192188169246,
       8.4268550035553886, 8.3958637618265204, 8.3510768084450202, 8.5220967761769817},
      {5.7727348933499822, 6.2750240492314360, 7.2059029840931608, 7.5059912507422064, 7.1195423698026152,
       7.0583154516107660, 6.6455993014007610, 6.6658193452694245, 6.2521584005677289, 5.9815493440242262,
       7.4644175898331611, 7.8858261196941859, 7.6747436215849545, 7.8410400674042373, 7.3364828640877953,
       7.6603820244380145, 7.2187530001625504, 6.9129955160175856, 8.2055676436260345, 8.5247423809720253,
       8.2760546291463157, 8.1942301097257477, 8.0787678426897145, 7.0114501305658568, 7.9317466462474044,
       7.8206312448555462, 7.7722534416415359, 7.7100817289579959, 7.6811693054383126, 7.6171071026798396,
       7.5770454733941204, 7.8648505105596351, 7.4379615418827187, 7.8926293609076001, 7.5031568672671343,
       7.2304694430602217, 8.5695293343535255, 8.8864360242645262, 8.5544117213949225, 8.5181283689844154,
       7.4761335821850565, 7.3822146119609862, 7.2034461579781439, 7.1063143458273945, 7.0764566772702446,
       7.0144737432667332, 6.9761124737385911, 7.9274006816947820, 7.9576359076119880, 8.2851259177947174,
       7.8671176569030727, 8.3954853033749135, 8.0243437151655037, 7.7580810737203993, 9.0126699397065391,
       9.5712728120955841, 8.9506865551572510, 9.0255204063217054, 8.9758207018017231, 8.9331125666294078,
       8.8960740994987884, 8.8584679446907586, 8.8306890943427945, 8.3665728798552301, 8.5931505482831199,
       8.5357032586038102, 8.6880147635446576, 8.6536219012090818, 8.6298114580536840, 8.6094026355049920,
       8.2777558899954435, 8.8665939883016947, 8.8308783235685979, 8.7195736929509362, 8.6171502206643211,
       8.4661615181209964, 7.5870611059884387, 7.3427193187153499, 7.0688978707909431, 7.8508664707727513,
       7.9846587421482891, 8.3580692788842370, 7.9222978002389457, 8.5079244082558443, 8.1369711481806881,
       7.8595593009695470, 9.0485748336654392, 9.3152150324706007, 8.9661826265674609, 8.6054342283124257,
       8.6789443759878058, 8.6524883280373626, 8.6128042561116995, 8.8380591221420683},
      {6.3967223203492471, 6.4643744718486795, 6.9927421183205194, 7.3321368995351728, 7.5373609509226815,
       7.3774901889567568, 6.9388847773821682, 6.9063815040299747, 6.5431762796599227, 6.3060152909603042,
       7.2554139099331145, 7.6222099841356767, 8.0073356121319694, 8.1301679069672748, 7.6167295453197843,
       7.8249772094081687, 7.4260558697592440, 7.1539356826839668, 7.9192746380838459, 8.1692851923070791,
       8.0623265262420389, 7.9695411291896869, 7.8389612491950524, 6.9585380346649721, 7.6932637562418149,
       7.5929191021643865, 7.5621161885697701, 7.5110936816522669, 7.4897396140661092, 7.4423075064353403,
       7.9355258242141673, 8.1645607693028488, 7.7140501361505622, 8.0402173439357707, 7.6834364517817493,
       7.4377723126569153, 8.3057230684776613, 8.5370260610013311, 8.2936295187657088, 8.2405281936390207,
       7.4801024399233986, 7.4228434783783399, 7.2546578941214515, 7.1754785289501521, 7.1724549162492766,
       7.1193535911225876, 7.0904407170571284, 7.7595931053437246, 8.2454409447775028, 8.1241206815655236,
       8.0817910048448152, 8.4767439373011726, 8.1331910691223737, 7.8918733450959371, 8.7299668829053090,
       9.1942723266186768, 8.8123590900032873, 8.6696847592051505, 8.6247094776893984, 8.5835133741404093,
       8.5474192509557056, 8.5117035862226071, 8.4850583090463623, 8.1791115956755931, 8.3106367207076932,
       8.2170953078436781, 8.3529663974284016, 8.3221643849253368, 8.3008103173391792, 8.2783226765813041,
       8.0594925933127417, 8.7182508905534135, 8.6180954657017885, 8.5476084801815091, 8.4425394030998504,
       8.3049679537575489, 7.7457973936910927, 7.5332033145043109, 7.2935864007812281, 7.7369162353600451,
       8.0729089454222152, 8.1768444493321564, 8.1299786777414713, 8.5812453267054210, 8.2346701974630729,
       7.9812571215415824, 8.7957294454213599, 8.9661826265674609, 8.8898376448712337, 8.5491205118048335,
       8.4782550678329471, 8.4529334941457748, 8.4123041771826443, 8.5336244403946253},
      {5.8471898355170531, 6.0612959463271601, 6.6427644673799131, 6.9275463423903281, 7.2355718739702821,
       7.0787242741594572, 6.6435204831915762, 6.6512685188966811, 6.2908976780017021, 6.0601619226096659,
       6.9356723860012650, 7.2656183212074597, 7.6231543280815917, 7.4661184001365148, 7.2915075825720432,
       7.5388725320002310, 7.1407076587087452, 6.8719881911486240, 7.8281896007890719, 7.8646612813338317,
       7.7365377769084382, 7.6469647712369895, 7.5477545914228301, 6.6945429901090803, 7.3608605443748285,
       7.2427526725437525, 7.2047689603756666, 7.1424080184663232, 7.1119835633233146, 7.0592602461024567,
       7.5677858566114660, 7.4937084718044513, 7.3918522366494717, 7.7497658008836590, 7.3892066318544281,
       7.1431640342779863, 8.1190187012012380, 8.1881828843239948, 8.0543897118569063, 7.9848479713740925,
       7.2178082056708597, 7.0978107448564023, 6.8999562707223916, 6.7918640314856331, 6.7671087938385446,
       6.6934094169373610, 6.6476781196099459, 7.3436641132070406, 7.8346143835508784, 7.7756550622482425,
       7.7047905193679069, 8.1484979113067801, 7.7869934972400827, 7.5400065557177252, 8.4533110515058301,
       8.7224076258802317, 8.4243986279861485, 8.3514552668966271, 8.3061015269292682, 8.2654722099661377,
       8.2261656954005300, 8.1906392598932367, 8.1628604095452708, 7.9633055756536830, 8.0009108293701612,
       7.8873381513175111, 8.0233984701280381, 7.9893831651525185, 7.9667062951688390, 7.9442186544109630,
       7.8722205383589090, 8.2382601462040324, 8.0987982067867996, 8.0207528653329945, 7.9094482347153319,
       7.7709306392440132, 7.1505340620772593, 6.9691204538451492, 6.7682423670102629, 7.3005784206746700,
       7.8584257277978287, 7.8238436362364494, 7.7546794531136918, 8.2569686089951464, 7.8867713647316524,
       7.6284455376716798, 8.5217192188169246, 8.6054342283124257, 8.5491205118048335, 8.3624152434368604,
       8.2794562497530215, 8.2369373438065097, 8.1947968963116065, 8.1808128565247227},
      {5.7045155047189162, 6.0316275069958145, 6.6380404949214604, 6.9683644380334862, 7.0683310842050844,
       6.9812144541028758, 6.5577271060326661, 6.5928759841799041, 6.2198439058955630, 5.9796597550408448,
       6.9294359313737095, 7.3177748518424570, 7.5475653621970267, 7.7131057922046473, 7.1911629284946139,
       7.4948420449761697, 7.0849607287870127, 6.8075488815758707, 7.6397839726635199, 7.9275890098290347,
       7.8359371859484011, 7.6562243880196448, 7.5730770662015541, 6.6178204510527969, 7.3997890510346052,
       7.2841375547727694, 7.2455870560188247, 7.1824705488435940, 7.1516676352489776, 7.0957319266472174,
       7.4836928392101338, 7.3457429314162255, 7.2875391753794769, 7.7087589265604732, 7.3387509115227836,
       7.0855275153728714, 7.9771003862147634, 8.2588581979785278, 8.1825132162823007, 7.9597147258211729,
       7.1134955949466399, 7.0165525614759181, 6.8302253010137752, 6.7272354926870772, 6.7036138282117070,
       6.6316152616138773, 6.5862619721922941, 7.3841042009443676, 7.7696078368464914, 7.6482875736345113,
       7.6189966916632210, 8.1256327131888479, 7.7573250579087363, 7.5041016617588250, 8.3752657100520249,
       8.8261539005643677, 8.5553560653408383, 8.4245869561204003, 8.3775328563954634, 8.3268888079295671,
       8.2907946847448617, 8.2550790200117650, 8.2273001696637991, 8.0744209770455413, 8.0405949012958260,
       7.9428958520134403, 8.0868929852091007, 8.0526893520993283, 8.0296349247555927, 8.0077140705835763,
       7.9976984379892588, 8.2514881701792540, 8.1373487055407434, 8.0717753722504977, 7.9661395085829794,
       7.8264892410314939, 7.1167079863275431, 6.9216883462143803, 6.7039918361175381, 7.3302468600060156,
       7.7919071494701155, 7.6983657366061005, 7.6675628230114841, 8.2344809682372695, 7.8586140559320805,
       7.5934858887502452, 8.4268550035553886, 8.6789443759878058, 8.4782550678329471, 8.2794562497530215,
       8.4270442327811921, 8.3658168640435662, 8.3257543336662962, 8.2409057509990760},
      {5.6455561834162795, 5.9855177512167925, 6.6085608342701416, 6.9381287615705052, 7.0135289487750425,
       6.9337823464721078, 6.5093502039102074, 6.5528139043484090, 6.1765694346831639, 5.9346844735250919,
       6.9061922748041713, 7.2903735588545482, 7.4903068511977438, 7.6686972972747540, 7.1424080184663232,
       7.4574255699397192, 7.0439534039180511, 6.7637071732318370, 7.6146507271105994, 7.9035902385393841,
       7.7945527542651591, 7.6263667194624949, 7.5449197574019822, 6.5813487705080362, 7.3712541848749771,
       7.2550359020272825, 7.2170521898591966, 7.1528016589664718, 7.1218099666918286, 7.0651182422784053,
       7.4247330673617213, 7.2954759897646095, 7.2387842653511862, 7.6724764752415169, 7.2990663890513439,
       7.0428198307463328, 7.9485655200551362, 8.2327797073881399, 8.1299786777414713, 7.9234322745022157,
       7.0677638470734490, 6.9708212641485030, 6.7841159957805282, 6.6801813929621403, 6.6567485071667978,
       6.5834271381714462, 6.5373178329381991, 7.3487660935713253, 7.7125381045272361, 7.5983986398887264,
       7.5700530029549027, 8.0899170484557512, 7.7182077725689320, 7.4617719850381166, 8.3454080414948759,
       8.7951626588355012, 8.5047120168749419, 8.3986976947558158, 8.3518328242566824, 8.3021331197367019,
       8.2658506684177446, 8.2301341025930963, 8.2025444814709356, 8.0283121223580700, 8.0148948691570467,
       7.9192746380838459, 8.0608153957102644, 8.0264216322831388, 8.0033672049394031, 7.9814463507673858,
       7.9496990932268545, 8.2033004972825978, 8.0912398508532739, 8.0273668773206044, 7.9219202428788904,
       7.7822690742358533, 7.0573706571190762, 6.8623510170059134, 6.6457880800807896, 7.2909407959861836,
       7.7340814013391972, 7.6475315578228491, 7.6180523477173061, 8.1997096474500868, 7.8198752290438831,
       7.5517229986153964, 8.3958637618265204, 8.6524883280373626, 8.4529334941457748, 8.2369373438065097,
       8.3658168640435662, 8.3066683135151269, 8.2666057831378570, 8.2131269006511118},
      {5.6047380877731214, 5.9416764934185347, 6.5753015451062851, 6.8980666817390102, 6.9568372243616183,
       6.8929642508289497, 6.4670200766437231, 6.5178538048811987, 6.1404753114984603, 5.8976455558486975,
       6.8642406059892949, 7.2518235106463793, 7.4449531112303848, 7.2397290598428761, 7.1017791520489686,
       7.4217099052066215, 7.0080485099591501, 6.7270462634612738, 7.5740214101474690, 7.8690081469780058,
       7.7586478603062590, 7.5887614657460158, 7.5071252744597006, 6.5511130940450553, 7.3325144568952290,
       7.2164854032733379, 7.1788796990110830, 7.1136843736266675, 7.0815591081803051, 7.0246781545410784,
       7.3791909992601097, 7.2610831274290337, 7.1989109641997189, 7.6367608105084193, 7.2633507243182471,
       7.0065369288815997, 7.9060466141086252, 8.1949852244458583, 8.0910506216274705, 7.8775117479489971,
       7.0473550245247578, 6.9369951883987868, 6.7463215128382465, 6.6429536966057174, 6.6214108503395321,
       6.5543254854259594, 6.5076493936068536, 7.3062476381705901, 7.6675628230114841, 7.5626829751556297,
       7.5290461286317170, 8.0525001228735249, 7.6804132896266495, 7.4237882728700306, 8.3011878746992345,
       8.7509424920398597, 8.4657830596693895, 8.3609032118135342, 8.3140383413144008, 8.2650946526060824,
       8.2291897586471805, 8.1934740939140838, 8.1658835717003715, 7.9888163785666597, 7.9776671728006230,
       7.8756221589656157, 8.0247212725255608, 7.9903284101899850, 7.9672739828462493, 7.9453531286742329,
       7.9068026299202874, 8.1513327453276290, 8.0430512768650662, 7.9795567617840044, 7.8739208981164870,
       7.7350266463766637, 7.0173081267418045, 6.8236112890261644, 6.6068600239667887, 7.2474770955479810,
       7.6889168905976417, 7.6120051223155549, 7.5776122599799800, 8.1634271961311313, 7.7834035484991233,
       7.5150620888448332, 8.3510768084450202, 8.6128042561116995, 8.4123041771826443, 8.1947968963116065,
       8.3257543336662962, 8.2666057831378570, 8.2276777270238561, 8.1758992042946890},
      {5.4354189303445146, 5.9137084138447662, 6.7270462634612738, 7.0798582978769522, 6.7385739276789174,
       6.7682423670102629, 6.3492902127184792, 6.4296036016072726, 6.0238794712907096, 5.7606404425464799,
       6.9936869128122101, 7.4551579730505058, 7.2688307125883629, 7.5349041248076647, 7.0069149367874317,
       7.3856157820219170, 6.9458767972756101, 6.6423869100198578, 7.7155621677738875, 8.0842473804140553,
       7.8444416880109440, 7.7970091298344002, 7.6668068071998210, 6.5885295690815067, 7.5105268950664072,
       7.3982770194112790, 7.3567029079564579, 7.2901847801745205, 7.2569254910106640, 7.1947537783271240,
       7.1766121021218705, 7.5657070384022811, 7.1061255671473660, 7.6133279247130776, 7.2206430396917076,
       6.9460655759556378, 8.0589249056353314, 8.4202409915677769, 8.0997434518242653, 7.9599039550469763,
       7.0086157470907855, 6.9407743663655497, 6.7652187543093874, 6.6661973531752556, 6.6452212934949300,
       6.5730339482170725, 6.5269246429838264, 7.4744332224274785, 7.5245109348532910, 7.9444078836367664,
       7.4840708471159649, 8.0725313880621599, 7.6891061198234452, 7.4177410474682794, 8.4790110836446111,
       9.0313784024976531, 8.4882707004272646, 8.5683957611818062, 8.5207739737794590, 8.4788227555103575,
       8.4432954189115126, 8.4066345091409485, 8.3792341172445912, 7.9411954922558632, 8.1488763697583870,
       8.0710193564388337, 8.2367481145807062, 8.2027337106967391, 8.1789232675413412, 8.1587027731269011,
       7.8421736405759557, 8.3034559221342246, 8.2949523211632314, 8.1749548603487749, 8.0738541904596826,
       7.9294794999039677, 7.0602050405941474, 6.8394849177964296, 6.5981671937699931, 7.3916630074236682,
       7.5426521605127697, 7.5228101245499381, 7.5335813224101429, 8.1879936550981913, 7.8000331930810516,
       7.5177076936398777, 8.5220967761769817, 8.8380591221420683, 8.5336244403946253, 8.1808128565247227,
       8.2409057509990760, 8.2131269006511118, 8.1758992042946890, 8.3792341172445912}};

  static constexpr double k1_ = 16.0;
  static constexpr double k3_ = 4.0;
};

} // namespace Dftd3
} // namespace Utils
} // namespace Scine

#endif // SCINE_DFTD3PARAMETERS_H
